
bl_info = {
    "name" : "Fence Maker",
    "author" : "Q@StudioPotpourri",
    "version" : (1,3),
    "blender" : (3, 0, 0),
    "location" : "Edit Mode > Createool > Fence Maker",
    "description" : "Make Fance/Torus structure from simple quad meshes.",
    "warning" : "",
    "doc_url" : "http://qcganime.web.fc2.com/BLENDER28/Addon11.html",
    "tracker_url" : "",
    "category" : "Mesh"
}

if "bpy" in locals():
    import imp
    imp.reload(objectmode_module)
    imp.reload(mesh_module)
    imp.reload(pole_module)
    imp.reload(common_module)
    imp.reload(modifier_module)
    imp.reload(translate_module)
else:
    from . import objectmode_module
    from . import mesh_module
    from . import pole_module
    from . import common_module
    from . import modifier_module
    from . import translate_module

import bpy
import math
import numpy as np
import mathutils
from mathutils import Vector, noise
from mathutils.noise import noise_vector
import bmesh
import random
import time
from bpy.props import * 
from math import sin, cos, sqrt, atan2

class MESH_OT_FenceMaker \
(bpy.types.Operator, \
 mesh_module.MeshModule, mesh_module.CurveModule, mesh_module.CurveCapModule, mesh_module.MeshSettingModule, mesh_module.InstancerModule, \
 common_module.CommonGeometryModule, common_module.CommonModule, common_module.LayoutModule, common_module.CollectionModule, common_module.CommonParameterModule, \
 modifier_module.ModifierModule):
    
    bl_idname = "mesh.make_fence"
    bl_label = "Fence Maker"
    bl_description = "Make Fance from simple quad meshes." 
    bl_options = {'REGISTER','UNDO'}

    CommonTypeSelection = common_module.CommonTypeSelection

    ActionType = [("FACE", "Face", "Face", "FACESEL", 1),
                  ("POLE", "Pole", "Pole", "EDGESEL", 2),
                  ("POLE_APPENDIX", "Pole Appendix", "Pole Appendix", "APPEND_BLEND", 3),
                  ("GUARDRAIL", "Guard Rail", "Guard Rail", "PROP_PROJECTED", 4),
                  ("TRUSS", "Truss", "Truss", "LINCURVE", 5),
                  ("CARSTOP", "Car Stop", "Car Stop", "SPHERECURVE", 6)]

    action : EnumProperty(
        name = "Action Type", default = "FACE",
        items = ActionType)

    fit_fence_width : BoolProperty(
        name = "Fit Width", default = True)

    fit_fence_height : BoolProperty(
        name = "Fit Height", default = False)

    h_margin : FloatProperty(
        name = "Margin H", default = 0, 
        soft_min = 0, step = 0.1)
        
    interval_horizontal : FloatProperty(
        name = "Width", default = 0.1,
        min = 0.01, step = 0.1, precision = 3)

    interval_vertical : FloatProperty(
        name = "Height", default = 0.1,
        min = 0.01, step = 0.1, precision = 3)

    interval_mesh_horizontal : FloatProperty(
        name = 'Interval', default = 0.05,
        min = 0.01, step = 0.1, precision = 3)

    interval_mesh_vertical : FloatProperty(
        name = 'Interval', default = 0.075,
        min = 0.01, step = 0.1, precision = 3)

    single_rope : BoolProperty(
        name = "Single", default = True)

    catenary_a : FloatProperty(
        name = "a", default = 10,
        min = 0.2, max = 2000)

    rope_type : EnumProperty(
        name = "Type", default = "ROPE",
        items = [("ROPE", "Rope", "Rope", "REMOVE", 1),
                ("CHAIN", "Chain", "Chain", "LINKED", 2)])

    chain_scale : FloatProperty(
        name = "Scale", default = 0.03,
        min = 0.01, max = 0.3, step = 0.1, precision = 3)

    chain_elongate : FloatProperty(
        name = "Elongate", default = 0.3,
        min = 0, max = 1, step = 0.3)

    rope_terminal_type : EnumProperty(
        name = "Terminal", default = "NONE",
        items = [("NONE", "None", "None", "", 1),
                 ("EYE_TERMINAL", "Eye Terminal", "Eye Terminal", "", 2),
                 ("KNOT", "Knot", "Knot", "", 3)])
    
    rope_knot_radius : FloatProperty(
        name = "Knot Radius", default = 0.05,
        min = 0.01, max = 0.5, step = 0.1)

    rope_terminal_size : FloatProperty(
        name = "Scale", default = 1,
        min = 0.3, max = 3, step = 0.3)

    rope_terminal_rot : FloatProperty(
        name = "Angle", default = 0, subtype="ANGLE",
        min = -3.14, max = 3.14, step = 0.3)

    interval_wire_vertical : FloatProperty(
        name = 'Height', default = 0.2,
        min = 0.02, step = 0.1, precision = 3)

    use_adjust_bezier_tips : BoolProperty(
        name = "Adjust Bezier Tips", default = True,
        description = "Adjust tips of bezier curves")

    fence_distort : FloatProperty(
        name = 'Distortion', default = 0,
        min = 0, max = 2, step = 0.3)

    fence_distort_offset : FloatProperty(
        name = 'Offset', default = 0,
        min = -1, max = 1, step = 0.3)

    fence_distort_freq : FloatProperty(
        name = 'Frequency', default = 0.3,
        min = 0.1, max = 2, step = 0.3 )

    fence_support_wire : BoolProperty(
        name = "Support", default = False)

    fence_support_num : IntProperty(
        name = "Number", default = 1,
        min = 0, max = 5)

    FenceType = [("CROSS", "Cross", "Cross", "MOD_MULTIRES", 1),
                 ("OBLIQUE", "Cross (Oblique)", "Corws (Oblique)", "X", 2),
                 ("WIRE_MESH", "Wire Mesh", "Wire Mesh", "ALIGN_FLUSH", 3),
                 ("ROPE", "Rope", "Rope", "REMOVE", 4),
                 ("BARBED_WIRE", "Barbed Wire", "Barbed Wire", "FREEZE", 5),
                 ("GUARD_PIPE", "Guard Pipe", "GuardPipe", "PROP_PROJECTED", 6)]

    fence_type : EnumProperty(
        name = "Fence Type", default = "CROSS",
        items = FenceType)

    curve_type : EnumProperty(
        name = "Curve Type",
        default = "POLY",
        items = [("POLY", "Poly", "Poly"),
                 ("BEZIER", "Bezier", "Bezier")])
    
    guardpipe_type : EnumProperty(
        name = "Type", default = "TYPEA",
        items = CommonTypeSelection[0:4])
    
    guardpipe_connector : BoolProperty(
        name = "Connector", default = True)

    guardpipe_connector_type : EnumProperty(
        name = "Type", default = "TYPEA",
        items = CommonTypeSelection[0:2])

    v_align_type : EnumProperty(
        name = "V Align", default = "0",
        items = [("0", "Top Base", "Top Base"),
                 ("1", "Bottom Base", "Bottom Base")])

    CrossOrnamentType = [("NONE", "None", "None", "", 0),
                         ("RING", "Ring", "Ring", "MESH_CIRCLE", 1),
                         ("HALF_RING", "Half Ring", "Half Ring", "SPHERECURVE", 2),
                         ("TURN_AROUND", "Turn Around", "Turn Around", "RESTRICT_SELECT_ON", 3),
                         ("QUAD", "Quad", "Quad", "KEYFRAME_HLT", 4),
                         ("TRIANGLE", "Triangle", "Triangle", "MARKER_HLT", 5),
                         ("TRIANGULAR", "Triangular", "Triangular", "LINCURVE", 6),
                         ("CAP", "Cap", "Cap", "MESH_CYLINDER", 7)]
        
    cross_ornament_type : EnumProperty(
        name = 'Ornament', default = "NONE",
        items = CrossOrnamentType)

    CapType = [("SIMPLE", "Simple", "Simple", "MESH_CYLINDER", 1),
               ("TAPER", "Taper", "Taper", "MESH_CONE", 2),
               ("ROUND", "Round", "Round", "SPHERECURVE", 3),
               ("HAT", "Hat", "Hat", "MOD_SHRINKWRAP", 4)]

    cross_cap_type : EnumProperty(
        name = 'Cap Type', default = "SIMPLE",
        items = CapType)

    cross_ornament_support: BoolProperty(
        name = 'Support', default = False)

    cross_ornament_support_num : IntProperty(
        name = 'Support Number', default = 6,
        min = 1, max = 8)

    cross_vertical : BoolProperty(
        name = "Vertical", default = True)

    cross_horizontal : BoolProperty(
        name = "Horizontal", default = True)

    ornament_radius : FloatProperty(
        name = "Radius", default = 0.05,
        min = 0.01, max = 0.5, step = 0.1, precision = 3)
            
    fill_top : BoolProperty(
        name = "Fill Top", default = True)

    fill_bottom : BoolProperty(
        name = "Fill Bottom", default = False)

    bevel_depth_mesh : FloatProperty(
        name = 'Bevel Depth',  default = 0.005,
        min = 0.001, step = 0.02, precision = 4)

    filling_mesh : FloatProperty(
        name = 'Filling', default = 0.5,
        min = 0.25, max = 1, step = 0.3)

    flip_wiremesh : BoolProperty(
        name = 'Flip HV', default = False)
    
    fit_truss : BoolProperty(
        name = 'Fit Length', default = False)

    flip_truss : BoolProperty(
        name = 'Flip', default = True)

    BevelType = [("CYLINDER", "Cylinder", "Cylinder", "MESH_CYLINDER", 1),
                 ("PRISM", "Prism", "Prism", "MESH_CUBE", 2),
                 ("RECT", "Rect Object", "Rect Object", "MATPLANE", 3),
                 ("LSHAPE", "L-Shape Object", "L-Shape Object", "DRIVER_ROTATIONAL_DIFFERENCE", 4)]

    type_bevel_truss : EnumProperty(
        name = "Type", default = "CYLINDER",
        items = BevelType[0:2])
    
    TrussAngleType = [("0", "Aspect Ratio", "Aspect Ratio"),
                      ("1", "Step Length", "Step Length")]

    angle_type_truss : EnumProperty(
        name = "Angle", default = "0",
        items = TrussAngleType)
        
    bevel_depth_truss : FloatProperty(
        name = 'Bevel Depth', default = 0.03,
        min = 0, step = 0.1, precision = 4)

    step_length_truss : FloatProperty(
        name = "Step Length", default = 0.3,
        min = 0.01, precision = 2)

    aspect_truss : FloatProperty(
        name = 'Aspect Ratio', default = 1,
        min = 0, step = 0.1, precision = 3)

    offset_truss : FloatProperty(
        name = 'Offset Factor', default = 1,
        min = 0, precision = 2)

    joint_size_truss : FloatProperty(
        name = 'Joint Size', default = 0.1,
        min = 0.01, max = 0.5, precision = 2)
    
    fence_bevel_type : EnumProperty(
        name = "Type", default = "CYLINDER",
        items = BevelType[0:3])

    object_aspect : FloatProperty(
        name = "Object Aspect", default = 1,
        min = 0.1, max = 5)

    resolution_u : IntProperty(
        name = 'Resolution U', default = 4,
        min = 2, max = 8)
    
    h_offset : FloatProperty(
        name = 'Offset H', default = 0,
        step = 1, precision = 2)

    v_offset : FloatProperty(
        name = 'Offset V', default = 0,
        step = 1, precision = 2)

    twist_radius_barb : FloatProperty(
        name = "Twist Radius", default = 1,
        min = 0.5, max = 2, step = 1, precision = 2)

    twist_width_barb : FloatProperty(
        name = "Twist Width", default = 6,
        step = 1, precision = 2)
    
    barb_interval : IntProperty(
        name = "Barb Interval", default = 22,
        min = 5, max = 100)

    barb_distort : FloatProperty(
        name = "Distortion", default = 1,
        min = 0, max = 10, step = 10, precision = 1)

    barb_distort_freq : FloatProperty(
        name = "Frequency", default = 0.3,
        min = 0.03, max = 0.5, step = 0.1, precision = 2)

    random_seed : IntProperty(
        name = "Random Seed", default = 0)

    bevel_depth_guardrail : FloatProperty(
        name = "Bevel Depth", default = 0.04,
        min = 0.001, step = 0.1, precision = 3)

    guardrail_num : IntProperty(
        name = "Number", default = 3,
        min = 1, max = 6)

    guardrail_resolution_u : IntProperty(
        name = "Resolution U", default = 8,
        min = 4, max = 16)

    guardrail_offset : FloatProperty(
        name = "Offset", default = 0.1,
        min = 0, max = 1, step = 0.3)

    guardrail_margin : FloatProperty(
        name = "Margin", default = 0.2, 
        min = 0, max = 0.5, step = 0.3)

    guardrail_slide : FloatProperty(
        name = "Slide", default = 0, 
        min = -0.5, max = 0.5, step = 0.3 )

    guardrail_connector : BoolProperty(
        name = "Connector", default = True)

    guardrail_connector_size : FloatProperty(
        name = "Size", default = 2,
        min = 1, max = 3, step = 0.3)

    guardrail_connector_radius_A : FloatProperty(
        name = "Radius A", default = 1.3,
        min = 1, max = 2, step = 0.3)

    #guardrail_connector_opposite : FloatProperty(
    #    name = "Opposite", default = 0, 
    #    min = 0, max = 1, step = 0.3)

    guardrail_connector_radius_B : FloatProperty(
        name = "Radius B",
        min = 0.1,
        max = 1,
        step = 0.1,
        default = 1)

    #guardrail_connector_radius_C : FloatProperty(
    #    name = "Radius C",
    #    min = 0.1,
    #    max = 1,
    #    step = 0.1,
    #    default = 0.3)

    GuardRailType = [("PIPE", "Pipe", "Pipe", "MESH_CYLINDER", 1),
                     ("RAIL", "Rail", "Rail", "PROP_PROJECTED", 2)] 

    guardrail_type : EnumProperty(
        name = "Type",
        default = "PIPE",
        items = GuardRailType)

    guardrail_rail_radius : FloatProperty(
        name ="Radius",
        min = 0,
        step = 0.3,
        default = 0.5)

    GuardRailPoleEndType = [("NONE", "None", "None", "", 0),
                            ("CAP", "Cap", "Cap", "MESH_CYLINDER", 1),
                            ("CURVE_1", "Curve 1", "Curve 1", "SPHERECURVE", 2),
                            ("CURVE_2", "Curve 2", "Curve 2", "SPHERECURVE", 3),
    ]
    
    guardrail_pipe_end : EnumProperty(
        name = "End Cap",
        default = "NONE",
        items = GuardRailPoleEndType)

    guardrail_pipe_captype : EnumProperty(
        name = "Cap Type",
        default = "SIMPLE",
        items = CapType)

    guardrail_endcap : BoolProperty(
        name = "End Cap",
        default = True)

    guardrail_endcap_length : FloatProperty(
        name = "Length",
        default = 0.5,
        min = 0,
        step = 0.3)
    
    guardrail_endcap_size : FloatProperty(
        name = "Size",
        default = 0.2,
        min = 0,
        step = 0.3)

    guardrail_vis : BoolProperty(
        name = "Vis",
        default = False)

    guardrail_vis_size : FloatProperty(
        name = "Size",
        min = 0.75,
        max = 3,
        step = 0.3,
        default = 1)

    DirectionType = [("HORIZONTAL", "Horizontal", "Horizontal", "", 0),
                     ("VERTICAL", "Vertical", "Vertical", "", 1)]
    
    guardrail_vis_direction : EnumProperty(
        name = "Direction",
        default = "HORIZONTAL",
        items = DirectionType)

    guardrail_vis_number : IntProperty(
        name = "Number", default = 1, 
        min = 1, max = 5)

    pole_not_vertical : BoolProperty(
        name = "Not Vertical", default = False)

    carstop_base_type : EnumProperty(
        name = 'Base Type', default = 'NONE',
        items = [("NONE", "None", "None", "", 1),
                ("BLOCK", "Block", "Block", "MESH_CUBE", 2),
                 ("CYLINDER", "Cylinder", "Cylinder", "MESH_CYLINDER", 3),
                 ("CUSTOM", "Custom", "Custom", "MESH_CUBE", 4)])

    pole_radius_ratio_horizontal : FloatProperty(
        name = "Radius", default = 1,
        min = 0.5, max = 2,
        precision = 3)

    pole_offset_horizontal : FloatProperty(
        name = "Offset", default = 0,
        min = -3, max = 3)

    pole_appendix_size : FloatProperty(
        name = "Size", default = 1,
        min = 0, max = 5, step = 0.1)

    bevel_depth_pole : FloatProperty( 
        name = 'Bevel Depth', default = 0.02,
        min = 0, step = 0.1, precision = 4)

    extend_pole_bottom : FloatProperty(
        name = 'Extend Bottom', default = 0,
        step = 0.1, precision = 4)

    pole_appendix_ypos : FloatProperty(
        name = "Position", default = 0.5,
        min = 0, max = 1.2, step = 0.3)

    carstop_type : EnumProperty(
        name = "Type", default = "TYPEA",
        items = CommonTypeSelection[0:1])

    carstop_curve_radius : FloatProperty(
        name = "Curve Radius", default = 0.15,
        min = 0.1, max = 0.4, step = 0.3)

    carstop_offset : FloatProperty(
        name = "Offset", default = 0,
        min = -0.3, max = 0.3, step = 0.3)

    carstop_ring: BoolProperty(
        name = "Ring", default = False)

    carstop_ring_radius_ratio : FloatProperty(
        name = "Radius", default = 0.3,
        min = 0.1, max = 0.8, step = 0.3)

    carstop_ring_type : EnumProperty(
        name = "Type", default = "TYPEA",
        items = CommonTypeSelection[0:2])

    #guardrail_switch : BoolProperty(
    #    name = "Switch",
    #    default = False)

    def draw(self, context):

        layout = self.layout

        layout.prop(self, "action");
        layout.separator()

        if (self.action == 'FACE'):

            layout.prop(self, "fence_type")

            if self.fence_type == 'CROSS':

                layout.prop(self, "fence_bevel_type")
                if (self.fence_bevel_type == 'RECT'):
                    layout.prop(self, "object_aspect")
                layout.prop(self, "v_align_type")
                layout.prop(self, "bevel_depth")
                
                row = layout.row()
                row.prop(self, "interval_horizontal")
                row.prop(self, "interval_vertical")
                row = layout.row()
                row.prop(self, "fit_fence_width")
                row.prop(self, "fit_fence_height")

                layout.prop(self, "h_margin")
                
                box = layout.box().column()

                #row = layout.row()
                box.prop(self, "cross_ornament_type")
                
                ornament_type = self.cross_ornament_type
                if (ornament_type != "NONE"):
                    if (ornament_type == 'CAP'):
                        box.prop(self, "cross_cap_type")
                        row = box.row()
                        row.prop(self, "cap_radius_A")
                        cap_type = self.cross_cap_type
                        if (cap_type == 'TAPER' or cap_type == 'ROUND' or cap_type == 'HAT'):
                            row.prop(self, "cap_radius_B")
                            
                    row = box.row()
                    row.prop(self, "resolution_u")
                    row.prop(self, "ornament_radius")
                    row = box.row()
                    row.prop(self, "cross_ornament_support")
                    if (self.cross_ornament_support == True):
                        row.prop(self, "cross_ornament_support_num")
                        
                row = box.row()
                row.prop(self, "fill_top")
                row.prop(self, "fill_bottom")

                row = layout.row()
                row.prop(self, "cross_horizontal")
                row.prop(self, "cross_vertical")

                row = layout.row()
                row.prop(self, "h_offset")
                row.prop(self, "v_offset")


            if self.fence_type == 'OBLIQUE':
                layout.prop(self, "fence_bevel_type")
                if (self.fence_bevel_type == 'RECT'):
                    layout.prop(self, "object_aspect")
                layout.prop(self, "bevel_depth")
                row = layout.row()
                row.prop(self, "interval_horizontal")
                row.prop(self, "interval_vertical")

                row = layout.row()
                row.prop(self, "cross_horizontal", text="Direction A")
                row.prop(self, "cross_vertical", text="Direction B")


                row = layout.row()
                row.prop(self, "h_offset", text="Offset A")
                row.prop(self, "v_offset", text="Offset B")
           
            if self.fence_type == 'WIRE_MESH':
                layout.prop(self, "curve_type")

                row = layout.row()
                row.prop(self, "bevel_depth_mesh")
                row.prop(self, "filling_mesh")
                
                row = layout.row()
                row.prop(self, "interval_mesh_horizontal")
                row.prop(self, "interval_mesh_vertical")
                row = layout.row()
                row.prop(self, "fit_fence_width")
                row.prop(self, "flip_wiremesh")

                col = self.newboxcol(layout)
                col.prop(self, "fence_distort")
                row = col.row()
                row.prop(self, "fence_distort_offset")
                row.prop(self, "fence_distort_freq")

                row = col.row()
                row.prop(self, "fence_support_wire")
                if (self.fence_support_wire == True):
                    row.prop(self, "fence_support_num")

            if self.fence_type == 'ROPE':
                layout.prop(self, "rope_type")
                row = layout.row()
                if (self.rope_type == "CHAIN"):
                    row.prop(self, "chain_scale")
                    row.prop(self, "chain_elongate")
                                                    
                layout.prop(self, "single_rope")

                col = self.newboxcol(layout)

                if (self.single_rope == True):
                    col.prop(self, "pole_appendix_ypos")
                else:
                    col.prop(self, "v_align_type")

                col.prop(self, "bevel_depth_mesh")
                
                if (self.single_rope == False):
                    col.prop(self, "interval_wire_vertical")
                    col.prop(self, "fit_fence_height")

                col.prop(self, "h_margin")
                col.prop(self, "catenary_a")

                col = self.newboxcol(layout)
                col.prop(self, "rope_terminal_type")
                if (self.rope_terminal_type == "EYE_TERMINAL"):
                    row = col.row()
                    row.prop(self, "rope_terminal_size")
                    row.prop(self, "rope_terminal_rot")
                if (self.rope_terminal_type == "KNOT"):
                    col.prop(self, "rope_knot_radius")
                    
                row = layout.row()
                row.prop(self, "fill_top")
                row.prop(self, "fill_bottom")
                layout.prop(self, "h_offset")

            if self.fence_type == 'BARBED_WIRE':
                col = self.newboxcol(layout)
                col.prop(self, "v_align_type")
                col.prop(self, "bevel_depth_mesh")
                col.prop(self, "interval_wire_vertical")
                col.prop(self, "fit_fence_height")
                row = col.row()
                row.prop(self, "fill_top")
                row.prop(self, "fill_bottom")
                col.prop(self, "h_offset")

                col = self.newboxcol(layout)
                row = col.row()
                row.prop(self, "twist_radius_barb")
                row.prop(self, "twist_width_barb")
                col.prop(self, "barb_interval")

                col = self.newboxcol(layout)
                row = col.row()
                row.prop(self, "barb_distort")
                row.prop(self, "barb_distort_freq")
                col.prop(self, "random_seed")

            if self.fence_type == 'GUARD_PIPE':
                col = self.newboxcol(layout)
                row = col.row()
                row.prop(self, "guardpipe_type")
                row.prop(self, "bevel_depth")
                col.prop(self, "h_offset")
                row = col.row()
                row.prop(self, "guardpipe_connector")
                if (self.guardpipe_connector == True):
                    row.prop(self, "guardpipe_connector_type")
        
        if (self.action == 'TRUSS'):
            layout.prop(self, "bevel_depth_truss")
            row = layout.row()
            row.prop(self, "fit_truss")
            row.prop(self, "flip_truss")
            layout.prop(self, "type_bevel_truss")
            layout.prop(self, "angle_type_truss")
            if (self.angle_type_truss == "0"):
                layout.prop(self, "aspect_truss")
            else:
                layout.prop(self, "step_length_truss")
            layout.prop(self, "offset_truss")
            layout.prop(self, "joint_size_truss")

        if (self.action == 'GUARDRAIL'):
            col = layout.column(align = True)
            
            col.prop(self, "guardrail_type")
            if (self.guardrail_type == "PIPE"):
                col.prop(self, "bevel_depth_guardrail")
                row = col.row()
                row.prop(self, "guardrail_num")
                row.prop(self, "guardrail_resolution_u")
                col.prop(self, "guardrail_margin")
                row = col.row()
                row.prop(self, "guardrail_offset")
                row.prop(self, "guardrail_slide")

                col = self.newboxcol(layout)
                col.prop(self, "guardrail_connector")
                if (self.guardrail_connector == True):
                    row = col.row()
                    row.prop(self, "guardrail_connector_size")
                    row.prop(self, "guardrail_connector_radius_A")
                    #col.prop(self, "guardrail_connector_opposite")
                    row = col.row()
                    row.prop(self, "guardrail_connector_radius_B")
                    #row.prop(self, "guardrail_connector_radius_C")

                    row = col.row()
                    row.prop(self, "guardrail_vis")
                    if (self.guardrail_vis == True):
                        row.prop(self, "guardrail_vis_size")
                        col.prop(self, "guardrail_vis_direction")
                        col.prop(self, "guardrail_vis_number")
                    
                col = self.newboxcol(layout)
                row = col.row()
                row.prop(self, "guardrail_pipe_end")
                if (self.guardrail_pipe_end == "CAP"):
                    col.prop(self, "guardrail_pipe_captype")
                    col.prop(self, "guardrail_endcap_length")
                    col.prop(self, "guardrail_endcap_size")
                    row = col.row()
                    row.prop(self, "cap_radius_A")
                    cap_type = self.guardrail_pipe_captype
                    if (cap_type == 'TAPER' or cap_type == 'ROUND' or cap_type == 'HAT'):
                        row.prop(self, "cap_radius_B")
                if ("CURVE" in self.guardrail_pipe_end):
                    col.prop(self, "guardrail_endcap_length")
                    
            if (self.guardrail_type == 'RAIL'):
                
                col.prop(self, "guardrail_rail_radius")
                row = col.row()
                row.prop(self, "guardrail_offset")
                row.prop(self, "guardrail_slide")
                
                col = self.newboxcol(layout)
                col.prop(self, "guardrail_connector")
                if (self.guardrail_connector == True):
                    row = col.row()
                    row.prop(self, "guardrail_vis")
                    if (self.guardrail_vis == True):
                        row.prop(self, "guardrail_vis_size")

                box = layout.box()
                row = box.row()
                row.prop(self, "guardrail_endcap")
                if (self.guardrail_endcap == True):
                    row.prop(self, "guardrail_endcap_length")

            col.prop(self, "use_adjust_bezier_tips")

        if (self.action == 'CARSTOP'):
            layout.prop(self, "carstop_type")
            row = layout.row()
            row.prop(self, "bevel_depth_pole")
            row.prop(self, "carstop_curve_radius")
            layout.prop(self, "h_margin")
            layout.prop(self, "extend_pole_bottom")
            col = self.newboxcol(layout)
            row = col.row()
            row.prop(self, "pole_not_vertical", text="Support")
            if (self.pole_not_vertical == True):
                row.prop(self, "pole_radius_ratio_horizontal")
                col.prop(self, "carstop_offset")
            col = self.newboxcol(layout)
            row = col.row()
            row.prop(self, "carstop_ring")
            if (self.carstop_ring == True):
                row.prop(self, "carstop_ring_radius_ratio")
                col.prop(self, "carstop_ring_type")
                row = col.row()
                row.prop(self, "pole_appendix_ypos")
                row.prop(self, "pole_appendix_size")

            col = self.newboxcol(layout)
            col.prop(self, "carstop_base_type")
            if (self.carstop_base_type != "NONE"):
                if (self.carstop_base_type == "CUSTOM"):
                    col.prop_search(self, "header_footer_custom_target", bpy.data, "objects")
                row = col.row()
                row.prop(self, "pole_header_footer_size")
                row.prop(self, "pole_base_aspect")
                row = col.row()
                row.prop(self, "pole_footer_taper")
                if (self.pole_footer_taper == True):
                    row.prop(self, "pole_footer_taper_facror")

        self.draw_common(layout)

    def invoke(self, context, event):

        obj = context.object
        fenceprops = obj.fenceprops

        if (self.action == "FACE"):
            if (fenceprops.fence_canvas_type == "GUARD_PIPE"):
                self.fence_type = "GUARD_PIPE"
                self.bevel_depth = 0.02
                self.h_offset = 0.07
                
        if (self.action == "GUARDRAIL"):

            self.guardrail_rail_radius = 0.35#obj.dimensions.z
            self.pole_vertical = True
            self.pole_not_vertical = False

        if (self.action == "CARSTOP"):
            self.bevel_depth_pole = 0.03
            self.carstop_curve_radius = 0.175
            self.pole_radius_ratio_horizontal = 0.55
            self.h_margin = 0.1

        return self.execute(context)



    def add_taper_modifier(self, object, taper_factor):

        mod = object.modifiers.new(name = "SimpleDeform", type="SIMPLE_DEFORM")
        mod.deform_method = "TAPER"
        mod.deform_axis = "Z"
        mod.factor = taper_factor

        return mod

    def apply_curve_modifier(self, object, curve_object, index):

        object_org = bpy.context.view_layer.objects.active
        bpy.context.view_layer.objects.active = curve_object

        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='DESELECT')
        for p in curve_object.data.splines[index].points:
            p.select = True

        bpy.ops.curve.duplicate_move()

        pre_object_list = [o for o in bpy.data.objects]
        
        bpy.ops.curve.separate()
        
        post_object_list = [o for o in bpy.data.objects]
        diff = set(pre_object_list) ^ set(post_object_list)

        new_obj = list(diff)[0]

        bpy.context.view_layer.objects.active = object_org

        #curve_object.data.splines[0], curve_object.data.splines[index] = curve_object.data.splines[index], curve_object.data.splines[0]

        mod = object.modifiers.new(name = "Curve", type="CURVE")
        mod.object = new_obj

        self.apply_modifier(object)

        bpy.data.objects.remove(new_obj)

        #curve_object.data.splines[0], curve_object.data.splines[index] = curve_object.data.splines[index], curve_object.data.splines[0]

    def apply_modifier(self, object):

        object_org = bpy.context.view_layer.objects.active
        bpy.context.view_layer.objects.active = object

        bpy.ops.object.modifier_apply(modifier="Curve", report=False)
        bpy.context.view_layer.objects.active = object_org
        bpy.ops.object.mode_set(mode='EDIT')

    
    def set_tilt_all(self, curveData, tilt, use_smooth = False):
        for s in curveData.splines:
            s.use_smooth = False
            for p in s.points:
                p.tilt += tilt

    

    def get_vlist(self, f):

        list_Z = sorted(f.verts[0:4], key = self.vertZ)

        # 菱形？
        if (math.isclose(list_Z[0].co.x, list_Z[2].co.x, abs_tol=1e-6) and math.isclose(list_Z[0].co.y, list_Z[2].co.y, abs_tol=1e-6)):
            if math.isclose(list_Z[0].co.x, list_Z[1].co.x, abs_tol=1e-6):
                if (list_Z[0].co.y < list_Z[1].co.y):
                    return [list_Z[0], list_Z[1], list_Z[2], list_Z[3]]
                else:
                    return [list_Z[1], list_Z[0], list_Z[3], list_Z[2]]
            else:
                if (list_Z[0].co.x < list_Z[1].co.x):
                    return [list_Z[0], list_Z[1], list_Z[2], list_Z[3]]
                else:
                    return [list_Z[1], list_Z[0], list_Z[3], list_Z[2]]

        if (math.isclose(list_Z[0].co.x, list_Z[1].co.x, abs_tol=1e-6) and math.isclose(list_Z[0].co.y, list_Z[1].co.y, abs_tol=1e-6)):
            if (list_Z[0].co.x == list_Z[2].co.x):
                if (list_Z[0].co.y < list_Z[2].co.y):
                    return [list_Z[0], list_Z[2], list_Z[1], list_Z[3]]
                else:
                    return [list_Z[2], list_Z[0], list_Z[3], list_Z[1]]
            else:
                if (list_Z[0].co.x < list_Z[2].co.x):
                    return [list_Z[0], list_Z[2], list_Z[1], list_Z[3]]
                else:
                    return [list_Z[2], list_Z[0], list_Z[3], list_Z[1]]

    
        if math.isclose(list_Z[0].co.x, list_Z[1].co.x, abs_tol=1e-6):
            sorted_list1 = sorted(list_Z[0:2], key = self.vertY)
            sorted_list2 = sorted(list_Z[2:4], key = self.vertY)
        else:
            sorted_list1 = sorted(list_Z[0:2], key = self.vertX)
            sorted_list2 = sorted(list_Z[2:4], key = self.vertX)
            
        vlist = sorted_list1 + sorted_list2

        return vlist

    def get_co_list(self, vlist):

        use_margin = False
        if (self.action == 'FACE' and (self.fence_type == 'CROSS')):
            use_margin = True
        if (self.action == 'CARSTOP'):
            use_margin = True

        if (use_margin == True):
            
            h0 = (vlist[1].co-vlist[0].co).normalized()
            h1 = (vlist[3].co-vlist[2].co).normalized()
        
            co_list = []
            co_list.append(vlist[0].co + h0 * self.h_margin)
            co_list.append(vlist[1].co - h0 * self.h_margin)
            co_list.append(vlist[2].co + h1 * self.h_margin)
            co_list.append(vlist[3].co - h1 * self.h_margin)

        else:
            co_list = [vlist[0].co, vlist[1].co, vlist[2].co, vlist[3].co]
            
        return co_list

    def do_ops_generate(self, function):
        
        pre_object_list = [o for o in bpy.data.objects]

        function()

        post_object_list = [o for o in bpy.data.objects]
        diff = set(pre_object_list) ^ set(post_object_list)

        new_obj = list(diff)[0]

        return new_obj

    def wire_distort_set_paramB(self, ini, i, f_idx):

        if (i == ini):
            random.seed(self.random_seed + f_idx)
            self.random_state_for_i = random.getstate()
            
        random.setstate(self.random_state_for_i)

        offset0 = 2 * 3.14159 * random.random()
        offset1 = 2 * 3.14159 * random.random()
        offset2 = 2 * 3.14159 * random.random()
        offset3 = 2 * 3.14159 * random.random()

        amp0 = random.random()
        amp1 = random.random() * 0.5
        amp2 = random.random() * 0.25
        amp3 = random.random() * 0.125

        freq = self.barb_distort_freq
        freq0 = 1 * freq * (1 if random.random() > 0.5 else -1)
        freq1 = 2 * freq * (1 if random.random() > 0.5 else -1)
        freq2 = 3 * freq * (1 if random.random() > 0.5 else -1)
        freq3 = 4 * freq * (1 if random.random() > 0.5 else -1)

        self.random_state_for_i = random.getstate()

        return (freq0, freq1, freq2, freq3), (offset0, offset1, offset2, offset3), (amp0, amp1, amp2, amp3)
        
    def get_v_ini_fin(self, num):
        v_ini = 1
        v_fin = num
        if (self.fill_top == True):
            if int(self.v_align_type) == 0:
                v_ini = 0
            else:
                v_fin = num + 1
                
        if (self.fill_bottom == True):
            if int(self.v_align_type) == 0:
                v_fin = num + 1
            else:
                v_ini = 0

        return v_ini, v_fin

    

    def make_rect_object(self, depth, aspect):
        newCurveData = bpy.data.curves.new('rect', type='CURVE')
        newCurveData.dimensions = '3D'
        newCurveData.resolution_u = 1

        polyline = newCurveData.splines.new('POLY')
        polyline.use_cyclic_u = True
        polyline.use_smooth = False
        points = polyline.points
        points.add(3)

        co_list = [Vector((-aspect, -1, 0, 1)) * depth,
                   Vector((+aspect, -1, 0, 1)) * depth,
                   Vector((+aspect, +1, 0, 1)) * depth,
                   Vector((-aspect, +1, 0, 1)) * depth]
        self.set_point_from_list3d(points, co_list)

        return bpy.data.objects.new(newCurveData.name, newCurveData)

    

    def proc_bevel_type(self, curveData, bevel_type, sc, action = ''):

        if (action == ''):
            action = self.action

        def set_use_smooth(curveData):
            for s in curveData.splines:
                s.use_smooth = self.smooth

        if (action == 'FACE'):

            if (self.fence_type == 'WIRE_MESH'):
                set_use_smooth(curveData)

            if (self.fence_type == 'ROPE'):
                set_use_smooth(curveData)

            if (self.fence_type == 'BARBED_WIRE'):
                set_use_smooth(curveData)

            if (bevel_type == 'CYLINDER'):
                set_use_smooth(curveData)

            if (bevel_type == 'PRISM'):
                curveData.bevel_resolution = 0
                self.set_tilt_all(curveData, 3.14159/4)

            if (bevel_type == 'RECT'):
                for s in curveData.splines:
                    s.use_smooth = False
                rectOB = self.make_rect_object(self.bevel_depth, self.object_aspect);
                sc.collection.objects.link(rectOB)
                curveData.bevel_mode = 'OBJECT'
                curveData.bevel_object = rectOB
                return rectOB

        if (self.action == 'TRUSS'):
            if (bevel_type == 'CYLINDER'):
                for s in curveData.splines:
                    s.use_smooth = self.smooth
                    
            if (bevel_type == 'PRISM'):
                curveData.bevel_resolution = 0
                for s in curveData.splines:
                    s.use_smooth = False
                    for p in s.points:
                        p.tilt += 3.14159/4

        if (self.action == 'GUARDRAIL'):
            for s in curveData.splines:
                s.use_smooth = self.smooth


    def set_bevel_depth(self, curveData):
        if (self.action == 'FACE'):
            if self.fence_type == 'CROSS':
                curveData.bevel_depth = self.bevel_depth
            if self.fence_type == 'OBLIQUE':
                curveData.bevel_depth = self.bevel_depth
            if self.fence_type == 'WIRE_MESH':
                curveData.bevel_depth = self.bevel_depth_mesh * self.filling_mesh * 2
            if self.fence_type == 'BARBED_WIRE':
                curveData.bevel_depth = self.bevel_depth_mesh
            if self.fence_type == 'GUARD_PIPE':
                curveData.bevel_depth = self.bevel_depth
                
        if (self.action == 'TRUSS'):
            curveData.bevel_depth = self.bevel_depth_truss
        if (self.action == 'GUARDRAIL'):
            curveData.bevel_depth = self.bevel_depth_guardrail
        if (self.action == 'CARSTOP'):
            curveData.bevel_depth = self.bevel_depth_pole
        
    def get_num_h(self, vlist, interval, fit, fit_offset = 0):
        h_dist = (vlist[1].co-vlist[0].co).length
        num_hf = h_dist / interval
        num_h = math.ceil(num_hf)
        if (fit == True):
            num_hf = num_h
            num_h += fit_offset
        return h_dist, num_hf, num_h

    def get_num_h_co(self, co_list, interval, fit, fit_offset = 0):
        h_dist = (co_list[1] - co_list[0]).length
        num_hf = h_dist / interval
        num_h = math.ceil(num_hf)
        if (fit == True):
            num_hf = num_h
            num_h += fit_offset
        return h_dist, num_hf, num_h

    def get_num_v(self, vlist, interval, fit, fit_offset = 0):
        v_dist = (vlist[2].co-vlist[0].co).length
        num_vf = v_dist / interval
        num_v = math.ceil(num_vf)
        if (fit == True):
            num_vf = num_v
            num_v += fit_offset
        return v_dist, num_vf, num_v

    def get_num_v_co(self, co_list, interval, fit, fit_offset = 0):
        v_dist = (co_list[2] - co_list[0]).length
        num_vf = v_dist / interval
        num_v = math.ceil(num_vf)
        if (fit == True):
            num_vf = num_v
            num_v += fit_offset
        return v_dist, num_vf, num_v

    

    def make_ring(self, curveData, center, radius, x_unit, y_unit, normal):

        bezline = curveData.splines.new('BEZIER')
        points = bezline.bezier_points
        points.add(3)

        points[0].co = center - y_unit*radius
        points[0].handle_left = center + (-y_unit-0.5521*x_unit)*radius
        points[0].handle_right = center + (-y_unit+0.5521*x_unit)*radius

        points[1].co = center + x_unit*radius
        points[1].handle_left = center + (x_unit-0.5521*y_unit)*radius
        points[1].handle_right = center + (x_unit+0.5521*y_unit)*radius

        points[2].co = center + y_unit*radius
        points[2].handle_left = center + (y_unit+0.5521*x_unit)*radius
        points[2].handle_right = center + (y_unit-0.5521*x_unit)*radius

        points[3].co = center - x_unit*radius
        points[3].handle_left = center + (-x_unit+0.5521*y_unit)*radius
        points[3].handle_right = center + (-x_unit-0.5521*y_unit)*radius

        bezline.use_cyclic_u = True

        unit_z = Vector((0,0,1))
        if (self.fence_bevel_type == 'PRISM'):
            self.adjust_bezier_tilt(bezline, unit_z, normal, offset=3.14159/4)
        else:
            self.adjust_bezier_tilt(bezline, unit_z, normal)

    def append_round_top(self, polyline, v_unit, pos, tilt, size, radius = 1.0):
        co_0 = pos.copy()
        co_0.resize_3d()
        t = 3.14159/2

        num = 10
        if (size < 1):
            num = 5

        points = polyline.points
        co = co_0 + v_unit * sin(t*0.02) * size * 0.2
        self.append_poly_from3d(points, co, tilt, cos(t*0.00)*radius)

        for i in range(1, num):
            f = i / num
            co = co_0 + v_unit * sin(t*f) * size * 0.2
            self.append_poly_from3d(points, co, tilt, cos(t*f)*radius)

        co = co_0 + v_unit * sin(t*0.95) * size * 0.2
        self.append_poly_from3d(points, co, tilt, cos(t*0.95)*radius)



    def get_support_base(self, tgt, co_list, unit, v_offset):
        if tgt == "TOP":
            co_0 = co_list[2].copy()
            co_1 = co_list[3].copy()
        if tgt == "BOTTOM":
            co_0 = co_list[0].copy()
            co_1 = co_list[1].copy()

        co_0 += unit*v_offset
        co_1 += unit*v_offset

        return co_0, co_1

    def make_ornament_ring_support(self, tgt, co_list, curveData, x_unit, y_unit, v_offset, num, num_max):

        co_0, co_1 = self.get_support_base(tgt, co_list, y_unit, v_offset)

        radius = self.ornament_radius - self.bevel_depth * 2
        for i in range(0,num):
            phi = (i+0.5) * 2*3.14159/num_max
            co_a = co_0 + math.cos(phi)*radius*x_unit + math.sin(phi)*radius*y_unit
            co_b = co_1 + math.cos(phi)*radius*x_unit + math.sin(phi)*radius*y_unit
            self.set_poly_from3d(curveData, co_a, co_b, 0)

    def make_ornament_turnaronud_support(self, tgt, co_list, curveData, x_unit, y_unit, v_offset):

        co_0, co_1 = self.get_support_base(tgt, co_list, x_unit, v_offset)
        if tgt == 'TOP':
            co_0 += y_unit * (0.075*self.ornament_radius-self.bevel_depth*2)
            co_1 += y_unit * (0.075*self.ornament_radius-self.bevel_depth*2)
        if tgt == 'BOTTOM':
            co_0 -= y_unit * (0.075*self.ornament_radius-self.bevel_depth*2)
            co_1 -= y_unit * (0.075*self.ornament_radius-self.bevel_depth*2)

        co_0 -= 0.125 * self.ornament_radius * x_unit
        co_1 -= 0.125 * self.ornament_radius * x_unit
        
        self.set_poly_from3d(curveData, co_0, co_1, 0)

    def make_ornament_quad_support(self, tgt, co_list, curveData, x_unit, y_unit, v_offset):

        co_0, co_1 = self.get_support_base(tgt, co_list, x_unit, v_offset)

        depth = self.bevel_depth
        radius = self.ornament_radius
        if tgt == "TOP":
            co_0 -= y_unit * radius 
            co_1 -= y_unit * radius 
            co_A0 = co_0 + y_unit*radius + x_unit*(radius-depth*2)
            co_A1 = co_1 + y_unit*radius + x_unit*(radius-depth*2)
            co_B0 = co_0 + y_unit* (2*radius-depth*2)
            co_B1 = co_1 + y_unit* (2*radius-depth*2)

        if tgt == "BOTTOM":
            co_0 += y_unit * radius 
            co_1 += y_unit * radius 
            co_A0 = co_0 - y_unit*radius + x_unit*(radius-depth*2)
            co_A1 = co_1 - y_unit*radius + x_unit*(radius-depth*2)
            co_B0 = co_0 - y_unit* (2*radius-depth*2)
            co_B1 = co_1 - y_unit* (2*radius-depth*2)
            
        self.set_poly_from3d(curveData, co_A0, co_A1, 0)
        self.set_poly_from3d(curveData, co_B0, co_B1, 0)

    def make_ornament_triangle_support(self, tgt, co_list, curveData, x_unit, y_unit, v_offset):

        co_0, co_1 = self.get_support_base(tgt, co_list, x_unit, v_offset)

        depth = self.bevel_depth
        radius = self.ornament_radius
        if tgt == "TOP":
            co_0 -= y_unit * radius 
            co_1 -= y_unit * radius 
            co_A0 = co_0 + y_unit*(1.732*radius-depth*2)+x_unit*(0.95*radius-depth*2)
            co_A1 = co_1 + y_unit*(1.732*radius-depth*2)+x_unit*(0.95*radius-depth*2)
            
        if tgt == "BOTTOM":
            co_0 += y_unit * radius 
            co_1 += y_unit * radius 
            co_A0 = co_0 - y_unit*(1.732*radius-depth*2)+x_unit*(0.95*radius-depth*2)
            co_A1 = co_1 - y_unit*(1.732*radius-depth*2)+x_unit*(0.95*radius-depth*2)
            
        self.set_poly_from3d(curveData, co_A0, co_A1, 0)

    def make_ornament_triangular_support(self, tgt, co_list, curveData, x_unit, y_unit, v_offset):

        co_0_base, co_1_base = self.get_support_base(tgt, co_list, x_unit, v_offset)

        radius = self.ornament_radius
        if tgt == "TOP":
            co_0_base -= y_unit * radius 
            co_1_base -= y_unit * radius 
            shapeA = Vector((0,+1.80))
            shapeB = Vector((0.432,+0.570))

        if tgt == "BOTTOM":
            co_0_base += y_unit * radius 
            co_1_base += y_unit * radius 
            shapeA = Vector((0,-1.80))
            shapeB = Vector((0.432,-0.570))

        co_0_base -= x_unit * self.bevel_depth*2
        co_1_base -= x_unit * self.bevel_depth*2

        shapeA += (shapeB-shapeA).normalized() * self.bevel_depth/radius
        
        ini = 0
        fin = self.cross_ornament_support_num
        num = max(1, fin - 1)
        
        for i in range(0,fin):
            co_offset = (i/num)*shapeB + (1-i/num)*shapeA
            co_0 = co_0_base + (co_offset[0]*x_unit + co_offset[1]*y_unit)*radius
            co_1 = co_1_base + (co_offset[0]*x_unit + co_offset[1]*y_unit)*radius
            self.set_poly_from3d(curveData, co_0, co_1, 0)

        
    def make_cross_ornament(self, tgt, co_list, curveData, normal, num_h, num_hf, v_offset):
        if (self.cross_ornament_type == "NONE"):
            if (tgt == "TOP"):
                co_0 = co_list[2].copy()
                co_1 = co_list[3].copy()
            if (tgt == "BOTTOM"):
                co_0 = co_list[0].copy()
                co_1 = co_list[1].copy()
            co_0 += normal*v_offset
            co_1 += normal*v_offset
                    
            unit_z = Vector((0,0,1))
            tilt = self.get_tilt((co_1-co_0).normalized(), unit_z, normal)
            self.set_poly_from3d(curveData, co_0, co_1, tilt)

        v_unit = (co_list[2]-co_list[0]).normalized()

        if (self.cross_ornament_type == "RING"):
            ini, fin = self.get_ini_fin_margin(num_h)
            if tgt == "TOP":
                for i in range(ini, fin):
                    co_c = (i*co_list[2] + (num_hf-i)*co_list[3])/num_hf
                    if (i == num_h):
                        co_c = co_list[2].copy()
                    co_c += normal*v_offset
                    self.make_ring(curveData, co_c, self.ornament_radius, v_unit, normal, normal)

            if tgt == "BOTTOM":
                for i in range(ini, fin):
                    co_c = (i*co_list[0] + (num_hf-i)*co_list[1])/num_hf
                    if (i == num_h):
                        co_c = co_list[0].copy()
                    co_c += normal*v_offset
                    self.make_ring(curveData, co_c, self.ornament_radius, v_unit, normal, normal)

            support_num = self.cross_ornament_support_num
            if (self.cross_ornament_support == True):
                self.make_ornament_ring_support(tgt, co_list, curveData, v_unit, normal, v_offset, support_num, support_num)


        if (self.cross_ornament_type == "CAP"):
            ini, fin = self.get_ini_fin_margin(num_h)
            if tgt == "TOP":
                for i in range(ini, fin):
                    co_c = (i*co_list[2] + (num_hf-i)*co_list[3])/num_hf
                    if (i == num_h):
                        co_c = co_list[2].copy()
                    co_c += normal*v_offset
                    self.make_cap(curveData, co_c, self.ornament_radius, self.bevel_depth, normal, v_unit, normal, "TOP", self.cross_cap_type)

            if tgt == "BOTTOM":
                for i in range(ini, fin):
                    co_c = (i*co_list[0] + (num_hf-i)*co_list[1])/num_hf
                    if (i == num_h):
                        co_c = co_list[0].copy()
                    co_c += normal*v_offset
                    self.make_cap(curveData, co_c, self.ornament_radius, self.bevel_depth, normal, v_unit, normal, "BOTTOM", self.cross_cap_type)


        if (self.cross_ornament_type == 'HALF_RING'):
            v_unit = (co_list[2]-co_list[0]).normalized()
            support_num = self.cross_ornament_support_num
            if (self.cross_ornament_support == True):
                self.make_ornament_ring_support(tgt, co_list, curveData, v_unit, normal, v_offset, int(support_num/2), support_num)

        if (self.cross_ornament_type == 'TURN_AROUND'):
            if (self.cross_ornament_support == True):
                self.make_ornament_turnaronud_support(tgt, co_list, curveData, normal, v_unit, v_offset)

        if (self.cross_ornament_type == 'QUAD'):
            if (self.cross_ornament_support == True):
                self.make_ornament_quad_support(tgt, co_list, curveData, normal, v_unit, v_offset)

        if (self.cross_ornament_type == 'TRIANGLE'):
            if (self.cross_ornament_support == True):
                self.make_ornament_triangle_support(tgt, co_list, curveData, normal, v_unit, v_offset)

        if (self.cross_ornament_type == 'TRIANGULAR'):
            if (self.cross_ornament_support == True):
                self.make_ornament_triangular_support(tgt, co_list, curveData, normal, v_unit, v_offset)
                
                    
    def adjust_by_ornament(self, co_0, co_1, v_unit):
        type = self.cross_ornament_type
        if (type == 'RING') or (type == 'HALF_RING') or (type == 'TRIANGULAR') or (type == 'TRIANGLE') or (type == 'QUAD'):
            if (self.fill_top == True):
                co_1 -= v_unit * self.ornament_radius
                
            if (self.fill_bottom == True):
                co_0 += v_unit * self.ornament_radius
                
                    
    def adjust_bezier_tilt(self, bezline, unit_z, normal, offset = 0):
        p0 = bezline.bezier_points[0]
        if (bezline.use_cyclic_u == False):
            p1 = bezline.bezier_points[1]
            tilt = self.get_tilt((p1.co - p0.co).normalized(), unit_z, normal)
        else:
            p1 = bezline.bezier_points[-1]
            tilt = self.get_tilt((p0.co - p1.co).normalized(), unit_z, normal)
            
        for p in bezline.bezier_points:
            p.tilt = tilt + offset

            
    def make_cross_vertical_line(self, co_list, curveData, normal, co_0, co_1, x_unit, y_unit):

        if (self.cross_vertical == False):
            return

        def set_bezpoints(p, x, y, base, shape, top=False):
            if (top == False):
                p.co           = base + (shape[0][0]*x+shape[0][1]*y)*radius
                p.handle_left  = base + (shape[1][0]*x+shape[1][1]*y)*radius
                p.handle_right = base + (shape[2][0]*x+shape[2][1]*y)*radius
            if (top == True):
                p.co           = base + (shape[0][0]*x-shape[0][1]*y)*radius
                p.handle_left  = base + (shape[2][0]*x-shape[2][1]*y)*radius
                p.handle_right = base + (shape[1][0]*x-shape[1][1]*y)*radius
                    
        shape = []
        if (self.cross_ornament_type == 'HALF_RING'):

            shape = [((0, -2), (-0.5521,-2), (+0.5521,-2)),
                     ((1, -1), (1, -1-0.5521), (1, -1+0.5521)),
                     ((0,  0), (+0.5521,0), (0, 1)),
                     ((0,  0), (0, 1), (+0.5521,0))]
            
        if (self.cross_ornament_type == 'TURN_AROUND'):

            shape = [((-0.500, 1.0), (-0.525, +1.1), (-0.475, +0.9)),
                     ((-0.250, 0.0), (-0.275, +0.1), (-0.225, -0.1)),
                     ((0     ,   0), (+0,000, -0.1), (+0.000, +0.1)),
                     ((0     ,   0), (+0.000, -0.1), (+0.000, +0.1))]

        if (self.cross_ornament_type == 'TRIANGLE'):
            shape = [((-1   , -1.732), (-1.1 , -1.732), (-0.9 , -1.732)),
                     ((0.8  , -1.732), (0.674, -1.732), (0.908, -1.732)),
                     ((0.946, -1.455), (1.038, -1.584), (0.853, -1.326)),
                     ((0     ,   0.0) , (0.138, -0.228), (0.0, 0.1)),
                     ((0     ,   0.0) , (0.0   , 0.1)   , (0.138, +0.228))]

        if (self.cross_ornament_type == 'QUAD'):
            shape = [((-0.950, -1.1), (-1.023, -1.021), (-0.876, -1.179)),
                     ((-0.074, -1.95), (-0.141, -1.930), (-0.028, -1.963)),
                     ((0.0734, -1.95), (0.0309, -1.966), (0.1373, -1.925)),
                     ((0.950, -1.086), (0.9269, -1.129), (0.9728, -1.044)),
                     ((0.950, -0.913), (0.9728, -0.955), (0.9157, -0.850)),
                     ((0.0   , 0.0)  , (0.3166, -0.316), (0.0, 0.1)),
                     ((0.0   , 0.0)  , (0.0, 0.1), (0.3166, +0.316))]


        if (self.cross_ornament_type == 'TRIANGULAR'):
            shape = [((-0.400, -1.0), (-0.400, -0.800), (-0.400, -1.304)),
                     ((-0.400, -1.8), (-0.400, -1.662), (-0.400, -1.868)),
                     ((-0.200, -1.954), (-0.312, -1.954), (-0.086, -1.954)),
                     ((0.0, -1.80), (-0.024, -1.848), (0.04, -1.716)),
                     ((0.432, -0.570), (0.332, -0.868), (0.454, -0.504)),
                     ((0.438, -0.416), (0.488, -0.468), (0.370, -0.348)),
                     ((0.066, -0.114), (0.098, -0.138), (0.032, -0.090)),
                     ((0    ,      0), (0, -0.06), (0, 0.06)),
                     ((0    ,      0), (0, -0.06), (0, 0.06))]

        if len(shape) == 0:
            return False

        point_num = (len(shape)-2)
        
        radius = self.ornament_radius
        bezline = curveData.splines.new('BEZIER')
        points = bezline.bezier_points
        points.add(1 + point_num * self.fill_top + point_num * self.fill_bottom)
        idx = 0
        if (self.fill_bottom == True):
            for i in range(0, point_num):
                set_bezpoints(points[idx+i],x_unit,y_unit,co_0,
                              shape[i])
            idx += point_num

        set_bezpoints(points[idx+0],x_unit,y_unit,co_0,shape[point_num+0])
        set_bezpoints(points[idx+1],x_unit,y_unit,co_1,shape[point_num+1])
        idx+= 2

        if (self.fill_top == True):
            for i in range(0, point_num):
                set_bezpoints(points[idx+i],x_unit,y_unit,co_1,
                              shape[point_num-i-1], top=True)

        unit_z = Vector((0,0,1))
        if (self.fence_bevel_type == "PRISM"):
            self.adjust_bezier_tilt(bezline, unit_z, normal, offset=3.14159/4)
        else:
            self.adjust_bezier_tilt(bezline, unit_z, normal)
            
        return True

        #if (self.cross_ornament_type == 'TURN_AROUND'):
        #
        #    shape = [((-0.500, 1.0), (-0.525, +1.1), (-0.475, +0.9)),
        #             ((-0.250, 0.0), (-0.275, +0.1), (-0.225, -0.1)),
        #             ((0     ,   0), (+0,000, -0.1), (+0.000, +0.1)),
        #             ((0     ,   0), (+0.000, -0.1), (+0.000, +0.1)),
        #             ((-0.250, 0,0), (-0.225, +0.1), (-0.275, -0.1)),
        #             ((-0.500,-1.0), (-0.475, -0.9), (-0.525, -1.1))]

        #    radius = self.ornament_radius
        #    bezline = curveData.splines.new('BEZIER')
        #    points = bezline.bezier_points
        #    points.add(1 + 2 * self.fill_top + 2 * self.fill_bottom)
        #    idx = 0

        #    if (self.fill_bottom == True):
        #        set_bezpoints(points[idx+0],x_unit,y_unit,co_0,shape[0])
        #        set_bezpoints(points[idx+1],x_unit,y_unit,co_0,shape[1])
        #        idx+= 2

        #    set_bezpoints(points[idx+0],x_unit,y_unit,co_0,shape[2])
        #    set_bezpoints(points[idx+1],x_unit,y_unit,co_1,shape[3])
        #    idx+= 2

        #    if (self.fill_top == True):
        #        set_bezpoints(points[idx+0],x_unit,y_unit,co_1,shape[4])
        #        set_bezpoints(points[idx+1],x_unit,y_unit,co_1,shape[5])
        #        idx+= 2
        
        #    unit_z = Vector((0,0,1))
        #    if (self.fence_bevel_type == "PRISM"):
        #        self.adjust_bezier_tilt(bezline, unit_z, normal, offset=3.14159/4)
        #    else:
        #        self.adjust_bezier_tilt(bezline, unit_z, normal)

        #return True
        
        # 単なる直線を作る時
        return False

    def get_ini_fin_margin(self, num_h):
        ini = 1; fin = num_h
        if (self.h_margin != 0):
            ini = 0; fin = num_h+1
        return ini, fin
            
    def make_fence_cross_1unit(self, co_list, curveData, normal):

        depth = self.bevel_depth
        unit_z = Vector((0,0,1))

        interval_h = self.interval_horizontal
        interval_v = self.interval_vertical

        h_dist, num_hf, num_h = self.get_num_h_co(co_list, interval_h, self.fit_fence_width)

        v_unit = (co_list[2] - co_list[0]).normalized()
        v_offset = self.v_offset * depth

        ini, fin = self.get_ini_fin_margin(num_h)
        for i in range(ini, fin):
            co_0 = (i*co_list[0] + (num_hf-i)*co_list[1])/num_hf
            co_1 = (i*co_list[2] + (num_hf-i)*co_list[3])/num_hf

            if (i == num_h):
                co_0 = co_list[0].copy()
                co_1 = co_list[2].copy()

            self.adjust_by_ornament(co_0, co_1, v_unit)

            co_0 += normal*v_offset
            co_1 += normal*v_offset

            if (self.make_cross_vertical_line(co_list, curveData, normal, co_0, co_1, normal, v_unit) == False):

                tilt = self.get_tilt((co_1-co_0).normalized(), unit_z, normal)
                self.set_poly_from3d(curveData, co_0, co_1, tilt)
                
        v_dist, num_vf, num_v = self.get_num_v_co(co_list, interval_v, self.fit_fence_height)

        if (self.cross_horizontal == False):
            return

        for i in range(1,num_v):
            if int(self.v_align_type) == 0:
                co_0 = (i*co_list[0] + (num_vf-i)*co_list[2])/num_vf
                co_1 = (i*co_list[1] + (num_vf-i)*co_list[3])/num_vf
            else:
                co_0 = (i*co_list[2] + (num_vf-i)*co_list[0])/num_vf
                co_1 = (i*co_list[3] + (num_vf-i)*co_list[1])/num_vf

            co_0 += normal*self.h_offset*depth
            co_1 += normal*self.h_offset*depth

            tilt = self.get_tilt((co_1-co_0).normalized(), unit_z, normal)
            self.set_poly_from3d(curveData, co_0, co_1, tilt)

        if (self.fill_top == True):
            self.make_cross_ornament("TOP", co_list, curveData, normal, num_h, num_hf, v_offset)
            
        if (self.fill_bottom == True):
            self.make_cross_ornament("BOTTOM", co_list, curveData, normal, num_h, num_hf, v_offset)


    def make_fence_oblique_1unit(self, co_list, curveData, normal):

        interval_h = self.interval_horizontal
        interval_v = self.interval_vertical

        h_dist, num_hf, num_h = self.get_num_h_co(co_list, interval_h, True)
        v_dist, num_vf, num_v = self.get_num_v_co(co_list, interval_v, True)

        depth = self.bevel_depth
        unit_z = Vector((0,0,1))

        for i in range(1, num_h + num_v):
            if (self.cross_horizontal == False):
                continue
            
            i_h_0 = min(i, num_h)
            i_v_0 = max(0, i-num_h)
            if (i_h_0 < num_h): 
                co_0 = (i_h_0*co_list[1] + (num_hf-i_h_0)*co_list[0])/num_hf
            else:
                co_0 = (i_v_0*co_list[3] + (num_vf-i_v_0)*co_list[1])/num_vf
                    
            i_h_1 = max(0, i-num_v)
            i_v_1 = min(i, num_v)
            if (i_v_1 == num_v):
                co_1 = (i_h_1*co_list[3] + (num_hf-i_h_1)*co_list[2])/num_hf
            else:
                co_1 = (i_v_1*co_list[2] + (num_vf-i_v_1)*co_list[0])/num_vf

            # unit != unit_z なはずなので、例外処理無しでいい
            co_0 += normal*self.h_offset*depth
            co_1 += normal*self.h_offset*depth
            tilt = self.get_tilt((co_1-co_0).normalized(), unit_z, normal)
            self.set_poly_from3d(curveData, co_0, co_1, tilt)

        for i in range(1, num_h + num_v):
            if (self.cross_vertical == False):
                continue
            
            i_h_0 = min(i, num_h)
            i_v_0 = max(0, i-num_h)
            if (i_h_0 < num_h): 
                co_0 = (i_h_0*co_list[3] + (num_hf-i_h_0)*co_list[2])/num_hf
            else:
                co_0 = (i_v_0*co_list[1] + (num_vf-i_v_0)*co_list[3])/num_vf
                    
            i_h_1 = max(0, i-num_v)
            i_v_1 = min(i, num_v)
            if (i_v_1 == num_v):
                co_1 = (i_h_1*co_list[1] + (num_hf-i_h_1)*co_list[0])/num_hf
            else:
                co_1 = (i_v_1*co_list[0] + (num_vf-i_v_1)*co_list[2])/num_vf

            co_0 += normal*self.v_offset*depth
            co_1 += normal*self.v_offset*depth
            tilt = self.get_tilt((co_1-co_0).normalized(), unit_z, normal)
            self.set_poly_from3d(curveData, co_0, co_1, tilt)

    def make_fence_wire_1unit(self, co_list, curveData, normal):

        if (self.flip_wiremesh == True):
            co_list[0], co_list[1], co_list[2], co_list[3] = co_list[0], co_list[2], co_list[1], co_list[3] 

        interval_h = self.interval_mesh_horizontal
        interval_v = self.interval_mesh_vertical

        h_dist, num_hf, num_h = self.get_num_h_co(co_list, interval_h, self.fit_fence_width, 1)
        
        h_unit = (co_list[1] - co_list[0]).normalized()

        # Fit関係は は独自計算
        v_dist, num_vf, num_v = self.get_num_v_co(co_list, interval_v, False)

        if (num_h == 0) or (num_v == 0):
            return

        # 縦方向は必ずFit扱いにする
        if (True):
            interval_v *= num_vf / num_v
            num_v += 1
            
        #v_unit = (co_list[2] - co_list[0]).normalized()
        #v_step = interval_v * v_unit

        dist = self.fence_distort * 0.1
        depth = self.bevel_depth_mesh
        normal_depth = normal * depth

        dist_normal_offset = - normal * self.fence_distort_offset 
        wave_length = 0.3 / self.fence_distort_freq 
        h_unit3 = h_unit * 3
        interval_h2 = interval_h*0.5
        depth_2 = depth*0.5

        def set_point_from3d(point, co):
            co_tmp = co.copy();
            co_tmp.resize_4d(); point.co = co_tmp

        for i in range(1,num_h):

            dist_factor_h = math.sin(i/(num_h) * 3.14159)
            
            if (i % 2 == 1):
                offset_count = -1
            else:
                offset_count = +1
                
            i_idx = i - i % 2
            co_0 = (i_idx*co_list[0] + (num_hf-i_idx)*co_list[1])/num_hf
            co_1 = (i_idx*co_list[2] + (num_hf-i_idx)*co_list[3])/num_hf

            co_0 -= normal * depth * 1 * offset_count
            co_0 -= offset_count * h_unit * 1.5 * depth
            co_1 -= normal * depth * 1 * offset_count
            co_1 -= offset_count * h_unit * 1.5 * depth

            polyline = curveData.splines.new('POLY')
            polyline.points.add((num_v-1)*3)
            points = polyline.points
            set_point_from3d(points[0], co_0)
            
            co_curr = co_0.copy()

            v_unit = (co_1 - co_0).normalized()
            v_step = (co_1 - co_0)/(num_v-1)
            v_step04 = v_step*0.4

            for j in range(1, num_v):
                normal_offset = normal_depth * offset_count

                co_curr += interval_h2 * (offset_count * h_unit) + v_step04
                co_curr += offset_count * h_unit3 * depth_2 - normal_offset
                co_curr0 = co_curr.copy()
                
                co_curr += interval_h2 * (offset_count * h_unit) + v_step04
                co_curr += offset_count * h_unit3 * depth_2 + normal_offset
                co_curr1 = co_curr.copy()
                
                co_curr += interval_v * 0.2 * v_unit + normal_offset * 2
                co_curr2 = co_curr.copy()

                if (self.fence_distort != 0):
                    dist_factor = dist_factor_h * sin(j/(num_v-1)*3.14159) * dist
                    co_curr0 += (noise_vector(co_curr0 / wave_length) + dist_normal_offset) * dist_factor
                    co_curr1 += (noise_vector(co_curr1 / wave_length) + dist_normal_offset) * dist_factor
                    co_curr2 += (noise_vector(co_curr2 / wave_length) + dist_normal_offset) * dist_factor
                    
                set_point_from3d(points[(j-1)*3+1], co_curr0)
                set_point_from3d(points[(j-1)*3+2], co_curr1)
                set_point_from3d(points[(j-1)*3+3], co_curr2)

                offset_count *= -1

        if (self.fence_support_wire == True):

            for i in range(1, self.fence_support_num+1):

                if (i == 1 and self.fence_support_num == 1):
                    dist_factor_h = 1
                else:
                    dist_factor_h = math.sin(i/(self.fence_support_num+1) * 3.14159)
                print("dist_factor_h = ", dist_factor_h, i, self.fence_support_num)
                
                i_idx = int(i * num_v / (self.fence_support_num+1))

                v0_step = (co_list[2] - co_list[0]) / num_v
                v1_step = (co_list[3] - co_list[1]) / num_v

                co_0 = (i_idx*co_list[0] + (num_v-1-i_idx)*co_list[2])/(num_v-1) - 0.2 * v0_step
                co_1 = (i_idx*co_list[1] + (num_v-1-i_idx)*co_list[3])/(num_v-1) - 0.2 * v1_step
                polyline = self.set_poly_resample_from3d(curveData, co_0, co_1, 0, 0.3*wave_length)

                if (self.fence_distort != 0):
                    pnum = len(polyline.points)
                    for j, p in enumerate(polyline.points):
                        dist_factor = dist_factor_h * sin(j/pnum*3.14159) * dist

                        pos = p.co.copy(); pos.resize_3d()
                        dist3d = (noise.noise_vector(pos / wave_length) + dist_normal_offset ) * dist_factor
                        dist3d.resize_4d(); dist3d[3] = 1
                        #print(dist3d)
                        p.co += dist3d

    def get_wire_base_pos(self, co_list, v_align_type, normal, depth, i, num_v, factor):
        if int(v_align_type) == 0:
            co_0_base = self.midpos(co_list[2], co_list[0], factor)
            co_1_base = self.midpos(co_list[3], co_list[1], factor)
        else:
            co_0_base = self.midpos(co_list[0], co_list[2], factor)
            co_1_base = self.midpos(co_list[1], co_list[3], factor)

        if (i == num_v):
            if int(v_align_type) == 0:
                co_0_base = co_list[0].copy()
                co_1_base = co_list[1].copy()
            else:
                co_0_base = co_list[2].copy()
                co_1_base = co_list[3].copy()

        co_0_base += normal * depth * self.h_offset
        co_1_base += normal * depth * self.h_offset

        return co_0_base, co_1_base

    def make_fence_rope_1unitB(self, context, co_list, curveData, normal):

        new_objects = []
        
        interval_v = self.interval_wire_vertical
        v_dist, num_vf, num_v = self.get_num_v_co(co_list, interval_v, self.fit_fence_height)
        v_unit = (co_list[2] - co_list[0]).normalized()

        depth = self.bevel_depth_mesh
        v_align_type = self.v_align_type
        
        v_ini, v_fin = self.get_v_ini_fin(num_v)

        def make_rope(co_0_base, co_1_base):
            
            points = curveData.splines.new('POLY').points
            points.add(1)

            points[0].co = co_0_base.to_4d()
            points[1].co = co_1_base.to_4d()
               

        if (self.single_rope == True):
            co_0_base, co_1_base = self.get_wire_base_pos(co_list, v_align_type, normal, depth, 0, 1, 1-self.pole_appendix_ypos)
            make_rope(co_0_base, co_1_base)

        else:
            for i in range(v_ini, v_fin):
                co_0_base, co_1_base = self.get_wire_base_pos(co_list, v_align_type, normal, depth, i, num_v, i/num_vf)
                make_rope(co_0_base, co_1_base)
            
        return new_objects
            

            

    def make_fence_barbed_wire_1unit(self, co_list, curveData, normal, f_idx):

        time0 = time.time()
        
        twist_length = self.bevel_depth_mesh *  self.twist_width_barb
        
        interval_v = self.interval_wire_vertical
        v_dist, num_vf, num_v = self.get_num_v_co(co_list, interval_v, self.fit_fence_height)
        v_unit = (co_list[2] - co_list[0]).normalized()

        depth = self.bevel_depth_mesh
        spiral_r = self.twist_radius_barb * depth
        distort = self.barb_distort * depth

        v_align_type = self.v_align_type

        count = 0
        barb_count = 0
        barb_interval = self.barb_interval
        random.seed(self.random_seed)

        def set_point_from3d(point, co):
            co_tmp = co.copy();
            co_tmp.resize_4d(); point.co = co_tmp

        def wire_distortB(i, h, num_h, freq, offset, amp):
            
            amp_base = sqrt(sin(3.14159*h/num_h))

            dis_x = 0
            dis_y = 0
            
            dis_x += sin(freq[0]*h + offset[0])*amp_base*amp[0]
            dis_y += cos(freq[0]*h + offset[0])*amp_base*amp[0]
            dis_x += sin(freq[1]*h + offset[1])*amp_base*amp[1]
            dis_y += cos(freq[1]*h + offset[1])*amp_base*amp[1]
            dis_x += sin(freq[2]*h + offset[2])*amp_base*amp[2]
            dis_y += cos(freq[2]*h + offset[2])*amp_base*amp[2]
            dis_x += sin(freq[3]*h + offset[3])*amp_base*amp[3]
            dis_y += cos(freq[3]*h + offset[3])*amp_base*amp[3]

            return dis_x, dis_y

        v_ini, v_fin = self.get_v_ini_fin(num_v)
        #v_ini, v_fin = 1, num_v
        for j in range(0, 2):

            for i in range(v_ini, v_fin):
                count += int(barb_interval * random.random())

                co_0_base, co_1_base = self.get_wire_base_pos(co_list, v_align_type, normal, depth, i, num_v, i/num_vf)

                h_dist = (co_1_base-co_0_base).length
                h_unit = (co_1_base-co_0_base).normalized()
                num_h_f = h_dist / twist_length
                num_h = math.ceil(num_h_f)

                if (True):
                    num_h_f = num_h 
                    num_h += 1
                    
                phase = 3.14159 * j + i%3
                co_0 = co_0_base + spiral_r*(v_unit*cos(phase) + normal*sin(phase))

                distort_freq, distort_offset, distort_amp = self.wire_distort_set_paramB(v_ini, i, f_idx)

                points = curveData.splines.new('POLY').points
                points.add(num_h)
                set_point_from3d(points[0], co_0)

                for h in range(1, num_h+1):
                    phase += 3.14159/4;
                    co_1 = co_0_base + h*h_unit*h_dist/num_h + spiral_r*(v_unit*math.cos(phase) + normal*math.sin(phase))
                    
                    ox0, oy0 = wire_distortB(i, h, num_h, distort_freq, distort_offset, distort_amp)
                    offset_x = normal*ox0 * distort
                    offset_y = v_unit*oy0 * distort
                    co_1 += offset_x + offset_y

                    set_point_from3d(points[h], co_1)

                    if (h>1 and h < num_h-2 and j == 0):
                        if (count % barb_interval == 0):

                            co_0_base_h = co_0_base + h*h_unit*h_dist/num_h + offset_x + offset_y
                            ox1, oy1 = wire_distortB(i, h+1, num_h, distort_freq, distort_offset, distort_amp)

                            dh = h_dist/num_h
                            dx = (ox1-ox0) * distort
                            dy = (oy1-oy0) * distort
                            
                            normal_b = normal - dx/dh * h_unit
                            v_unit_b = v_unit - dy/dh * h_unit
                            h_unit_b = h_unit + dx/dh * normal + dy/dh * v_unit

                            for k in range(0,2):
                                phase = 3.15159*k + barb_count
                                co_b0 = co_0_base_h + 8*spiral_r*(v_unit_b*cos(phase+3.14159/8) + normal_b*sin(phase+3.14159/8))

                                bpoints = curveData.splines.new('POLY').points
                                bpoints.add(20)
                                
                                for sp in range(1, 21):
                                    phase += 0.7853975 #3.14159/4
                                    radius_sp = 2
                                    barb_offset0 = Vector((0,0,0))
                                    barb_offset1 = Vector((0,0,0))
                                    if (sp == 1):
                                        barb_offset0 = -h_unit_b*depth*0.5
                                        
                                    if (sp == 20):
                                        radius_sp = 8
                                        phase += 0.39269875 #3.14159/8
                                        barb_offset1 = h_unit_b*depth*0.5
                                            
                                    co_b1 = co_0_base_h + sp*h_unit_b*depth*0.5 + radius_sp*spiral_r*(-v_unit_b*sin(phase) + normal_b*cos(phase))
                                    
                                    if (sp == 1):
                                        set_point_from3d(bpoints[0], co_b0+barb_offset0)
                                    set_point_from3d(bpoints[sp], co_b1+barb_offset1)
                                        
                                    
                                    co_b0 = co_b1.copy()
                                bpoints[0].radius = 0.3
                                bpoints[-1].radius = 0.3
                            barb_count += 1

                        count += 1
                        
                    co_0 = co_1.copy()

        time1 = time.time()
        print(time1-time0)

    def make_fence_guard_pipe_1unit(self, splines, co_list, curveData, normal, f_idx, type):
        verts = []

        co_list_local = []
        for co in co_list:
            co_list_local.append(co.copy())

        co_list_local[0] += self.h_offset*(co_list[1]-co_list[0]).normalized()
        co_list_local[1] -= self.h_offset*(co_list[1]-co_list[0]).normalized()
        co_list_local[2] += self.h_offset*(co_list[3]-co_list[2]).normalized()
        co_list_local[3] -= self.h_offset*(co_list[3]-co_list[2]).normalized()

        min_x = min_z = 99
        max_x = max_z = -99

        new_s_list = []
        
        for s in splines:
            new_s = curveData.splines.new('BEZIER')
            new_s_list.append(new_s)
            new_s.use_cyclic_u = s.use_cyclic_u
            points = s.bezier_points
            p = new_s.bezier_points.add(len(points) - 1)
            for p, new_p in zip(s.bezier_points, new_s.bezier_points):
                self.copy_bezier_point_from(new_p, p)
                min_x = min(min_x, p.co.x)
                max_x = max(max_x, p.co.x)
                min_z = min(min_z, p.co.z)
                max_z = max(max_z, p.co.z)


        parts_length = max_x - min_x
        f_length = ((co_list_local[2]+co_list_local[0])/2 - (co_list_local[3]+co_list_local[1])/2).length

        scaling_by_extend = {'A':True, 'B':True,
                             'C':True, 'D':False}[type]

        if (scaling_by_extend == True):
            extend = (f_length - parts_length)/2
            for s in new_s_list:
                points = s.bezier_points
                for p in points:
                    co = p.co
                    nx = (co.x - min_x)/(max_x-min_x)
                    if (nx < 0.1):
                        p.co.x           -= extend
                        p.handle_left.x  -= extend
                        p.handle_right.x -= extend
                    if (nx > 0.9):
                        p.co.x           += extend
                        p.handle_left.x  += extend
                        p.handle_right.x += extend
        else:
            scaling = (parts_length/f_length)
            for s in new_s_list:
                points = s.bezier_points
                for p in points:
                    co = p.co
                    p.co.x           *= scaling
                    p.handle_left.x  *= scaling
                    p.handle_right.x *= scaling
                                            
                        
                        

        # Recalc Range for defoemed shape
        min_x = min_z = 99
        max_x = max_z = -99
        for s in new_s_list:
            points = s.bezier_points
            for p in points:
                min_x = min(min_x, p.co.x)
                max_x = max(max_x, p.co.x)
                min_z = min(min_z, p.co.z)
                max_z = max(max_z, p.co.z)
                
                
        for s in new_s_list:
            points = s.bezier_points
            for p in points:
                co, handle_left, handle_right = p.co, p.handle_left, p.handle_right

                nx = (co.x - min_x)/(max_x-min_x)
                nz = (co.z - min_z)/(max_z-min_z)

                p.co = nz * ((co_list_local[3]-co_list_local[2]) * nx + co_list_local[2]) + (1-nz) * ((co_list_local[1]-co_list_local[0]) * nx + co_list_local[0])

                nx = (handle_left.x - min_x)/(max_x-min_x)
                nz = (handle_left.z - min_z)/(max_z-min_z)
                p.handle_left = nz * ((co_list_local[3]-co_list_local[2]) * nx + co_list_local[2]) + (1-nz) * ((co_list_local[1]-co_list_local[0]) * nx + co_list_local[0])

                nx = (handle_right.x - min_x)/(max_x-min_x)
                nz = (handle_right.z - min_z)/(max_z-min_z)
                p.handle_right = nz * ((co_list_local[3]-co_list_local[2]) * nx + co_list_local[2]) + (1-nz) * ((co_list_local[1]-co_list_local[0]) * nx + co_list_local[0])

        connect_pos_list = [0.333, 0.666]
        for pos in connect_pos_list:
            posA = (pos*co_list_local[0] + (1-pos)*co_list_local[2])
            posB = (pos*co_list_local[1] + (1-pos)*co_list_local[3])
            dir = (posB - posA).normalized()
            verts.append(posA)
            verts.append(posA - dir*0.01)
            verts.append(posB)
            verts.append(posB + dir*0.01)

        return verts
            

    def adjust_bezier_tips(self, curveData):

        for s in curveData.splines:
            if len(s.bezier_points) > 2:
                print("!!")
                s.bezier_points[0].handle_right_type = "ALIGNED"
                s.bezier_points[0].handle_left_type = "ALIGNED"
                s.bezier_points[0].handle_left = (s.bezier_points[0].co * 3  - s.bezier_points[1].handle_left) / 2
                #s.bezier_points[0].handle_right = Vector((0,0,0))
                #s.bezier_points[0].handle_left = Vector((0,0,0))

                s.bezier_points[-1].handle_right_type = "ALIGNED"
                s.bezier_points[-1].handle_left_type = "ALIGNED"
                s.bezier_points[-1].handle_left = (s.bezier_points[-1].co + s.bezier_points[-2].handle_right) / 2
        

    def convert_to_bezier(self, curveOB, object, preserve_tilt = False):
        bpy.context.view_layer.objects.active = curveOB
        if (preserve_tilt == True):
            tilt_list_list = []
            for s in curveOB.data.splines:
                tilt_list = []
                for p in s.points:
                    tilt_list.append(p.tilt)

                tilt_list_list.append(tilt_list)
                #print("tilt_list_len0=", len(tilt_list))

        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.curve.spline_type_set(type='BEZIER')
        bpy.ops.curve.handle_type_set(type='AUTOMATIC')
        if (preserve_tilt == True):
            for i, s in enumerate(curveOB.data.splines):
                tilt_list = tilt_list_list[i]

                print("tilt_list_len", len(tilt_list))
                print("bez_len", len(s.bezier_points))

                for j,b in enumerate(s.bezier_points):
                    b.tilt = tilt_list[j]

        if (self.use_adjust_bezier_tips == True):
            self.adjust_bezier_tips(curveOB.data)
                    
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.context.view_layer.objects.active = object

    def is_face_valid(self, f):
        if (f.select == False):
            return False

        if (len(f.verts) != 4):
            print("f.verts != 4")
            return False
        
        return True

    def make_fence(self, context, object, data, material):

        vertices_for_connector = []
        new_objects = []

        curveData = self.make_curve('fence')
        self.adjust_resolution_u(curveData)
        bm = bmesh.from_edit_mesh(data)

        knot_objects = []

        for f in bm.faces:
            
            if self.is_face_valid(f) == False:
                continue

            normal = f.normal

            vlist = self.get_vlist(f)
            co_list = self.get_co_list(vlist)
            
            if self.fence_type == 'CROSS':
                self.make_fence_cross_1unit(co_list, curveData, normal)
                
            if self.fence_type == 'OBLIQUE':
                self.make_fence_oblique_1unit(co_list, curveData, normal)
                
            if self.fence_type == 'WIRE_MESH':
                self.make_fence_wire_1unit(co_list, curveData, normal)

            if self.fence_type == 'ROPE':
                knot_objects += self.make_fence_rope_1unitB(context, co_list, curveData, normal)

            if self.fence_type == 'BARBED_WIRE':
                self.make_fence_barbed_wire_1unit(co_list, curveData, normal, f.index)
            
            if self.fence_type == "GUARD_PIPE":
                col = self.get_asset_colelction(context, "FenceParts")
                guardpipe_type = self.guardpipe_type[4:5]
                pipeOB = self.load_object(context, "GuardPipe."+guardpipe_type, col, use_curve = True)
                splines = pipeOB.data.splines

                vertices_for_connector += self.make_fence_guard_pipe_1unit(splines, co_list, curveData, normal, f.index, guardpipe_type)


        self.set_bevel_depth(curveData)

        rectOB = self.proc_bevel_type(curveData, self.fence_bevel_type, context.scene)
        
        curveOB = self.make_curve_object('fence', curveData, material, context)
        new_objects.append(curveOB)

        if (self.guardpipe_connector == True and vertices_for_connector != []):
            type = self.guardpipe_connector_type[4:5]
            partsOB, connectOB, mod_dist = self.make_instancer_basic(context, vertices_for_connector, "guardpipe_connector_shape."+type, "guardpipe_connector", 1)
            connector_offset = {'A':0.05, 'B':0.05,
                                'C':0.05, 'D':0.00}[guardpipe_type]

            print(connector_offset, type)
            
            self.modifier_input(mod_dist, "Translation", Vector((-connector_offset+self.bevel_depth,0,0)))
            new_objects.append(connectOB)

        for knot in knot_objects:
            knot.parent = curveOB
            new_objects.append(knot)

        if self.fence_type == 'ROPE':

            h_margin = self.h_margin
            if (self.rope_terminal_type == "KNOT"):

                knotOB = self.load_object(context, "rope_terminal_knot")
                mod_knot = self.add_GN_modifier(knotOB, "KnotScale")
                self.modifier_attrib(mod_knot, "Selection", "Ring")
                self.modifier_input(mod_knot, "Expand", float(self.rope_knot_radius-0.105))
                self.modifier_input(mod_knot, "Radius", float(self.bevel_depth_mesh))
                #h_margin += self.rope_knot_radius + 0.04
                
            mod_catenary = self.add_GN_modifier(curveOB, "Catenary")
            self.modifier_input(mod_catenary, "Tension", self.catenary_a)
            self.modifier_input(mod_catenary, "Offset", h_margin)

            if (self.rope_terminal_type == "KNOT"):
                mod_cap = self.add_GN_modifier(curveOB, "CurveCaps")
                self.modifier_input(mod_cap, "Scale", float(1))
                self.modifier_input(mod_cap, "Object", knotOB)
            

            end_offset = 0
            if (self.rope_terminal_type == "EYE_TERMINAL"):
                terminal_obj = self.load_object(context, "rope_terminal_shape")
                mod_terminal = self.add_GN_modifier(curveOB, "ObjectAtCurveEnd")
                self.modifier_input(mod_terminal, "Object", terminal_obj)
                self.modifier_input(mod_terminal, "Scale", self.rope_terminal_size)
                self.modifier_input(mod_terminal, "Angle", self.rope_terminal_rot)
                end_offset = self.rope_terminal_size * 0.2

            if (self.rope_type == "CHAIN"):
                end_offset += self.chain_scale * 0.5
                mod_chain = self.add_GN_modifier(curveOB, "Chain")
                self.modifier_attrib(mod_chain, "Selection_L", "L")
                self.modifier_attrib(mod_chain, "Selection_R", "R")
                self.modifier_input(mod_chain, "Scale", self.chain_scale) 
                self.modifier_input(mod_chain, "Elongate", self.chain_elongate)
                self.modifier_input(mod_chain, "Radius", self.bevel_depth_mesh)
                self.modifier_input(mod_chain, "OffsetFactor", float(2))
                self.modifier_input(mod_chain, "Start", float(end_offset))
                self.modifier_input(mod_chain, "End", float(end_offset))
            
            if (self.rope_type == "ROPE"):
                mod_rope = self.add_GN_modifier(curveOB, "Rope")
                self.modifier_input(mod_rope, "Radius", self.bevel_depth_mesh)
                self.modifier_input(mod_rope, "Start", float(end_offset))
                self.modifier_input(mod_rope, "End", float(end_offset))
                                    
                
        if self.fence_type == 'WIRE_MESH':
            if self.curve_type == "BEZIER":
                self.convert_to_bezier(curveOB, object)

        if (rectOB != None):
            rectOB.parent = curveOB
            new_objects.append(rectOB)
            
        return new_objects

    def get_init_co_truss(self, vlist, normal, b_vec, b_unit):
        bevel_depth = self.bevel_depth_truss
        offset      = self.offset_truss
        flip        = self.flip_truss
        
        if abs(normal.dot((1,0,0))) > 0.5:
            if normal.dot((1,0,0)) > 0 and flip:
                co_0 = vlist[0].co.copy()
                co_0 += b_unit * bevel_depth * offset
                offset_count = 1;
            else:
                co_0 = vlist[0].co.copy() + b_vec
                co_0 -= b_unit * bevel_depth * offset
                offset_count = -1;
        elif abs(normal.dot((0,1,0))) > 0.5:
            if normal.dot((0,1,0)) > 0 and flip:
                co_0 = vlist[0].co.copy()
                co_0 += b_unit * bevel_depth * offset
                offset_count = 1;
            else:
                co_0 = vlist[0].co.copy() + b_vec
                co_0 -= b_unit * bevel_depth * offset
                offset_count = -1;
        else:
            if normal.dot((0,0,1)) > 0 and flip:
                co_0 = vlist[0].co.copy()
                co_0 += b_unit * bevel_depth * offset
                offset_count = 1;
            else:
                co_0 = vlist[0].co.copy() + b_vec
                co_0 -= b_unit * bevel_depth * offset
                offset_count = -1;

        return co_0, offset_count

    def make_truss_face(self, vlist, curveData, normal, dir):

        # a : along long axis
        # b : along short axis

        if (dir == 'VERTICAL'):
            b_vec  = (vlist[1].co-vlist[0].co)
            b_dist = (vlist[1].co-vlist[0].co).length
            b_unit = (vlist[1].co-vlist[0].co).normalized()

            a_vec  = (vlist[2].co - vlist[0].co)
            a_dist = (vlist[2].co - vlist[0].co).length
            a_unit = (vlist[2].co - vlist[0].co).normalized()

        if (dir == 'HORIZONTAL'):
            b_vec  = (vlist[2].co-vlist[0].co)
            b_dist = (vlist[2].co-vlist[0].co).length
            b_unit = (vlist[2].co-vlist[0].co).normalized()

            a_vec  = (vlist[1].co - vlist[0].co)
            a_dist = (vlist[1].co - vlist[0].co).length
            a_unit = (vlist[1].co - vlist[0].co).normalized()

        bevel_depth = self.bevel_depth_truss
        offset      = self.offset_truss
        aspect      = self.aspect_truss

        co_0, offset_count = self.get_init_co_truss(vlist, normal, b_vec, b_unit)
        co_1 = co_0.copy()

        if (self.angle_type_truss == "0"):
            a_step = b_dist * aspect
        else:
            a_step = self.step_length_truss

        num = int(a_dist / a_step)
        if (num == 0):
            if (self.fit_truss == True):
                num = 1
            else:
                return
            
        if (self.fit_truss == True):
            a_step *= (a_dist / a_step) / num

        polyline = curveData.splines.new('POLY')
        polyline.points.add(num*2)
        self.set_point_from3d(polyline.points[0], co_0)

        for count in range(num):
            co_1 += offset_count * b_vec
            co_1 -= offset_count * b_unit * bevel_depth * offset * 2
            co_1 += a_unit * (1-self.joint_size_truss) * a_step
            self.set_point_from3d(polyline.points[count*2+1], co_1)

            co_1 += a_unit * self.joint_size_truss * a_step
            self.set_point_from3d(polyline.points[count*2+2], co_1)

            offset_count *= -1
            if (count > 100):
                break

        unit_z = Vector((0,0,1))
        unit_0 = (polyline.points[1].co - polyline.points[0].co).normalized()
        unit_0.resize_3d()
        
        tilt = self.get_tilt(unit_0, unit_z, normal)
        #if (math.isclose(unit_0.dot(Vector((1,0,0))), 0, abs_tol=1e-6)) and (math.isclose(unit_0.dot(Vector((0,1,0))), 0, abs_tol=1e-6)):
        #    tilt = math.atan2(normal.y, normal.x)
        #else:
        #    tilt = self.get_tilt(unit_0, unit_z, normal)

        for p in polyline.points:
            p.tilt = tilt

    def make_truss(self, context, object, data, material):

        new_objects = []

        curveData = self.make_curve('truss')
        bm = bmesh.from_edit_mesh(data)

        for f in bm.faces:
            if self.is_face_valid(f) == False:
                continue

            normal = f.normal
            vlist = self.get_vlist(f)

            h_dist = (vlist[1].co-vlist[0].co).length
            v_dist = (vlist[2].co-vlist[0].co).length

            if h_dist > v_dist:
                self.make_truss_face(vlist, curveData, normal, 'HORIZONTAL')
            else:
                self.make_truss_face(vlist, curveData, normal, 'VERTICAL')

        self.set_bevel_depth(curveData)
        self.proc_bevel_type(curveData, self.type_bevel_truss, context.scene)
        
        curveOB = self.make_curve_object('fence', curveData, material, context)
        new_objects.append(curveOB)

        return new_objects








    def interpolate_pos(self, co_list, j, num):
        if (num != 1):
            co = (j*co_list[0] + (num-1-j)*co_list[1])/(num-1)
        else:
            co = (0.5*co_list[0] + 0.5*co_list[1])
        return co
                    

    def make_guardrail_main(self, curveData, edge_list_list, margin, slide, n_offset, num, radius):

        for edge_list in edge_list_list:

            num_egde = len(edge_list)
            fin = num
            num_line = max(1, fin - 1)
            for j in range(0, fin):
        
                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(num_egde - 1)

                for i in range(0, num_egde):
                    e = edge_list[i]
                    normal = self.get_edge_normal(e)
                    ztilt = self.get_edge_ztilt(e)

                    co_list = self.get_co_list_from_edge(e, margin, slide)
                    co = self.interpolate_pos(co_list, j, num)
                    co += normal * n_offset
                    self.set_point_from3d(points[i], co)
                    points[i].tilt = ztilt
                    points[i].radius = radius

    def make_pipe_end_curve_rim(self, points, co00, co01, co10, co11, size, normal, n_offset):
        t0 = (co00-co10).normalized()
        t1 = (co01-co11).normalized()
        bn = (co00 + t0*size - co01 - t1*size)
        co00 += normal * n_offset
        co01 += normal * n_offset

        self.set_point_from3d(points[0], co00)
        self.set_point_from3d(points[1], co00 + t0*(size-0.1))
        self.set_point_from3d(points[2], co00 + t0*size - bn*0.1)
        self.set_point_from3d(points[3], co01 + t1*size + bn*0.1)
        self.set_point_from3d(points[4], co01 + t1*(size-0.1))
        self.set_point_from3d(points[5], co01)

    def make_guardrail_pipe_end_curve_1(self, curveData, edge_list_list, margin, slide, n_offset, num, radius):

        s_index_list = []
        if (num == 1):
            return s_index_list

        def make_main_rim(points, e0, e1, margin, slide):

            size = self.guardrail_endcap_length
            normal = self.get_edge_normal(e0)
            
            co0_list = self.get_co_list_from_edge(e0, margin, slide)
            co1_list = self.get_co_list_from_edge(e1, margin, slide)
            co00 = co0_list[0]
            co01 = co0_list[1]
            co10 = co1_list[0]
            co11 = co1_list[1]
            self.make_pipe_end_curve_rim(points, co00, co01, co10, co11, size, normal, n_offset)

        for edge_list in edge_list_list:

            polyline = curveData.splines.new('POLY')
            points = polyline.points
            points.add(5)
                
            e0 = edge_list[0]; e1 = edge_list[1]
            make_main_rim(points, e0, e1, margin, slide)
            s_index_list.append(self.get_spline_index(polyline))

            polyline = curveData.splines.new('POLY')
            points = polyline.points
            points.add(5)
                
            e0 = edge_list[-1]; e1 = edge_list[-2]
            make_main_rim(points, e0, e1, margin, slide)
            s_index_list.append(self.get_spline_index(polyline))

        def make_main(points, e0, e1, margin, slide, arch):
                                  
            size = self.guardrail_endcap_length
            normal = self.get_edge_normal(e0)

            co0_list = self.get_co_list_from_edge(e0, margin, slide)
            co1_list = self.get_co_list_from_edge(e1, margin, slide)

            co0 = (j*co0_list[0] + (num_line-j)*co0_list[1])/num_line + normal * n_offset
            co1 = (j*co1_list[0] + (num_line-j)*co1_list[1])/num_line + normal * n_offset
            self.set_point_from3d(points[0], co0)
            self.set_point_from3d(points[1], co0 + (co0 - co1).normalized() * ((size - 0.03) + 0.25*arch))
            
        for edge_list in edge_list_list:

            num_egde = len(edge_list)
            fin = num
            num_line = max(1, fin-1)
            
            for j in range(1, fin-1):

                x = j/num_line - 0.5
                arch = 0.25 - x*x
        
                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(1)

                e0 = edge_list[0]; e1 = edge_list[1]
                make_main(points, e0, e1, margin, slide, arch)

                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(1)

                e0 = edge_list[-1]; e1 = edge_list[-2]
                make_main(points, e0, e1, margin, slide, arch)

        return s_index_list

    def make_guardrail_pipe_end_curve_2(self, curveData, edge_list_list, margin, slide, n_offset, num, radius):

        print("num=", num)
        s_index_list = []
        if (num == 1):
            return s_index_list

        def make_main_rim(points, e0, e1, margin, slide, i, num):

            size = self.guardrail_endcap_length
            normal = self.get_edge_normal(e0)
            
            co0_list = self.get_co_list_from_edge(e0, margin, slide)
            co1_list = self.get_co_list_from_edge(e1, margin, slide)
            
            co00 = self.interpolate_pos(co0_list, i*2, num)
            co01 = self.interpolate_pos(co0_list, i*2+1, num)
            co10 = self.interpolate_pos(co1_list, i*2, num)
            co11 = self.interpolate_pos(co1_list, i*2+1, num)
                
            self.make_pipe_end_curve_rim(points, co00, co01, co10, co11, size, normal, n_offset)

        for edge_list in edge_list_list:

            for i in range(0, int(num/2)):

                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(5)
                
                e0 = edge_list[0]; e1 = edge_list[1]
                make_main_rim(points, e0, e1, margin, slide, i, num)
                s_index_list.append(self.get_spline_index(polyline))

                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(5)
                
                e0 = edge_list[-1]; e1 = edge_list[-2]
                make_main_rim(points, e0, e1, margin, slide, i, num)
                s_index_list.append(self.get_spline_index(polyline))

        return s_index_list
    

    def make_guardrail_pipe_end_simple(self, curveData, edge_list_list, margin, slide, n_offset, num, radius):
        for edge_list in edge_list_list:

            num_egde = len(edge_list)
            fin = num
            num_line = max(1, fin-1)

            def make_main(points, e0, e1, margin, slide, arch, normal):
                                  
                size = self.guardrail_endcap_length

                co0_list = self.get_co_list_from_edge(e0, margin, slide)
                co1_list = self.get_co_list_from_edge(e1, margin, slide)
        
                if (num != 1):
                    co0 = (j*co0_list[0] + (num_line-j)*co0_list[1])/num_line + normal * n_offset
                    co1 = (j*co1_list[0] + (num_line-j)*co1_list[1])/num_line + normal * n_offset
                else:
                    co0 = (0.5*co0_list[0] + 0.5*co0_list[1]) + normal * n_offset
                    co1 = (0.5*co1_list[0] + 0.5*co1_list[1]) + normal * n_offset

                x_unit = (co0_list[1]-co0_list[0]).normalized()
                y_unit = (co0 - co1).normalized()
                co_end = co0 + y_unit * (size + 0.25*arch)

                self.set_point_from3d(points[0], co0)
                
                self.set_point_from3d(points[1], co_end)
            
                return co_end, x_unit, y_unit
                
            for j in range(0, fin):

                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(1)

                e0 = edge_list[0]; e1 = edge_list[1]
                normal = self.get_edge_normal(e0)
                center, x_unit, y_unit = make_main(points, e0, e1, margin, slide, 0, normal)

                cap_size = self.guardrail_endcap_size
                self.make_cap(curveData, center, cap_size, radius, x_unit, y_unit, normal, 'TOP', self.guardrail_pipe_captype)

                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(1)

                e0 = edge_list[-1]; e1 = edge_list[-2]
                normal = self.get_edge_normal(e0)
                center, x_unit, y_unit = make_main(points, e0, e1, margin, slide, 0, normal)
                
                self.make_cap(curveData, center, cap_size, radius, x_unit, y_unit, normal, 'TOP', self.guardrail_pipe_captype)
                

    def get_spline_index(self, s):
        return list(s.id_data.splines).index(s)

    def adjust_guardrail_pipe_end_curve(self, curveData, spline_list):
        for s_index in spline_list:
            s = curveData.splines[s_index]
            p0 = s.bezier_points[0]
            p1 = s.bezier_points[1]
            p1.handle_left_type = 'FREE'
            p1.handle_right_type = 'FREE'
            p1.handle_left = p0.co / 3 + p1.co * 2 / 3
            p1.handle_right = p1.co + (p1.co - p1.handle_left).normalized() * 0.05

            p5 = s.bezier_points[5]
            p4 = s.bezier_points[4]
            p4.handle_left_type = 'FREE'
            p4.handle_right_type = 'FREE'
            p4.handle_right = p5.co / 3 + p4.co * 2 / 3
            p4.handle_left = p4.co + (p4.co - p4.handle_right).normalized() * 0.05
            

    def make_guardrail_end_shape(self):
        railData = bpy.data.curves.new('railend_shape', type='CURVE')
        railData.dimensions = '3D'
        railData.resolution_u = 1

        polyline = railData.splines.new('POLY')
        polyline.use_smooth = False
        points = polyline.points
        points.add(7)
        polydata=(
            (-0.0278,0.514,0,1),(-0.0019,0.514,0,1),(0.1062,0.3738,0,1),
            (0.1062,0.2262,0,1),(0.1062,-0.2262,0,1),(0.1062,-0.3738,0,1),
            (-0.0019,-0.514,0,1),(-0.0278,-0.514,0,1),
        )
        #polydata=(
        #    (-0.0498,0.6853,0,1),(-0.0153,0.6853,0,1),(0.1289,0.4984,0,1),
        #    (0.1289,0.3016,0,1),(0.1289,-0.3016,0,1),(0.1289,-0.4984,0,1),
        #    (-0.0153,-0.6853,0,1),(-0.0498,-0.6853,0,1),
        #)
        
        for p, pdata in zip (points, polydata):
            p.co = pdata

        return bpy.data.objects.new(railData.name, railData)
        
    def make_guardrail_end(self, context, material, endData, main_splines, edge_list_list, idx, radius, curve_radius):

        curve_radius = radius * curve_radius

        for e_list, s in zip(edge_list_list, main_splines):
            polyline = endData.splines.new('POLY')
            points = polyline.points

            e = e_list[idx]
            bezpoint = s.bezier_points[idx]
            tilt = bezpoint.tilt
            if (idx == 0):
                tilt *= -1

            print("tilt(end)=", tilt)
            normal = self.get_edge_normal(e)
            if (idx == 0):
                tangent = (bezpoint.handle_right - bezpoint.co).normalized()
                normal -= (normal.dot(tangent)*tangent)
                normal = normal.normalized()
                points.add(12)
            
                co = bezpoint.co + tangent*0.2*curve_radius; co.resize_4d()
                points[12].co = co
                co = bezpoint.co + tangent*0.0*curve_radius; co.resize_4d()
                points[11].co = co
                co = bezpoint.co - tangent*1.3*curve_radius; co.resize_4d()
                points[10].co = co
                for i in range(1, 10):
                    theta = 3.14159 * (9-i) / 8
                    co = bezpoint.co - tangent*1.5*curve_radius - tangent*sin(theta)*curve_radius + normal*(cos(theta)-1)*curve_radius; co.resize_4d()
                    points[i].co = co
                co = bezpoint.co + (-tangent*1.3 - normal*2)*curve_radius; co.resize_4d()
                points[0].co = co
                

            if (idx == -1):
                tangent = (bezpoint.co - bezpoint.handle_left).normalized()
                normal -= (normal.dot(tangent)*tangent)
                normal = normal.normalized()

                points.add(12)
                co = bezpoint.co - tangent*0.2*curve_radius; co.resize_4d()
                points[0].co = co
                co = bezpoint.co - tangent*0.0*curve_radius; co.resize_4d()
                points[1].co = co
                co = bezpoint.co + tangent*1.3*curve_radius; co.resize_4d()
                points[2].co = co
                for i in range(3, 12):
                    theta = 3.14159 * (i-3) / 8
                    co = bezpoint.co + tangent*1.5*curve_radius + tangent*sin(theta)*curve_radius + normal*(cos(theta)-1)*curve_radius; co.resize_4d()
                    points[i].co = co
                co = bezpoint.co - (-tangent*1.3 + normal*2)*curve_radius; co.resize_4d()
                points[12].co = co
                    

            for i in range(0, 13):
                points[i].tilt = tilt
                points[i].radius = radius

        #return self.make_object(endData.name, endData, material, context)
        

    def make_guardrail_connector_pipe(self, curveData, edge_list_list, margin, slide, n_offset, radius, main_num):
        # コネクト部
        radius_A = self.guardrail_connector_radius_A
        radius_B = self.guardrail_connector_radius_B
        #radius_C = self.guardrail_connector_radius_C
        #opposite = self.guardrail_connector_opposite
        for edge_list in edge_list_list:

            num_egde = len(edge_list)
            fin = self.guardrail_num
            num_line = max(1, fin - 1)
            for j in range(0, fin):
                for i in range(0, num_egde):
                    e = edge_list[i]
                    normal = self.get_edge_normal(e)
                    co_list = self.get_co_list_from_edge(e, margin, slide)
                    if (main_num != 1):
                        co_0 = (j*co_list[0] + (num_line-j)*co_list[1])/num_line
                    else:
                        co_0 = (0.5*co_list[0] + 0.5*co_list[1])
                        
                    co_1 = co_0 + normal * n_offset
                    polyline = self.set_poly_from3d(curveData, co_1, co_0, 0, radius = radius_B)
                    points = polyline.points
                    #if (opposite != 0):
                    #    co_0 = co_0 - normal * opposite * 0.1
                    #    self.append_poly_from3d(points, co_0, 0, radius_C)
                    #    co_0 = co_0 - normal * opposite * 0.9
                    #    self.append_poly_from3d(points, co_0, 0, radius_C)
                    #    self.append_round_top(polyline, -normal, co_0, 0, radius_C * 0.2, radius = radius_C)

        vertices_for_vis = []

        # 形状に沿ってパイプ部
        connect_size = radius * self.guardrail_connector_size
        # カーブを追加するので、直接ループしちゃ駄目
        for i in range(0, main_num):
            bezline = curveData.splines[i]
            if (bezline.type == 'POLY'):
                continue
            #print(bezline)
            for p in bezline.bezier_points:
                co_0 = p.co+(p.handle_left-p.co).normalized()*connect_size
                co_1 = p.co+(p.handle_right-p.co).normalized()*connect_size
                self.set_poly_from3d(curveData, co_0, co_1, 0, radius = radius_A)
                
                tangent = (co_1 - co_0).normalized()
                normal = tangent.cross((0,0,1)).normalized()
                v = [0, 0, 1]
                theta = +p.tilt# + 3.14159/2

                rot_matrix_base = self.get_rot_matrix_along_tangent(tangent, normal, theta)

                if (self.guardrail_vis == True):
                    binormal = Vector(tuple(np.dot(rot_matrix_base, [0, 1, 0])))

                    vis_num = self.guardrail_vis_number
                    for j in range(0, vis_num):
                        f = 0.4 / vis_num + 0.05
                        co_list = [(1-f)*co_0+f*co_1, f*co_0+(1-f)*co_1]
                        
                        if (self.guardrail_vis_direction == 'HORIZONTAL'):
                            
                            co_vis = self.interpolate_pos(co_list, j, vis_num)
                            co_vis += normal * radius * radius_A
                            vertices_for_vis += self.set_instancer_verts(rot_matrix_base, co_vis)
                            
                        if (self.guardrail_vis_direction == 'VERTICAL'):
                        
                            rot_matrixB = self.rotation_matrix(tangent, -3.14159/2+theta)
                            rot_matrix = np.dot(rot_matrixB, rot_matrix_base)
                            
                            co_vis0 = self.interpolate_pos(co_list, j, vis_num)
                            co_vis = co_vis0 + binormal * radius * radius_A
                            vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis)
                            
                            rot_matrixB = self.rotation_matrix(tangent, +3.14159/2+theta)
                            rot_matrix = np.dot(rot_matrixB, rot_matrix_base)

                            co_vis = co_vis0 - binormal * radius * radius_A
                            vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis)
                            
        return vertices_for_vis

    def evaluate_bezier(self, p0, p1, eval_pos):
        bez_list = mathutils.geometry.interpolate_bezier(p0.co, p0.handle_right, p1.handle_left, p1.co, 11)
        idx = int(eval_pos * 10)
        f = eval_pos * 10 - idx
        pos = (1-f)*bez_list[idx] + f*bez_list[idx+1]
        return pos
    
    def make_guardrail_connector_rail(self, curveData, edge_list_list, slide, n_offset, radius):

        # コネクト部
        for edge_list in edge_list_list:
            num_egde = len(edge_list)
            for i in range(0, num_egde):
                e = edge_list[i]
                normal = self.get_edge_normal(e)
                co_list = self.get_co_list_from_edge(e, 0, slide)
                co_0 = (co_list[0] + co_list[1])/2

        vertices_for_vis = []
        vertices_for_connect = []

        # 形状に沿ってパイプ部
        connect_size = radius * self.guardrail_connector_size * 0.2
        # カーブを追加するので、直接ループしちゃ駄目
        num = len(curveData.splines)
        for i in range(0,num):
            bezline = curveData.splines[i]
            if (bezline.type == 'POLY'):
                continue
            #print(bezline)
            bez_index = 0
            for p in bezline.bezier_points:
                if (bez_index > 0):
                    p_minus = bezline.bezier_points[bez_index-1]
                    dist = (p.co - p_minus.co).length
                    eval_pos = 1-connect_size/dist*0.5
                    pos0 = self.evaluate_bezier(p_minus, p, eval_pos)
                    co_0 = p.co+(pos0-p.co).normalized()*connect_size
                else:
                    co_0 = p.co+(p.handle_left-p.co).normalized()*connect_size
                
                if (bez_index < len(bezline.bezier_points) - 1):
                    p_plus = bezline.bezier_points[bez_index+1]
                    dist = (p_plus.co-p.co).length
                    eval_pos = connect_size/dist*0.5
                    pos1 = self.evaluate_bezier(p, p_plus, eval_pos)
                    co_1 = p.co+(pos1-p.co).normalized()*connect_size
                else:
                    co_1 = p.co+(p.handle_right-p.co).normalized()*connect_size

                tangent = (co_1 - co_0).normalized()
                normal = tangent.cross((0,0,1)).normalized()
                v = [0, 0, 1]
                theta = +p.tilt# + 3.14159/2
                
                rot_matrixA = np.array([[tangent.x, 0, normal.x],
                                        [tangent.y, 0, normal.y],
                                        [tangent.z, 1, normal.z]])
                rot_matrixB = self.rotation_matrix(tangent, theta)

                rot_matrix = np.dot(rot_matrixB, rot_matrixA)
                
                offset = np.dot(rot_matrix, v)
                #print(offset)
                #print(tuple(offset))
                co_0 -= Vector(tuple(offset)) * radius * 0.035
                co_1 -= Vector(tuple(offset)) * radius * 0.035
                
                self.set_poly_from3d(curveData, co_0, co_1, p.tilt, radius = radius)

                if (self.guardrail_vis == True):

                    binormal = Vector(tuple(np.dot(rot_matrix, [0, 1, 0])))

                    co_vis_0 = 0.2*co_0 + 0.8*co_1
                    co_vis_1 = 0.8*co_0 + 0.2*co_1
                    co_vis_connect = 0.5*co_0 + 0.5*co_1 - normal * radius * 0.005

                    vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis_0)
                    vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis_1)

                    vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis_0 + (normal*0.13*0.75 + binormal*0.3)*radius)
                    vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis_1 + (normal*0.13*0.75 + binormal*0.3)*radius)

                    vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis_0 + (normal*0.13*0.75 - binormal*0.3)*radius)
                    vertices_for_vis += self.set_instancer_verts(rot_matrix, co_vis_1 + (normal*0.13*0.75 - binormal*0.3)*radius)
                    
                    vertices_for_connect += self.set_instancer_verts(rot_matrix, co_vis_connect)
                    
                bez_index += 1
                    
        return vertices_for_vis, vertices_for_connect

        
    def make_guardrail(self, context, object, data, material):

        col = self.get_asset_colelction(context, "FenceParts")

        new_objects = []
        
        curveData = self.make_curve('guardrail')
        self.adjust_resolution_u(curveData)
        bm = bmesh.from_edit_mesh(data)

        n_offset = self.guardrail_offset

        edge_list_list = self.make_edge_list_list(bm)
        for e_list in edge_list_list:
            self.adjust_elist_direction(e_list)

        if (len(edge_list_list) == 0):
            return {'FINISHED'}

        margin = self.guardrail_margin
        slide = self.guardrail_slide

        if (self.guardrail_type == 'PIPE'):
            radius = self.bevel_depth_guardrail
            curveData.twist_mode = "MINIMUM"
            self.make_guardrail_main(curveData, edge_list_list, margin, slide, n_offset, self.guardrail_num, 1)

            main_num = len(curveData.splines)
            
            s_index_list = []
            if (self.guardrail_pipe_end == 'CURVE_1'):
                s_index_list = self.make_guardrail_pipe_end_curve_1(curveData, edge_list_list, margin, slide, n_offset, self.guardrail_num, 1)
            if (self.guardrail_pipe_end == 'CURVE_2'):
                s_index_list = self.make_guardrail_pipe_end_curve_2(curveData, edge_list_list, margin, slide, n_offset, self.guardrail_num, 1)

            if (self.guardrail_pipe_end == 'CAP'):
                curveDataCap = self.make_curve('guardrail_cap')
                self.make_guardrail_pipe_end_simple(curveDataCap, edge_list_list, margin, slide, n_offset, self.guardrail_num, radius)
                self.set_bevel_depth(curveDataCap)
                self.proc_bevel_type(curveDataCap, 'CYLINDER', context.scene)
                capOB = self.make_curve_object(curveDataCap.name, curveDataCap, material, context)

            self.set_bevel_depth(curveData)
            self.proc_bevel_type(curveData, 'CYLINDER', context.scene)

            curveOB = self.make_curve_object(curveData.name, curveData, material, context)
            new_objects.append(curveOB)
            self.convert_to_bezier(curveOB, object, preserve_tilt = True)

            if (s_index_list != []):
                self.adjust_guardrail_pipe_end_curve(curveData, s_index_list)

            curveData = curveOB.data
            if (self.guardrail_connector == True):
                vertices_for_vis = self.make_guardrail_connector_pipe(curveData, edge_list_list, margin, slide, n_offset, radius, main_num)

                if (vertices_for_vis != []):
                    partsOB, visOB, mod_dist = self.make_instancer_vis(context, vertices_for_vis, "vis_shape.A", "guardrail_vis", 4*radius*self.guardrail_vis_size)
                    self.modifier_input(mod_dist, "Translation", Vector((-0.01, 0, 0)))

                    new_objects.append(visOB)
                   

            if (self.guardrail_pipe_end == 'CAP'):
                capOB.parent = curveOB

        if (self.guardrail_type == 'RAIL'):
            
            
            radius = self.guardrail_rail_radius
            curveData.use_fill_caps = False
            curveData.twist_mode = "Z_UP"

            self.make_guardrail_main(curveData, edge_list_list, margin, slide, n_offset, 1, radius)

            self.set_bevel_depth(curveData)
            self.proc_bevel_type(curveData, 'CYLINDER', context.scene)

            curveOB = self.make_curve_object('guardrail', curveData, material, context)
            new_objects.append(curveOB)
            self.convert_to_bezier(curveOB, object, preserve_tilt = True)

            if "rail_shape" in bpy.data.objects:
                rail_shape_OB = bpy.data.objects["rail_shape"]
            else:
                rail_shape_OB = self.load_object(context, "rail_shape", col_to_belong=col, use_curve = True)
                
            curveData.bevel_mode = 'OBJECT'
            curveData.bevel_object = rail_shape_OB
            
            if (self.guardrail_connector == True):
                vertices_for_vis, vertices_for_connector = self.make_guardrail_connector_rail(curveData, edge_list_list, slide, n_offset, radius)

                if (vertices_for_vis != []):
                    partsOB, visOB, mod_dist = self.make_instancer_vis(context, vertices_for_vis, "vis_shape.A", "guardrail_vis", 0.5*radius*self.guardrail_vis_size)
                    new_objects.append(visOB)

                if (vertices_for_connector != []):
                    partsOB, connectOB, mod_dist = self.make_instancer_basic(context, vertices_for_connector, "guardrail_connect_shape", "guardrail_connector", 0.5*radius)
                    new_objects.append(connectOB)
                    
            if (self.guardrail_endcap == True):

                if "railend_shape" in bpy.data.objects:
                    end_shape_OB = bpy.data.objects["railend_shape"]
                else:
                    end_shape_OB = self.load_object(context, "railend_shape", col_to_belong=col, use_curve = True)
                    
                endData = bpy.data.curves.new('railEnd', type='CURVE')
                endData.dimensions = '3D'
                endData.resolution_u = 2
                self.make_guardrail_end(context, material, endData, curveOB.data.splines, edge_list_list, 0, radius, self.guardrail_endcap_length)
                self.make_guardrail_end(context, material, endData, curveOB.data.splines, edge_list_list, -1, radius, self.guardrail_endcap_length)
                
                endOB = self.make_curve_object(endData.name, endData, material, context)
            
                endData.bevel_mode = 'OBJECT'
                endData.bevel_object = end_shape_OB

                self.convert_to_bezier(endOB, object, preserve_tilt = True)
                endOB.parent = curveOB
                

        return new_objects

    def make_carstop_A(self, co_list, curveData, footer_co_list, footer_tilt_list, tilt):

        bezline = curveData.splines.new('BEZIER')
        points = bezline.bezier_points

        cr = self.carstop_curve_radius

        p = points

        co_0 = co_list[0] + (co_list[0]-co_list[2]).normalized() * self.extend_pole_bottom
        co_1 = co_list[1] + (co_list[1]-co_list[3]).normalized() * self.extend_pole_bottom
        
        if (co_list[3]-co_list[2]).length > cr * 2.2:
            p.add(5)
            p[0].co = co_0
            p[0].handle_left  = p[0].co - (co_list[2]-co_list[0])/3
            p[0].handle_right = p[0].co + (co_list[2]-co_list[0])/3

            p[1].co = co_list[2] + (co_list[0] - co_list[2]).normalized() * cr * 0.9
            p[1].handle_right = co_list[2] + (co_list[0] - co_list[2]).normalized() * cr * 0
            p[1].handle_left = p[1].co - (p[1].handle_right - p[1].co)
        
            p[2].co = co_list[2] + (co_list[3]-co_list[2]).normalized() * cr 
            p[2].handle_left  = co_list[2] + (co_list[3]-co_list[2]).normalized() * cr
            p[2].handle_right = p[2].co - (p[2].handle_left - p[2].co)

            p[3].co = co_list[3] - (co_list[3]-co_list[2]).normalized() * cr 
            p[3].handle_right = co_list[3] - (co_list[3]-co_list[2]).normalized() * cr * 0.9
            p[3].handle_left = p[3].co - (p[3].handle_right - p[3].co)
        
            
            p[4].co = co_list[3] + (co_list[1] - co_list[3]).normalized() * cr
            p[4].handle_left = co_list[3] + (co_list[1] - co_list[3]).normalized() * cr * 0
            p[4].handle_right = p[4].co - (p[4].handle_left - p[4].co)
            
            p[5].co = co_1
            p[5].handle_left  = p[5].co + (co_list[3] - co_list[1])/3
            p[5].handle_right = p[5].co - (co_list[3] - co_list[1])/3

        else:
            p.add(4)
            p[0].co = co_0
            p[0].handle_left  = p[0].co - (co_list[2]-co_list[0])/3
            p[0].handle_right = p[0].co + (co_list[2]-co_list[0])/3

            p[1].co = co_list[2] + (co_list[0] - co_list[2]).normalized() * cr * 0.9
            p[1].handle_right = co_list[2] + (co_list[0] - co_list[2]).normalized() * cr * 0
            p[1].handle_left = p[1].co - (p[1].handle_right - p[1].co)
        
            p[2].co = (co_list[2] + co_list[3])/2
            p[2].handle_left  = p[2].co - (co_list[3] - co_list[2]).normalized() * 0.02
            p[2].handle_right = p[2].co + (co_list[3] - co_list[2]).normalized() * 0.02
            
            p[3].co = co_list[3] + (co_list[1] - co_list[3]).normalized() * cr
            p[3].handle_left = co_list[3] + (co_list[1] - co_list[3]).normalized() * cr * 0
            p[3].handle_right = p[3].co - (p[3].handle_left - p[3].co)
            
            p[4].co = co_1
            p[4].handle_left  = p[4].co + (co_list[3] - co_list[1])/3
            p[4].handle_right = p[4].co - (co_list[3] - co_list[1])/3

        footer_co_list.append(co_0)
        footer_tilt_list.append(tilt)
        footer_co_list.append(co_1)
        footer_tilt_list.append(tilt)
        
        if (self.pole_not_vertical == True):
            c0 = (co_list[0]+co_list[2])/2 + (co_list[2]-co_list[0]).normalized() * self.carstop_offset
            c1 = (co_list[1]+co_list[3])/2 + (co_list[3]-co_list[1]).normalized() * self.carstop_offset
            self.set_poly_from3d(curveData, c0, c1, 0, radius = self.pole_radius_ratio_horizontal)

        
    def set_carstop_ring_verts(self, co_list):

        verts = []

        co_0_base = self.midpos(co_list[0], co_list[2], self.pole_appendix_ypos)
        co_1_base = self.midpos(co_list[1], co_list[3], self.pole_appendix_ypos)
        tangent = self.midpos((co_list[1]-co_list[0]), (co_list[3]-co_list[2]), self.pole_appendix_ypos).normalized()
        
        verts.append(co_0_base)
        verts.append(co_0_base - tangent * 0.1)
        verts.append(co_1_base)
        verts.append(co_1_base + tangent * 0.1)

        return verts;
                                     
    def make_carstop(self, context, object, data, material):

        new_objects = []

        curveData = self.make_curve('carstop')
        self.adjust_resolution_u(curveData)
        
        bm = bmesh.from_edit_mesh(data)

        footer_co_list = []
        footer_tilt_list = []

        vertices_for_ring = []

        for f in bm.faces:
            if self.is_face_valid(f) == False:
                continue

            v_list = self.get_vlist(f)
            co_list = self.get_co_list(v_list)

            normal = f.normal
            tilt = atan2(normal.y, normal.x)

            if self.carstop_type == 'TYPEA':
                self.make_carstop_A(co_list, curveData, footer_co_list, footer_tilt_list, tilt)

            if self.carstop_ring == True:
                vertices_for_ring += self.set_carstop_ring_verts(co_list)#, curveData)

        self.set_bevel_depth(curveData)
        rectOB = self.proc_bevel_type(curveData, self.fence_bevel_type, context.scene)
        curveOB = self.make_curve_object('carstop', curveData, material, context)
        new_objects.append(curveOB)
        
        self.adjust_resolution_u(curveData)

        ring_name = {"TYPEA":"PoleConnectorRing.A",
                     "TYPEB":"PoleConnectorRing.B"}[self.carstop_ring_type]
        
        if vertices_for_ring != []:
            radius = self.bevel_depth_pole
            radius_ratio = self.carstop_ring_radius_ratio
            
            partsOB, ringOB, mod_dist = self.make_instancer_curve(context, vertices_for_ring, ring_name, "carstop_connector.Ring", self.pole_appendix_size, 8, self.bevel_depth_pole * radius_ratio)
            
            self.modifier_input(mod_dist, "Translation", Vector((radius*(1+radius_ratio), 0, 0)))

            new_objects.append(ringOB)

        if (self.carstop_base_type != "NONE" and footer_co_list != []):

            footerOB = self.make_header_footer_instance(context, footer_co_list, footer_tilt_list, material, "carstop.FOOTER", self.pole_header_footer_size)
            new_objects.append(footerOB)

            dataname = {'BLOCK':'Cube',
                        'CYLINDER':'Cylinder',
                        'CUSTOM':None}[self.carstop_base_type]

            instanceeOB = self.make_instanceeOB(context, material, function = None, dataname = dataname)

            if (instanceeOB == None):
                if self.header_footer_custom_target in bpy.data.objects:
                    instanceeOB = bpy.data.objects[self.header_footer_custom_target]

            aspect = self.pole_base_aspect
            mod_dist = self.add_GN_modifier(footerOB, "EdgeDistributor.Taper")
            if (self.pole_footer_taper == True):
                self.modifier_input(mod_dist, "Taper", self.pole_footer_taper_facror)

            
            scale = Vector((1, 1, aspect)) * self.pole_header_footer_size * self.bevel_depth_pole

            self.modifier_input(mod_dist, "Object", instanceeOB)
            self.modifier_input(mod_dist, "Scale", scale)

            #mod_edge = self.add_edge_split_modifier(instanceeOB)
        
        return new_objects
            
    def execute(self, context):

        object = context.object
        data = object.data
        material = self.get_material(object)

        if (self.action == 'FACE'):
            new_objects = self.make_fence(context, object, data, material)

        if (self.action == 'TRUSS'):
            new_objects = self.make_truss(context, object, data, material)

        if (self.action == 'GUARDRAIL'):
            new_objects = self.make_guardrail(context, object, data, material)

        if (self.action == 'CARSTOP'):
            new_objects = self.make_carstop(context, object, data, material)
            
        self.move_objects_to_collection(context, object, new_objects)
        self.set_parenting_(object, new_objects)

        return {'FINISHED'}





class MESH_PT_FenceMaker(bpy.types.Panel):
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Fence Maker"
    bl_category = "Create"
    bl_context = "mesh_edit"

    def draw_header(self, context):
        layout = self.layout
        layout.label(icon="PLUGIN")

    #def pgettext_dummy(self, text):
    #    return text

    def draw(self, context):

        def pgettext_dummy(text):
            return text

        use_translate = context.preferences.view.use_translate_interface
        pgettext = {True:bpy.app.translations.pgettext,
                    False:pgettext_dummy}[use_translate]

        layout = self.layout

        col = layout.column(align = True)
        col.operator(MESH_OT_FenceMaker.bl_idname, text=pgettext("Make Face"), icon="FACESEL").action = 'FACE'
        col.operator(pole_module.MESH_OT_FenceMaker_Pole.bl_idname, text=pgettext("Make Poles"), icon="EDGESEL").action = 'POLE'
        col.operator(pole_module.MESH_OT_FenceMaker_Pole.bl_idname, text=pgettext("Make Pole Appendix"), icon="APPEND_BLEND").action = 'POLE_APPENDIX'

        col.separator()
        
        col.operator(MESH_OT_FenceMaker.bl_idname, text=pgettext("Make Truss"), icon="LINCURVE").action = 'TRUSS'

        col.separator()
        
        col.operator(MESH_OT_FenceMaker.bl_idname, text=pgettext("Make Guard Rail"), icon="PROP_PROJECTED").action = 'GUARDRAIL'
        col.operator(MESH_OT_FenceMaker.bl_idname, text=pgettext("Make Car Stopper"), icon="SPHERECURVE").action = 'CARSTOP'

class FenceProperties(bpy.types.PropertyGroup):

    fence_canvas_type : StringProperty(
        default = "NotDefined")

    bottom_offset : FloatProperty(
        default = 0)


class FenceMakerPreferences(bpy.types.AddonPreferences):
    bl_idname = __name__

    use_collection : BoolProperty(
        name = "use_collection", default = True,
        description = "Link new objects to collection of src objects")

    def draw(self, context):

        pgettext = translate_module.get_pgettext(context)
        layout = self.layout
        row = layout.row()

        row.prop(self, "use_collection")
        
classes = (
    FenceMakerPreferences,
    
    MESH_OT_FenceMaker,
    pole_module.MESH_OT_FenceMaker_Pole,
    MESH_PT_FenceMaker,
    objectmode_module.OBJECT_OT_FenceMaker,
    objectmode_module.OBJECT_PT_FenceMaker,

    FenceProperties,
)

def baseObj_poll(self, object):
    if (object == bpy.context.object):
        return False
    
    return object in list(bpy.context.view_layer.objects)

def register():
    bpy.app.translations.register(__name__, translate_module.translate_dict_FenceMaker)
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls);

        #bpy.types.Scene.fencemaker_custom_target = bpy.props.PointerProperty(name = "Custom Object", description="Target Object for Custom", type = bpy.types.Object, poll = baseObj_poll)
    
    bpy.types.Object.fenceprops = PointerProperty(type = FenceProperties)
    
def unregister():
    bpy.app.translations.unregister(__name__)
    from bpy.utils import unregister_class
    for cls in classes:
        unregister_class(cls)
        
    del bpy.types.Object.fenceprops

    #del bpy.types.Scene.fencemaker_custom_target

if __name__ == "__main__":
    register()
