import bpy
import math

from bpy.props import *
from mathutils import Vector
import numpy as np


CommonTypeSelection = [("TYPEA", "Type A", "Type A", "EVENT_A", 1),
                       ("TYPEB", "Type B", "Type B", "EVENT_B", 2),
                       ("TYPEC", "Type C", "Type C", "EVENT_C", 3),
                       ("TYPED", "Type D", "Type D", "EVENT_D", 4),
                       ("TYPEE", "Type E", "Type E", "EVENT_E", 5)]

class CollectionModule:

    def move_to_collection(self, obj, tgt_collection):

        col_org = obj.users_collection
        for col in col_org:
            col.objects.unlink(obj)
            
        tgt_collection.objects.link(obj)

    def move_objects_to_collection(self, context, object, new_objects):
        
        addon_name = __name__.split(".")[0]
        preferences = context.preferences.addons[addon_name].preferences
        if (preferences.use_collection):
            src_collection = object.users_collection[0]
            for new_obj in new_objects:
                if (new_obj != None):
                    self.move_to_collection(new_obj, src_collection)

class LayoutModule:

    def newboxcol(self, layout):
        box = layout.box()
        col = box.column(align = True)
        return col

class CommonGeometryModule:

    def search_face(self, f, e0, edge_list, face_tag):

        if (face_tag[f.index] == True):
            return 
        face_tag[f.index] = True
        edge_list.append(e0)

        #print(f.index, e0.index)

        is_end = True
        for e in f.edges:
            for f1 in e.link_faces:
                #print("e_index=", e.index, "f1.index=", f1.index)
                if f1 == f:
                    continue
                if f1.select == False:
                    continue
                if (face_tag[f1.index] == True):
                    continue

                is_end = False
                self.search_face(f1, e, edge_list, face_tag)


        #e0 ̔Α
        #print("f=", f.index, "is_end = ", is_end)
        if (is_end == True):
            l_edge = [l.index for l in f.loops if l.edge == e0][0]
            opposite_edge = f.loops[(l_edge+2) % 4].edge
            edge_list.append(opposite_edge)


    def vertX(self, vert):
        return vert.co[0]

    def vertY(self, vert):
        return vert.co[1]

    def vertZ(self, vert):
        return vert.co[2]

    def get_vertical_vlist(self, e):
        list_Z = sorted(e.verts, key = self.vertZ)
        return list_Z;

    def get_edge_ztilt(self, e):
        v_list = self.get_vertical_vlist(e)
        binormal = (v_list[1].co - v_list[0].co).normalized()
        dot = binormal.dot((0,0,1))
        normal = self.get_edge_normal(e)
        if (normal.z > 0):
            return -math.acos(dot)
        return math.acos(dot)

    def adjust_resolution_u(self, curveData):
        if (self.action == 'FACE'):
            if (self.fence_type == 'CROSS'):
                type = self.cross_ornament_type
                if (type != 'None'):
                    curveData.resolution_u = self.resolution_u

        if (self.action == 'GUARDRAIL'):
            curveData.resolution_u = self.guardrail_resolution_u

        if (self.action == 'CARSTOP'):
            curveData.resolution_u = 8

    

class CommonParameterModule:

    pole_header_footer_size : FloatProperty(
        name = 'Size', default = 1,
        min = 0, step = 0.1)

    pole_base_aspect : FloatProperty(
        name = 'Aspect', default = 1, 
        min = 0.1, max = 10, step = 0.3)

    pole_footer_taper : BoolProperty(
        name = "Taper", default = False)
    
    pole_footer_taper_facror : FloatProperty(
        name = "Factor", default = -0.5,
        min = -2, max = 1, step = 0.3 )

    header_footer_custom_target : StringProperty(
        name = "Custom Object",
        description="Target Object for Custom")

class CommonModule:

    resolution : IntProperty(
        name = 'Resolution', default = 2,
        min = 1, max = 5)

    bevel_depth : FloatProperty(
        name = 'Bevel Depth', default = 0.01,
        min = 0.001, step = 0.1,
        precision = 3)

    smooth : BoolProperty(
        name = 'Smooth Shading', default = True)

    use_material : BoolProperty(
        name = "Material", default = False)

    parenting : BoolProperty(
        name = "Parenting", default = True)

    def draw_common(self, layout):

        row = layout.row()
        row.prop(self, "parenting")
        row.prop(self, "smooth")

        row = layout.row()
        row.prop(self, "resolution")
        row.prop(self, "use_material")

    def make_mesh(self, name, vertices, edges, faces):
        
        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(vertices, edges, faces)
        return mesh

    def make_object(self, data, material, context, col_to_belong = None):

        obj =  bpy.data.objects.new(data.name, data)
        
        if (col_to_belong):
            col_to_belong.objects.link(obj)
        else:
            context.scene.collection.objects.link(obj)

        self.set_material(obj, material)
        
        return obj

    def set_parenting_(self, parent_object, object):

        if (self.parenting == True):
            if (type(object) == list):
                for obj in object:
                    obj.parent = parent_object
                    
            else:
                object.parent = parent_object

    def get_material(self, object):
        material = None
        if len(object.material_slots) != 0:
            material = object.material_slots[0].material
        return material

    def set_material(self, object, material):
        if (self.use_material == False):
            return
        if (material == None):
            return
            
        object_org = bpy.context.view_layer.objects.active
        bpy.context.view_layer.objects.active = object
        bpy.ops.object.material_slot_add()
        if (material != None):
            object.data.materials[0] = material
        bpy.context.view_layer.objects.active = object_org

    def rotation_matrix(self, axis, theta):
        axis = np.asarray(axis)
        axis = axis / math.sqrt(np.dot(axis, axis))
        a = math.cos(theta / 2.0)
        b, c, d = -axis * math.sin(theta / 2.0)
        aa, bb, cc, dd = a * a, b * b, c * c, d * d
        bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
        return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                         [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                         [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

    def get_rot_matrix_from_2v(self, tangent, normal):
        binormal = tangent.cross(normal).normalized()
        rot_matrix = np.array([[tangent.x, binormal.x, normal.x],
                               [tangent.y, binormal.y, normal.y],
                               [tangent.z, binormal.z, normal.z]])
        return rot_matrix

    def get_rot_matrix_along_tangent(self, tangent, normal, theta):
        rot_matrixA = np.array([[tangent.x, 0, normal.x],
                                [tangent.y, 0, normal.y],
                                [tangent.z, 1, normal.z]])
        rot_matrixB = self.rotation_matrix(tangent, theta)
        return np.dot(rot_matrixB, rot_matrixA)

    
