
import bpy
from bpy.props import * 

import numpy as np

from os import path
from mathutils import Vector, Matrix
from math import sin, cos, sqrt, atan2, isclose

class CurveCapModule:

    cap_radius_A : FloatProperty(
        name = "Radius A",
        min = 1,
        max = 5,
        default = 1.5)

    cap_radius_B : FloatProperty(
        name = "Radius B",
        min = 0.1,
        max = 5,
        default = 1.5)

    def adjust_poly_tilt(self, polyline, unit_z, normal, offset = 0):
        #print(offset)
        p0 = polyline.points[0]
        if (polyline.use_cyclic_u == False):
            p1 = polyline.points[1]
            tmp = (p1.co - p0.co); tmp.resize_3d()
            tilt = self.get_tilt(tmp.normalized(), unit_z, normal)
        else:
            p1 = polyline.points[-1]
            tmp = (p1.co - p0.co); tmp.resize_3d()
            tilt = self.get_tilt(tmp.normalized(), unit_z, normal)
            
        for p in polyline.points:
            p.tilt = tilt + offset

    def append_round_top(self, polyline, v_unit, pos, tilt, size, radius = 1.0):
        co_0 = pos.copy()
        co_0.resize_3d()
        t = 3.14159/2

        num = 10
        if (size < 1):
            num = 5

        points = polyline.points
        co = co_0 + v_unit * sin(t*0.02) * size * 0.2
        self.append_poly_from3d(points, co, tilt, cos(t*0.00)*radius)

        for i in range(1, num):
            f = i / num
            co = co_0 + v_unit * sin(t*f) * size * 0.2
            self.append_poly_from3d(points, co, tilt, cos(t*f)*radius)

        co = co_0 + v_unit * sin(t*0.95) * size * 0.2
        self.append_poly_from3d(points, co, tilt, cos(t*0.95)*radius)

    def make_cap(self, curveData, center, ornament_size, radius, x_unit, y_unit, normal, pos, type):

        if (pos == 'BOTTOM'):
            y_unit = -y_unit

        polyline = curveData.splines.new('POLY')
        points = polyline.points
        line_list = [polyline]
        unit_z = Vector((0,0,1))
        
        if (type == 'SIMPLE' or type == 'ROUND' or type == 'HAT'):
            points.add(1)
            tmp = center-y_unit*(ornament_size-0.00)*0.1; tmp.resize_4d()
            points[0].co = tmp
            tmp = center+y_unit*(ornament_size-0.00)*0.9; tmp.resize_4d()
            points[1].co = tmp

            points[0].radius = self.cap_radius_A
            points[1].radius = self.cap_radius_A

            if (type == 'ROUND'):
                self.append_round_top(polyline, y_unit, tmp, 0, radius * self.cap_radius_A * self.cap_radius_B, radius = self.cap_radius_A)

            if (type == 'HAT'):
                polyline2 = curveData.splines.new('POLY')
                points = polyline2.points
                points[0].co = tmp - ornament_size * self.cap_radius_A * self.cap_radius_B * 0.05 * Vector((y_unit.x,y_unit.y, y_unit.z,0))
                points[0].radius = self.cap_radius_A * self.cap_radius_B
                self.append_round_top(polyline2, y_unit, tmp, 0, radius * self.cap_radius_A * self.cap_radius_B * 5, radius = self.cap_radius_A*self.cap_radius_B)

                unit_z = Vector((0,0,1))
                self.adjust_poly_tilt(polyline2, unit_z, normal)

                line_list.append(polyline2)
                
        if (type == 'TAPER'):
            points.add(5)

            tmp = center-y_unit*(ornament_size-0.00)*0.1; tmp.resize_4d()
            points[0].co = tmp
            tmp = center-y_unit*(ornament_size-0.005)*0.1; tmp.resize_4d()
            points[1].co = tmp
            tmp = center-y_unit*(ornament_size-0.01)*0.1; tmp.resize_4d()
            points[2].co = tmp
            tmp = center+y_unit*(ornament_size-0.01)*0.9; tmp.resize_4d()
            points[3].co = tmp
            tmp = center+y_unit*(ornament_size-0.005)*0.9; tmp.resize_4d()
            points[4].co = tmp
            tmp = center+y_unit*(ornament_size-0.00)*0.9; tmp.resize_4d()
            points[5].co = tmp

            points[0].radius = 1.0
            points[1].radius = self.cap_radius_A
            points[2].radius = self.cap_radius_A
            points[3].radius = self.cap_radius_B * self.cap_radius_A
            points[4].radius = self.cap_radius_B * self.cap_radius_A
            points[5].radius = self.cap_radius_B

            #if (pos == 'BOTTOM'):
            #    points[0].radius = self.cap_radius_B
            #    points[1].radius = self.cap_radius_B * self.cap_radius_A
            #    points[2].radius = self.cap_radius_B * self.cap_radius_A
            #    points[3].radius = self.cap_radius_A
            #    points[4].radius = self.cap_radius_A
            #    points[5].radius = 1.0

        # őŜtilt͂̂łł45x
        self.adjust_poly_tilt(polyline, unit_z, normal)

        return line_list


class CurveModule:

    def make_curve(self, name):

        curveData = bpy.data.curves.new(name, type='CURVE')
        curveData.dimensions = '3D'
        curveData.resolution_u = 2
        curveData.use_fill_caps = True
        curveData.bevel_resolution = self.resolution
        
        return curveData

    def make_curve_object(self, name, data, material, context):
        curveOB = bpy.data.objects.new(name, data)
        context.scene.collection.objects.link(curveOB)
        self.set_material(curveOB, material)
        return curveOB

    #def set_transform(self, curveOB, object):
    #    if (self.copy_transform == True):
    #        curveOB.location = object.location
    #        curveOB.rotation_euler = object.rotation_euler
    #        curveOB.rotation_quaternion = object.rotation_quaternion
    #        curveOB.rotation_axis_angle = object.rotation_axis_angle
    #        curveOB.rotation_mode = object.rotation_mode
    #        curveOB.scale = object.scale
   

    def get_tilt(self, unit, unit_z, normal):

        if (isclose(unit.dot(Vector((1,0,0))), 0, abs_tol=1e-6)) and (isclose(unit.dot(Vector((0,1,0))), 0, abs_tol=1e-6)):
            tilt = atan2(normal.y, normal.x)
        else:
            unit_a = unit.cross(unit_z).normalized()
            unit_b = unit_a.cross(unit)
            tilt = - atan2(normal.dot(unit_b), normal.dot(unit_a))
        return tilt
        
    def set_poly_resample_from3d(self, curveData, co_0, co_1, tilt, wave_length, radius = 1):
        length = (co_1-co_0).length
        num = max(1, int(length / wave_length))

        polyline = curveData.splines.new('POLY')
        points = polyline.points
        points.add(num)

        for i in range(0, num+1):
            self.set_point_from3d(points[i], (num-i)/num * co_0 + i/num * co_1)
            points[i].tilt = tilt
            points[i].radius = radius
        
        return polyline

    def set_poly_from3d(self, curveData, co_0, co_1, tilt, radius = 1):
        polyline = curveData.splines.new('POLY')
        points = polyline.points
        points.add(1)
        points[0].tilt = tilt
        points[1].tilt = tilt
        points[0].radius = radius
        points[1].radius = radius
        self.set_point_from3d(points[0], co_0)
        self.set_point_from3d(points[1], co_1)
        return polyline

    def append_poly_from3d(self, points, co, tilt = 0, radius = 1):
        points.add(1)
        self.set_point_from3d(points[-1], co)
        points[-1].tilt = tilt
        points[-1].radius = radius
        

    def set_point_from3d(self, point, co):
        co_tmp = co.copy();
        co_tmp.resize_4d(); point.co = co_tmp

    def set_point_from_list3d(self, points, co_list):
        for p,co in zip(points, co_list):
            self.set_point_from3d(p, co)    

    def copy_point_from(self, p0, p1):
        p0.co = p1.co
        p0.radius = p1.radius
        p0.tilt = p1.tilt

    def copy_bezier_point_from(self, p0, p1, handle_free = True):
        self.copy_point_from(p0, p1)

        if (handle_free == True):
            p0.handle_left_type = 'FREE'
            p0.handle_right_type = 'FREE'
        else:
            p0.handle_left_type = p1.handle_left_type
            p0.handle_right_type = p1.handle_right_type
            
        p0.handle_left = p1.handle_left
        p0.handle_right = p1.handle_right


    def search_most_straight(self, s0, s_list):

        v0 = (s0.points[-1].co - s0.points[-2].co).normalized()
        dot_list = []
        for s1 in s_list:
            if (s0.points[-1].co == s1.points[0].co):
                v1 = (s1.points[1].co - s1.points[0].co).normalized()
            if (s1.points[-1].co == s1.points[-1].co):
                v1 = (s1.points[-2].co - s1.points[-1].co).normalized()
                    
            dot_list.append(v0.dot(v1))
            
        zipped = zip(dot_list, s_list)
        sorted_list = sorted(zipped, key = lambda x:x[0])
        
        z_list = list(sorted_list)
        return z_list[-1]

    def copy_points_from(self, p0, p1):
        p0.co = p1.co
        p0.radius = p1.radius
        p0.tilt = p1.tilt
        
    def invert_poly(self, points):
        num = len(points)
        co_list = []
        radius_list = []
        tilt_list = []
        for i in range(num-1, -1, -1):
            co_list    .append(points[i].co.copy())
            radius_list.append(points[i].radius)
            tilt_list  .append(points[i].tilt)
            
        for i in range(0, num):
            points[i].co     = co_list[i]
            points[i].radius = radius_list[i]
            points[i].tilt   = -tilt_list[i]

    def connect_poly(self, points0, points1):
        num0 = len(points0)
        num1 = len(points1)
        tilt_base = points0[-1].tilt
        points0.add(num1-1)
        for i in range(1, num1):
            self.copy_points_from(points0[num0+i-1], points1[i])
            while( points0[num0+i-1].tilt > tilt_base + 3*3.14159/4):
                points0[num0+i-1].tilt -= + 3.14159
                
            while( points0[num0+i-1].tilt < tilt_base - 3*3.14159/4):
                points0[num0+i-1].tilt += + 3.14159
                
        # ]̉\邩痼0ɂĂ
        points1[0].radius = 0
        points1[-1].radius = 0
    
    def merge_poly_unit(self, curveData):
        proc = False
        splines = curveData.splines
        for s0 in splines:
            candid = []
            for s1 in splines:
                if (s0 == s1): 
                    continue
                
                if (s0.points[0].radius == 0 or s1.points[0].radius == 0):
                    continue
                        
                if (s0.points[-1].co == s1.points[0].co):
                    candid.append(s1)

                if (s0.points[-1].co == s1.points[-1].co):
                    candid.append(s1)
                    
            if (candid != []):
                dot_value, tgt = self.search_most_straight(s0, candid)            
                if (dot_value > self.pole_merge_threshold):
                    if (s0.points[-1].co == tgt.points[-1].co):
                        self.invert_poly(tgt.points)
                        
                    self.connect_poly(s0.points, tgt.points)
                    proc = True

        # serch inverted
        for s0 in splines:
            candid = []
            for s1 in splines:
                if (s0 == s1): 
                    continue
                if (s0.points[0].radius == 0 or s1.points[0].radius == 0):
                    continue
                        
                if (s0.points[0].co == s1.points[0].co):
                    candid.append(s1)

                if (s0.points[0].co == s1.points[-1].co):
                    candid.append(s1)

            if (candid != []):
                self.invert_poly(s0.points)
                dot_value, tgt = self.search_most_straight(s0, candid)            
                if (dot_value > self.pole_merge_threshold):
                    if (s0.points[-1].co == tgt.points[-1].co):
                        self.invert_poly(tgt.points)
                        
                    self.connect_poly(s0.points, tgt.points)
                    proc = True

        for s0 in splines:
            if s0.points[0].co == s0.points[-1].co:
                v0 = (s0.points[-1].co - s0.points[-2].co).normalized()
                v1 = (s0.points[1].co - s0.points[0].co).normalized()
                if (v0.dot(v1) > self.pole_merge_threshold):
                    s0.use_cyclic_u = True
                    
        for s in splines:
            if s.points[0].radius == 0:
                splines.remove(s)
                
        return proc

    def merge_poly(self, curveData):
        count = 0
        while(True):
            if (self.merge_poly_unit(curveData) == False):
                break
                count += 1
                if (count > 100):
                    break

class InstancerModule:

    def set_instancer_verts(self, rot_matrix, co):

        verts = []

        verts.append(co)
        verts.append(co + Vector(np.dot(rot_matrix, (0, 0, 1))) * 0.1)
        
        return verts

    def make_instancer_object(self, context, vertices, name):

        if (vertices == []):
            return None
        
        edges = []
        for index in range(0, len(vertices), 2):
            edges.append([index, index+1])

        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(vertices, edges, [])

        #bpy.ops.object.mode_set(mode="OBJECT")
        #bpy.ops.object.mode_set(mode="EDIT")

        new_obj = self.make_object(mesh, None, context)

        return new_obj


    def make_instancer_vis(self, context, vertices_for_vis, name_instancee, name_instancer, size):
        
        col = self.get_asset_colelction(context, "FenceParts")

        partsOB = self.load_object(context, name_instancee, col)
        visOB = self.make_instancer_object(context, vertices_for_vis, name_instancer)
        mod_dist = self.add_GN_modifier(visOB, "EdgeDistributor.Vis")

        self.modifier_attrib(mod_dist, "NUT", "NUT")
        self.modifier_attrib(mod_dist, "EXTEND", "EXTEND")

        self.modifier_input(mod_dist, "Object", partsOB)
        self.modifier_input(mod_dist, "Scale", size)
        
        return partsOB, visOB, mod_dist

    def make_instancer_curve(self, context, vertices, name_instancee, name_instancer, size, resolution, radius):

        col = self.get_asset_colelction(context, "FenceParts")

        partsOB = self.load_object(context, name_instancee, col, use_curve=True)
        instancerOB = self.make_instancer_object(context, vertices, name_instancer)
        mod_dist = self.add_GN_modifier(instancerOB, "EdgeDistributor.Curve")
        
        self.modifier_input(mod_dist, "Object", partsOB)
        self.modifier_input(mod_dist, "Scale", size)
        self.modifier_input(mod_dist, "Resolutuion", resolution)
        self.modifier_input(mod_dist, "Radius", radius)
        
        return partsOB, instancerOB, mod_dist

    def make_instancer_basic(self, context, vertices, name_instancee, name_instancer, size, use_curve = False):

        col = self.get_asset_colelction(context, "FenceParts")

        partsOB = self.load_object(context, name_instancee, col, use_curve)
        instancerOB = self.make_instancer_object(context, vertices, name_instancer)
        mod_dist = self.add_GN_modifier(instancerOB, "EdgeDistributor.Basic")
        
        self.modifier_input(mod_dist, "Object", partsOB)
        self.modifier_input(mod_dist, "Scale", Vector((1,1,1))*size)
        
        return partsOB, instancerOB, mod_dist

    def make_instanceeOB(self, context, material, function = None, dataname = None):
        col = self.get_asset_colelction(context, "FenceParts")

        instanceeOB = None
        if (function != None):
            mesh_instancee = getattr(self, function)()
            instanceeOB = self.make_object(mesh_instancee, material, context, col)
        if (dataname != None):
            instanceeOB = self.load_object(context, dataname, col)

        return instanceeOB


    def make_header_footer_instance(self, context, co_list, co_tilt, material, name, size):

        vertices = []

        for co,tilt in zip(co_list, co_tilt):

            rot_matrix = Matrix(self.rotation_matrix((0,0,1), tilt))

            vertices.append(co)
            vertices.append(co + rot_matrix @ Vector((0.1,0,0)))

        instancerOB = self.make_instancer_object(context, vertices, name)

        bpy.ops.object.mode_set(mode="OBJECT")
        bpy.ops.object.mode_set(mode="EDIT")

        return instancerOB

class MeshSettingModule:

    def get_asset_colelction(self, context, collection_name):

        if collection_name in bpy.data.collections:
            return bpy.data.collections[collection_name]

        col = bpy.data.collections.new(collection_name)
        context.scene.collection.children.link(col)

        context.view_layer.layer_collection.children[collection_name].hide_viewport = True
        col.hide_render = True

        return col

    def load_mesh(self, meshname):

        if not meshname:
            return

        if (meshname in bpy.data.meshes):
            return

        dirname = path.dirname(__file__)
        libfile = dirname + ("\FenceMakerV1p3.blend")
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.meshes = [meshname]

    def load_curve(self, curvename):

        if (not curvename):
            return

        if (curvename in bpy.data.curves):
            return

        dirname = path.dirname(__file__)
        libfile = dirname + ("\FenceMakerV1p3.blend")
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.curves = [curvename]

    def load_object(self, context, dataname, col_to_belong = None, use_curve = False):

        if (dataname in bpy.data.objects):
            return bpy.data.objects[dataname]
        
        if (dataname == ""):
            print("No Mesh Name")
            return None

        if (use_curve == True):
            self.load_curve(dataname)
            obj = bpy.data.objects.new(dataname, bpy.data.curves[dataname])
        else:
            self.load_mesh(dataname)
            obj = bpy.data.objects.new(dataname, bpy.data.meshes[dataname])

        if (col_to_belong):
            col_to_belong.objects.link(obj)
        else:
            context.scene.collection.objects.link(obj)

        return obj


class MeshModule:
    
    def make_edge_list_list(self, bm):

        fnum = len(bm.faces)
        
        face_tag = [False]*fnum
        edge_list_all = []
        for f in bm.faces:

            if (face_tag[f.index] == True):
                continue
            face_tag[f.index] = True
                    
            if f.select == False:
                continue

            count = 0
            for e in f.edges:
                if (count >= 2):
                    continue
                for f1 in e.link_faces:
                    if f1 == f:
                        continue
                    if f1.select == False:
                        continue
                    
                    edge_list = []
                    self.search_face(f1, e, edge_list, face_tag)
                    edge_list_all.append(edge_list)
                    count += 1;

            #[X^[gꍇ
            if count == 1:
                edge_list = []
                e0 = edge_list_all[-1][0]
                l_edge = [l.index for l in f.loops if l.edge == e0][0]
                opposite_edge = f.loops[(l_edge+2) % 4].edge
                edge_list.append(opposite_edge)
                edge_list_all.append(edge_list)

            #Ɨ
            if count == 0:
                
                def z_directional(e):
                    v = (e.verts[1].co - e.verts[0].co).normalized()
                    return abs(v.dot(Vector((0,0,1))))
                    
                list_e = sorted(f.edges, key = z_directional)

                #print("list_e", list_e)

                for e in list_e:
                    print(e, z_directional(e))

                edge_list = [list_e[-1], list_e[-2]]
                edge_list_all.append(edge_list)
                edge_list_all.append([])

        #for e_list in edge_list_all:
        #    print([e.index for e in e_list])

        #2̔{ɂȂĂ͂
        num = len(edge_list_all)

        edge_list_combined = []
        for i in range(0, num, 2):
            edge_list_all[i].reverse()
            edge_list_combined.append(edge_list_all[i] + edge_list_all[i+1])

        return edge_list_combined

    def get_center_f(self, f):
        center = Vector((0,0,0))
        for v in f.verts:
            center += v.co
        center /= len(f.verts)
        return center

    def get_edge_info(self, e):
        co0 = e.verts[0].co.copy()
        co1 = e.verts[1].co.copy()
        normal = self.get_edge_normal(e)
        return co0, co1, normal

    def get_sorted_by_z(self, e, co0, co1):
        v0 = e.verts[0]; v1 = e.verts[1]
        if (co0.z > co1.z):
            co0, co1 = co1, co0
            v0 = e.verts[1]; v1 = e.verts[0]
        return v0, v1, co0, co1

    def search_edge_cont(self, bm, e, v):
        e_candidates = [e for e in v.link_edges]
        e_candidates.remove(e)
        for l in e.link_loops:
            e_next = l.link_loop_next.edge
            e_candidates.remove(e_next) if e_next in e_candidates else None
            e_prev = l.link_loop_prev.edge
            e_candidates.remove(e_prev) if e_prev in e_candidates else None

        return e_candidates
    
    def adjust_elist_direction(self, e_list):
        
        co_0 = self.get_vertical_vlist(e_list[0])[0].co
        co_1 = self.get_vertical_vlist(e_list[1])[0].co
        binormal = (co_1-co_0).normalized()
        normal = self.get_edge_normal(e_list[0])

        if (binormal.cross(normal).z > 0):
            e_list.reverse()

    def get_co_list_from_edge(self, e, margin, slide):
        v_list = self.get_vertical_vlist(e)
        co_list = [v.co.copy() for v in v_list]
        co_org = [co_list[0], co_list[1]]
        co_list[0] = co_org[0] + (co_org[1]-co_org[0]) * (margin + slide)
        co_list[1] = co_org[1] + (co_org[0]-co_org[1]) * (margin - slide)
        return co_list

    def get_edge_normal(self, edge):
        n = Vector((0,0,0))
        count = 0
        for f in edge.link_faces:
            n += f.normal
            count += 1

        if (count == 0):
            return Vector((1,0,0))

        n /= count
        e_n = (edge.verts[1].co - edge.verts[0].co).normalized()
        n -= n.dot(e_n) * e_n
        
        if (n.length == 0):
            return Vector((1,0,0))

        return n.normalized()

    def midpos(self, co_0, co_1, factor):
        return (1-factor)*co_0 + factor*co_1


