
import bpy

from os import path
from mathutils import Vector

class ModifierModule():

    def load_NodeGroup(self, node_name):

        if node_name in bpy.data.node_groups:
            return bpy.data.node_groups[node_name]

        dirname = path.dirname(__file__)
        libfile = dirname + ("\FenceMakerV1p3.blend")

        #print(libfile)
        
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.node_groups = [node_name]
            
        return bpy.data.node_groups[node_name]

    #def add_geometry_node_modifier(self, object, node_name):
    def add_GN_modifier(self, object, node_name, modifier_name = "GeometryNode"):
        mod = object.modifiers.new(name = modifier_name, type="NODES")
        old_node = mod.node_group
        if (old_node != None):
            bpy.data.node_groups.remove(mod.node_group)
        node = self.load_NodeGroup(node_name)
        mod.node_group = node

        return mod

    def modifier_input(self, mod, label, value):

        for input in mod.node_group.inputs:
            if input.name == label:
                if (type(value) == Vector):
                    mod[input.identifier][0] = value[0]
                    mod[input.identifier][1] = value[1]
                    mod[input.identifier][2] = value[2]
                else:
                    mod[input.identifier] = value
                
    def modifier_attrib(self, mod, label, name):

        for input in mod.node_group.inputs:
            if input.name == label:
                mod[input.identifier + "_use_attribute"] = 1
                mod[input.identifier + "_attribute_name"] = name

    def add_transform_modifier(self, object, translation=(0,0,0), rotation=(0,0,0), scale=(1,1,1)):
        
        mod = object.modifiers.new(name = "GoemetryNodes", type="NODES")
        group = mod.node_group
        group.name = "Fence Maker Connector"
        links = group.links
        nodes = group.nodes
        trans = nodes.new(type="GeometryNodeTransform")

        trans.inputs['Translation'].default_value = translation
        trans.inputs['Rotation'].default_value = rotation
        trans.inputs['Scale'].default_value = scale

        # N̑S폜
        for l in links:
            links.remove(l)
            
        links.new(nodes['Group Input'].outputs[0], trans.inputs[0])
        links.new(trans.outputs[0], nodes['Group Output'].inputs[0])

    def add_taper_modifier(self, object, taper_factor):

        mod = object.modifiers.new(name = "SimpleDeform", type="SIMPLE_DEFORM")
        mod.deform_method = "TAPER"
        mod.deform_axis = "Z"
        mod.factor = taper_factor

        return mod

    def add_edge_split_modifier(self, object):

        mod = object.modifiers.new(name = "EdgeSplit", type="EDGE_SPLIT")

        return mod
