
import bpy
from bpy.props import * 
from mathutils import Vector, Matrix

from . import mesh_module
from . import common_module

class OBJECT_OT_FenceMaker(bpy.types.Operator, mesh_module.MeshModule, common_module.CommonModule):
    bl_idname = "object.add_fence_template"
    bl_label = "Add Fence Template"
    bl_description = "Description"
    bl_options = {'REGISTER', 'UNDO'}

    action : StringProperty(
        name = "action")

    CanvasType = [("FENCE", "Fence", "Fence", "", 1),
                  ("GUARDRAIL", "Guard Rail", "Guard Rail", "", 2),
                  ("GUARD_PIPE", "Guard Pipe", "Guard Pipe", "", 3),
                  ("CARSTOPPER", "Car Stopper", "Car Stopper", "", 4),]
    
    type : EnumProperty(
        name = "type", default = "GUARDRAIL",
        items = CanvasType)

    guardrail_interval : EnumProperty(
        name = "Interval", default = "2.0m",
        items = [("2.0m", "2.0m", "2.0m", "", 1),
                 ("4.0m", "4.0m", "4.0m", "", 2)])

    fence_height : EnumProperty(
        name = "Height", default = "1.2m",
        items = [("0.8m", "0.8m", "0.8m", "", 1),
                 ("0.9m", "0.9m", "0.9m", "", 2),
                 ("1.0m", "1.0m", "1.0m", "", 3),
                 ("1.2m", "1.2m", "1.2m", "", 4),
                 ("1.5m", "1.5m", "1.5m", "", 5),
                 ("1.8m", "1.8m", "1.8m", "", 6)])

    fence_interval : EnumProperty(
        name = "Interval", default = "2.0m",
        items = [("1.0m", "1.0m", "1.0m", "", 1),
                 ("1.5m", "1.5m", "1.5m", "", 2),
                 ("2.0m", "2.0m", "2.0m", "", 3),
                 ("2.5m", "2.5m", "2.5m", "", 4),
                 ("3.0m", "3.0m", "3.0m", "", 5)])

    fence_unit_num : IntProperty(
        name = "Number", default = 1,
        min = 1, max = 20)

    fence_bottom_offset : FloatProperty(
        name = "Offset", default = 0.1,
        min = 0, max = 0.2,
        precision = 3)

    guardpipe_interval : EnumProperty(
        name = "Interval", default = "3.0m",
        items = [("2.5m", "2.5m", "2.5m", "", 1),
                 ("3.0m", "3.0m", "3.0m", "", 2),
                 ("3.5m", "3.5m", "3.5m", "", 3)])

    guardpipe_offset : EnumProperty(
        name = "Offset", default = "25cm",
        items = [("25cm", "25cm", "25cm", "", 1),
                 ("30cm", "30cm", "30cm", "", 2)])

    carstopper_interval : EnumProperty(
        name = "Interval", default = "1.0m",
        items = [("0.7m", "0.7m", "0.7m", "", 1),
                 ("1.0m", "1.0m", "1.0m", "", 2),
                 ("1.5m", "1.5m", "1.5m", "", 3),
                 ("2.0m", "2.0m", "2.0m", "", 4),
                 ("3.0m", "3.0m", "3.0m", "", 5),]
    )

    carstopper_height : EnumProperty(
        name = "Height", default = "0.65m",
        items = [("0.65m", "0.65m", "0.65m", "", 1),
                 ("0.80m", "0.80m", "0.80m", "", 2),]
    )
                 
    show_wire : BoolProperty(
        name = "Show Wire", default = True)

    set_active : BoolProperty(
        name = "Set Active", default = True,
        description = "Set new canvas to active")

    face_flip : BoolProperty(
        name = "Flip", default = False)

    convert_threshold : FloatProperty(
        name = "Threthold", default = 0.2,
        description = "Merge threshold to remove too small segment",
        min = 0, max = 2)

    use_cursor : BoolProperty(
        name = "Use Cursor", default = False)

    def draw(self, context):
        layout = self.layout

        if (self.action == "CLEAR"):
            return
        
        col = layout.column(align = True)
        col.prop(self, "type")

        if (self.type == "FENCE"):
            col.prop(self, "fence_height")
            col.prop(self, "fence_interval")
            col.prop(self, "fence_bottom_offset")
            
        if (self.type == "GUARDRAIL"):
            col.prop(self, "guardrail_interval")
            
        if (self.type == "GUARD_PIPE"):
            col.prop(self, "guardpipe_interval")
            col.prop(self, "guardpipe_offset")

        if (self.type == "CARSTOPPER"):
            col.prop(self, "carstopper_interval")
            col.prop(self, "carstopper_height")

        if (self.action == "SET"):
            col.prop(self, "fence_unit_num")
            
        if (self.action == "CONVERT"):
            col.prop(self, "convert_threshold")

        if (self.action == "SET"):
            col.prop(self, "use_cursor")

        row = col.row()
        row.prop(self, "show_wire")
        if (self.action == "CONVERT"):
            row.prop(self, "set_active")
            
        row = col.row()
        row.prop(self, "face_flip")
        row.prop(context.space_data.overlay, "show_face_orientation")

    def set_object_status(self, context, new_obj):
        
        bpy.context.view_layer.objects.active = new_obj
        bpy.ops.object.select_all(action='DESELECT')
        new_obj.select_set(True)
        if (self.show_wire == True):
            new_obj.show_wire = True

        new_obj.fenceprops.fence_canvas_type = self.type
        if (self.type == "GUARD_PIPE"):
            offset = float(self.guardpipe_offset[0:2])*0.01
            new_obj.fenceprops.bottom_offset = offset
        else:
            new_obj.fenceprops.bottom_offset = self.fence_bottom_offset

        new_obj.color[3] = 0.3


    def convert_annotation(self, context):

        scene = context.scene
        obj_orig = context.object

        if obj_orig and obj_orig.type == 'GPENCIL':
            gpen_src = obj_orig.data
            matrix_world = obj_orig.matrix_world
        else:
            if (scene.grease_pencil == None):
                return {'FINISHED'}
            
            gpen_src = scene.grease_pencil
            matrix_world = Matrix()

        gpen = bpy.data.objects.new(name = "GPen_tmp", object_data = gpen_src.copy())
        scene.collection.objects.link(gpen)

        gpen.matrix_world = matrix_world
        bpy.context.view_layer.objects.active = gpen

        if (self.type == 'GUARDRAIL'):
            interval = float(self.guardrail_interval[0:3])
            height = 0.35
            offset = 0.425

        if (self.type == 'FENCE'):
            interval = float(self.fence_interval[0:3])
            height = float(self.fence_height[0:3])
            offset = self.fence_bottom_offset

        if (self.type == 'GUARD_PIPE'):
            interval = float(self.guardpipe_interval[0:3])
            offset = float(self.guardpipe_offset[0:2])*0.01
            height = 0.8 - offset

        if (self.type == 'CARSTOPPER'):
            interval = float(self.carstopper_interval[0:3])
            offset = 0
            height = float(self.carstopper_height[0:4])
            
        bpy.ops.object.mode_set(mode='EDIT_GPENCIL')
        bpy.ops.gpencil.select_all(action='SELECT')
        bpy.ops.gpencil.stroke_sample(length=interval)

        #newobj = bpy.context.view_layer.objects.active

        bpy.ops.object.mode_set(mode='OBJECT')

        pre_object_list = [o for o in bpy.data.objects]

        bpy.ops.gpencil.convert(type='POLY')

        post_object_list = [o for o in bpy.data.objects]
        diff = set(pre_object_list) ^ set(post_object_list)

        new_obj = list(diff)[0]

        new_obj.name = "fence_canvas"
        self.set_object_status(context, new_obj)

        bpy.data.objects.remove(gpen)
        bpy.ops.object.convert(target="MESH")
        bpy.ops.object.mode_set(mode='EDIT')
        
        context.scene.tool_settings.mesh_select_mode = (True, False, False)
        bpy.ops.mesh.select_all(action='SELECT')

        # curvêɎ͂łǂ 
        bpy.ops.mesh.remove_doubles(threshold = self.convert_threshold)
        
        vec = Vector((0,0,1))

        bpy.ops.mesh.extrude_region_move(MESH_OT_extrude_region={"use_normal_flip":False, "use_dissolve_ortho_edges":False, "mirror":False}, TRANSFORM_OT_translate={"value":vec*height, "orient_type":'GLOBAL', "orient_matrix":((1, 0, 0), (0, 1, 0), (0, 0, 1)), "orient_matrix_type":'GLOBAL'})

        bpy.ops.mesh.select_all(action='SELECT')

        if (self.face_flip == True):
            bpy.ops.mesh.flip_normals()
        
        bpy.ops.transform.translate(value=vec*offset, orient_type='GLOBAL', orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), orient_matrix_type='GLOBAL')

        bpy.ops.object.mode_set(mode='OBJECT')

        if (self.set_active == False):

            bpy.context.view_layer.objects.active = obj_orig
            if (obj_orig):
                obj_orig.select_set(True)
        
            new_obj.select_set(False)

        return{'FINISHED'}

    def construct_data(self, width, height0, height1):

        unit_num = self.fence_unit_num
        pole_num = unit_num + 1

        vertices = []
        for i in range(0, pole_num):
            lx = width * (-0.5*unit_num + i)
            vertices.append((lx, 0, height0))
            vertices.append((lx, 0, height1))

        faces = []
        for i in range(0, unit_num):
            faces.append([i*2+0, i*2+2, i*2+3, i*2+1])

        return vertices, faces

    def exe_clear_annotation(self, context):

        layers = context.scene.grease_pencil.layers
        for layer in layers:
            layers.remove(layer)

    def execute(self, context):

        if (self.action == "CLEAR"):
            self.exe_clear_annotation(context)
            return {'FINISHED'}

        if (self.action == 'CONVERT'):
            return self.convert_annotation(context)

        if (self.type == "FENCE"):
            width  = float(self.fence_interval[0:3])
            offset = self.fence_bottom_offset
            bottom = offset
            top    = offset + float(self.fence_height[0:3])

        if (self.type == "GUARDRAIL"):
            width = float(self.guardrail_interval[0:3])
            bottom = 0.6 - 0.35/2
            top    = 0.6 + 0.35/2

        if (self.type == "GUARD_PIPE"):
            width = float(self.guardpipe_interval[0:3])
            bottom = float(self.guardpipe_offset[0:2])*0.01
            top    = 0.8

        if (self.type == "CARSTOPPER"):
            width = float(self.carstopper_interval[0:3]) + 0.2
            bottom = 0
            top    = float(self.carstopper_height[0:4])

        vertices, faces = self.construct_data(width, bottom, top)
            
        mesh = self.make_mesh('fence_canvas', vertices, [], faces)
        new_obj = self.make_object(mesh, None, context)

        if (self.use_cursor == True):
            new_obj.location = context.scene.cursor.location

        self.set_object_status(context, new_obj)

        if (self.face_flip == True):
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.flip_normals()
            bpy.ops.object.mode_set(mode='OBJECT')
       
        return {'FINISHED'}
    

class OBJECT_PT_FenceMaker(bpy.types.Panel):
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Fence Maker"
    bl_category = "Create"
    bl_context = "objectmode"
    
    def draw_header(self, context):
        layout = self.layout
        layout.label(icon="PLUGIN")

    def draw(self, context):
        
        def pgettext_dummy(text):
            return text

        use_translate = context.preferences.view.use_translate_interface
        pgettext = {True:bpy.app.translations.pgettext,
                    False:pgettext_dummy}[use_translate]

        layout = self.layout

        col = layout.column(align = True)
        col.operator(OBJECT_OT_FenceMaker.bl_idname, text=pgettext("Set Template"), icon="FACESEL").action = 'SET'
        col.separator()
        col.operator(OBJECT_OT_FenceMaker.bl_idname, text=pgettext("Convert from Annotation"), icon="GREASEPENCIL").action = 'CONVERT'
        col.operator(OBJECT_OT_FenceMaker.bl_idname, text=pgettext("Clear Annotation"), icon="REMOVE").action = "CLEAR"
        
