
import bpy
from bpy.props import * 

from . import mesh_module
from . import common_module
from . import modifier_module

import time
import numpy as np
import bmesh
import mathutils
from mathutils import Vector, Matrix
from math import sin, cos, sqrt, atan2, ceil

class MESH_OT_FenceMaker_Pole \
(bpy.types.Operator, \
 mesh_module.MeshModule, mesh_module.CurveModule, mesh_module.CurveCapModule, mesh_module.MeshSettingModule, mesh_module.InstancerModule, \
 common_module.CommonGeometryModule, common_module.CommonModule, common_module.LayoutModule, common_module.CollectionModule, common_module.CommonParameterModule, \
 modifier_module.ModifierModule):

    bl_idname = "fencemaker.make_pole"
    bl_label = "Fence Maker"
    bl_description = "Make Poles for Fence from simple quad meshes." 
    bl_options = {'REGISTER','UNDO'}

    CommonTypeSelection = common_module.CommonTypeSelection

    action : EnumProperty(
        name = "Action Type", default = "POLE",
        items = [("POLE", "Pole", "EDGESEL", 1),
                 ("POLE_APPENDIX", "Pole Appendix", "Pole Appendix", "APPEND_BLEND", 2)] )
    
    BevelType = [("CYLINDER", "Cylinder", "Cylinder", "MESH_CYLINDER", 1),
                 ("PRISM", "Prism", "Prism", "MESH_CUBE", 2),
                 ("RECT", "Rect Object", "Rect Object", "MATPLANE", 3),
                 ("LSHAPE", "L-Shape Object", "L-Shape Object", "DRIVER_ROTATIONAL_DIFFERENCE", 4)]

    fill_top : BoolProperty(
        name = "Fill Top", default = True)

    fill_bottom : BoolProperty(
        name = "Fill Bottom", default = False)

    fence_support_num : IntProperty(
        name = "Number", default = 1,
        min = 0, max = 5)

    pole_bevel_type : EnumProperty(
        name = "Type", default = "CYLINDER",
        items = BevelType)

    bevel_depth_pole : FloatProperty( 
        name = 'Bevel Depth', default = 0.02,
        min = 0, step = 0.1, precision = 4)

    pole_solidify : BoolProperty(
        name = 'Solidify', default = True)

    pole_solidify_thickness : FloatProperty(
        name = 'Thickness', default = 0.005,
        min = 0, max = 0.1, step = 0.1,
        precision = 3)
    
    tilt_pole : BoolProperty(
        name = 'Tilt', default = False)

    pole_distort : FloatProperty(
        name = 'Distortion', default = 0,
        min = 0, max = 2,
        step = 0.3)

    pole_distort_freq : FloatProperty(
        name = 'Frequency', default = 0.3,
        min = 0.1, max = 1,
        step = 0.3)

    pole_not_vertical : BoolProperty(
        name = "Not Vertical", default = False)

    pole_vertical_threshold : FloatProperty(
        name = "Threshold", default = 1e-5, 
        min = 1e-5, max = 1, step = 0.3,
        precision = 5)

    pole_vertical : BoolProperty(
        name = "Vertical", default = True)

    extend_pole_top : FloatProperty(
        name = 'Extend Top', default = 0,
        step = 0.1,
        precision = 4)

    extend_pole_bottom : FloatProperty(
        name = 'Extend Bottom', default = 0,
        step = 0.1,
        precision = 4)

    extend_pole_rim : FloatProperty(
        name = 'Extend Rim', default = 0,
        min = 0, step = 0.1, precision = 4)

    pole_vis : BoolProperty(
        name = "Vis", default = False)

    pole_vis_size : FloatProperty(
        name = "Size", default = 1,
        min = 0, max = 3,
        step = 0.3)

    pole_vis_offset_A : FloatProperty(
        name = "Offset A", default = 0,
        min = -3, max = 3,
        step = 0.3)

    pole_vis_offset_B : FloatProperty(
        name = "Offset B", default = 0,
        min = -3, max = 3,
        step = 0.3)

    pole_radius_ratio_horizontal : FloatProperty(
        name = "Radius", default = 1,
        min = 0.5, max = 2,
        precision = 3)

    pole_offset_horizontal : FloatProperty(
        name = "Offset", default = 0,
        min = -3, max = 3)
            
    pole_top_cap : EnumProperty( 
        name = "Cap Type", default = "NONE",
        items = [("NONE", "None", "None", "", 1),
                 ("ROUND", "Round", "Round", "SPHERECURVE", 2),
                 ("CAP", "Cap", "Cap", "MESH_CYLINDER",3)] )

    pole_cap_size : FloatProperty(
        name = "Cap Size", default = 1,
        min = 0.1)

    CapType = [("SIMPLE", "Simple", "Simple", "MESH_CYLINDER", 1),
               ("TAPER", "Taper", "Taper", "MESH_CONE", 2),
               ("ROUND", "Round", "Round", "SPHERECURVE", 3),
               ("HAT", "Hat", "Hat", "MOD_SHRINKWRAP", 4)]
        
    pole_cap_type : EnumProperty(
        name = 'Cap Type', default = 'SIMPLE',
        items = CapType)

    pole_footer_type : EnumProperty(
        name = 'Footer', default = 'BLOCK',
        items = [("BLOCK", "Block", "Block", "MESH_CUBE", 1),
                 ("CYLINDER", "Cylinder", "Cylinder", "MESH_CYLINDER", 2),
                 ("CUSTOM", "Custom", "Custom", "MESH_CUBE", 3)])

    pole_header_type : EnumProperty(
        name = 'Header', default = 'REFLECTOR',
        items = [("REFLECTOR", "Reflector", "Reflector", "", 1),
                 ("CUSTOM", "Custom", "Custom", "", 2)])
   
    #pole_header_footer_size : FloatProperty(
    #    name = 'Size', default = 1,
    #    min = 0, step = 0.1)

    #pole_base_aspect : FloatProperty(
    #    name = 'Aspect', default = 1, 
    #    min = 0.1, max = 10, step = 0.3)

    pole_merge : BoolProperty(
        name = "Merge", default = True)

    pole_merge_threshold : FloatProperty(
        name = "Threshold", default = 0.9, 
        min = -1, max = 1, step = 0.3)

    pole_support_size : FloatProperty(
        name = "Size", default = 0.5,
        min = 0, step = 0.3)

    pole_support_angle : FloatProperty(
        name = "Angle", default = 0.2,
        min = 0, max = 0.5, step = 0.3)

    pole_support_offset : FloatProperty(
        name = "Offset", default = 0.2,
        min = 0, max = 0.5, step = 0.3)

    pole_support_extend : FloatProperty(
        name = "Extend", default = 0.1,
        min = 0, max = 0.5, step = 0.1)

    PoleAppendixType = [("HEADER", "Header", "Header"),
                        ("FOOTER", "Footer", "Footer"),
                        ("CONNECTOR", "Connector", "Connector"),
                        ("SUPPORT", "Support", "Support")]

    pole_appendix_type : EnumProperty(
        name = "Type", default = "HEADER",
        items = PoleAppendixType)

    pole_connector_type : EnumProperty(
        name = "Type",
        default = "CLIPFENCE",
        items = [("CLIPFENCE", "Clip Fence", "Clip Fence", "APPEND_BLEND", 1),
                 ("RING_A", "Ring A", "Ring A", "MESH_CIRCLE", 2),
                 ("RING_B", "Ring B", "Ring B", "MESH_CIRCLE", 3)])
    
    pole_appendix_connector_extend : FloatProperty(
        name = "Extend", default = 0,
        min = -2, max = 2, step = 0.3)

    pole_appendix_connector_scaleB : FloatProperty(
        name = "ScaleB", default = 1,
        min = 0.5, max = 2, step = 0.1)

    pole_appendix_connector_offset : FloatProperty(
        name = "Offset Scale", default = 1,
        min = 0, step = 0.3)
    
    pole_appendix_connector_for_ring : BoolProperty(
        name = "Connector for Ring", default = False)

    pole_appendix_connector_extend_for_ring : FloatProperty(
        name = "Extend", default = 0,
        min = -2, max = 2, step = 0.3)

    pole_appendix_size : FloatProperty(
        name = "Size", default = 1,
        min = 0, max = 5, step = 0.1)

    pole_appendix_ypos : FloatProperty(
        name = "Position", default = 0.5,
        min = 0, max = 1.2, step = 0.3)

    pole_appendix_resolution_u : IntProperty(
        name = "Resolution U", default = 4,
        min = 1, max = 12)
    
    pole_reflector_type : EnumProperty(
        name = "Reflector",
        default = "TYPEA",
        items = CommonTypeSelection[0:3])

    interval_mesh_vertical : FloatProperty(
        name = 'Interval', default = 0.075,
        min = 0.01, step = 0.1, precision = 3)

    def draw(self, context):

        layout = self.layout

        layout.prop(self, "action");
        layout.separator()

        if (self.action == 'POLE'):
            layout.prop(self, "pole_bevel_type")
            if (self.pole_bevel_type == 'RECT'):
                layout.prop(self, "object_aspect")
            layout.prop(self, "bevel_depth_pole")
            if (self.pole_bevel_type == 'LSHAPE'):
                row = layout.row()
                row.prop(self, "pole_solidify")
                if (self.pole_solidify == True):
                    row.prop(self, "pole_solidify_thickness")
            else:
                layout.prop(self, "tilt_pole")
                row = layout.row()
                row.prop(self, "pole_distort")
                row.prop(self, "pole_distort_freq")
            box = layout.box()
            col = box.column(align = True)
            row = col.row()
            row.prop(self, "pole_not_vertical")
            row.prop(self, "pole_vertical_threshold")
            if (self.pole_not_vertical == True):
                if (self.pole_bevel_type != 'LSHAPE'):
                    row = col.row()
                    row.prop(self, "pole_radius_ratio_horizontal")
                    row.prop(self, "pole_offset_horizontal")
                col.prop(self, "extend_pole_rim")
            
            box = layout.box()
            col = box.column(align = True)
            #col2 = col.column(align = True)
            col.prop(self, "pole_vertical")
            col.prop(self, "extend_pole_top")
            col.prop(self, "extend_pole_bottom")
            
            if (self.pole_bevel_type == 'LSHAPE'):
                box = layout.box()
                col = box.column(align = True)
                row = col.row()
                row.prop(self, "pole_vis")
                if (self.pole_vis == True):
                    row.prop(self, "pole_vis_size")
                    row = col.row()
                    row.prop(self, "pole_vis_offset_A")
                    row.prop(self, "pole_vis_offset_B")
                    col.prop(self, "fence_support_num")
            else:
                col.prop(self, "pole_top_cap")
                if (self.pole_top_cap == 'ROUND'):
                    col.prop(self, "pole_cap_size")
                if (self.pole_top_cap == 'CAP'):
                    col.prop(self, "pole_cap_type")
                    col.prop(self, "pole_cap_size")
                    row = col.row()
                    row.prop(self, "cap_radius_A")
                    cap_type = self.pole_cap_type
                    if (cap_type == 'TAPER' or cap_type == 'ROUND' or cap_type == "HAT"):
                        row.prop(self, "cap_radius_B")
            
            #box = layout.box()
            #col = box.column(align = True)
            #col.prop(self, "pole_base_type")
            #if (self.pole_base_type != "NONE"):
            #    row = col.row()
            #    row.prop(self, "pole_header_footer_size")
            #    row.prop(self, "pole_base_aspect")
            #    row = col.row()
            #    row.prop(self, "pole_footer_taper")
            #    if (self.pole_footer_taper == True):
            #        row.prop(self, "pole_footer_taper_facror")
                            
            row = layout.row()
            row.prop(self, "pole_merge")
            row.prop(self, "pole_merge_threshold")

        if (self.action == 'POLE_APPENDIX'):
            col = layout.column(align = True)
            
            col.prop(self, "pole_appendix_type")

            if (self.pole_appendix_type == 'CONNECTOR'):
                col.prop(self, "pole_connector_type")
            if (self.pole_appendix_type == 'SUPPORT'):
                col.prop(self, "pole_bevel_type")
                if (self.pole_bevel_type == 'LSHAPE'):
                    row = col.row()
                    row.prop(self, "pole_solidify")
                    if (self.pole_solidify == True):
                        row.prop(self, "pole_solidify_thickness")
                        
            col.prop(self, "bevel_depth_pole")
            col.prop(self, "pole_vertical_threshold")
            
            if (self.pole_appendix_type in ['HEADER', 'FOOTER']):

                col=self.newboxcol(layout)

                if (self.pole_appendix_type == 'HEADER'):
                    col.prop(self, "extend_pole_top")

                if (self.pole_appendix_type == 'FOOTER'):
                    col.prop(self, "extend_pole_bottom")
                
                col.prop(self, "tilt_pole")

                if (self.pole_appendix_type == 'HEADER'):
                    col.prop(self, "pole_header_type")

                    if (self.pole_header_type == "REFLECTOR"):
                        col.prop(self, "pole_reflector_type")

                    if (self.pole_header_type == "CUSTOM"):
                        col.prop_search(self, "header_footer_custom_target", bpy.data, "objects")

                    col.prop(self, "pole_header_footer_size")

                if (self.pole_appendix_type == "FOOTER"):
                    col.prop(self, "pole_footer_type")

                    if (self.pole_footer_type == "CUSTOM"):
                        col.prop_search(self, "header_footer_custom_target", bpy.data, "objects")
                        col.prop(self, "pole_header_footer_size")
                    else:
                        row = col.row()
                        row.prop(self, "pole_header_footer_size")
                        row.prop(self, "pole_base_aspect")
                        row = col.row()
                        row.prop(self, "pole_footer_taper")
                        if (self.pole_footer_taper == True):
                            row.prop(self, "pole_footer_taper_facror")
                
            if (self.pole_appendix_type == 'CONNECTOR'):
                col = self.newboxcol(layout)
                col.prop(self, "bevel_depth", text="Wire Bevel Depth")
                
                if (self.pole_connector_type == "CLIPFENCE"):
                    col.prop(self, "fence_support_num")
                    row = col.row()
                    row.prop(self, "pole_appendix_connector_extend")
                    row.prop(self, "pole_appendix_connector_scaleB")
                    col.prop(self, "pole_appendix_connector_offset")
                    row = col.row()
                    row.prop(self, "pole_appendix_connector_for_ring")
                    if (self.pole_appendix_connector_for_ring == True):
                        row.prop(self, "ornament_radius")
                        col.prop(self, "pole_appendix_connector_extend_for_ring")
                    row = col.row()
                    row.prop(self, "fill_top")
                    row.prop(self, "fill_bottom")
                
                if ("RING" in self.pole_connector_type):
                    row = col.row()
                    row.prop(self, "pole_appendix_ypos")
                    row.prop(self, "pole_appendix_size")
                    col.prop(self, "pole_appendix_resolution_u")
                    
            if (self.pole_appendix_type == 'SUPPORT'):
                row = layout.row()
                row.prop(self, "pole_support_size")
                row.prop(self, "pole_support_angle")
                row = layout.row()
                row.prop(self, "pole_support_offset")
                row.prop(self, "pole_support_extend")

        self.draw_common(layout)
                
    def invoke(self, context, event):

        obj = context.object

        if (self.action == "POLE"):

            fenceprops = obj.fenceprops
            canvas_type = fenceprops.fence_canvas_type

            if (canvas_type == "GUARDRAIL"):
                self.extend_pole_top = 0      #-0.075
                self.extend_pole_bottom = 0.5 #0.5
                self.bevel_depth_pole = 0.057 #0.057

            if (canvas_type in ["FENCE", "GUARD_PIPE"]):
                self.extend_pole_bottom = fenceprops.bottom_offset

            if (canvas_type == "GUARD_PIPE"):
                self.bevel_depth_pole = 0.114/2
                self.extend_pole_top = 0.02
                self.pole_top_cap = "ROUND"

        return self.execute(context)

    def set_bevel_depth(self, curveData):

        if (self.action == 'POLE'):
            curveData.bevel_depth = self.bevel_depth_pole
        if (self.action == 'POLE_APPENDIX'):
            if (self.pole_appendix_type == 'CONNECTOR'):
                curveData.bevel_depth = self.bevel_depth
            else:
                curveData.bevel_depth = self.bevel_depth_pole

    def proc_bevel_type(self, curveData, bevel_type, sc, action = ''):

        if (action == ''):
            action = self.action

        def set_use_smooth(curveData):
            for s in curveData.splines:
                s.use_smooth = self.smooth

        if (action == 'POLE' or action == 'POLE_SUPPORT'):
            if (bevel_type == 'CYLINDER'):
                set_use_smooth(curveData)

            if (bevel_type == 'PRISM'):
                curveData.bevel_resolution = 0
                for s in curveData.splines:
                    s.use_smooth = False
                    for p in s.points:
                        p.tilt += 3.14159/4

            if (bevel_type == 'RECT'):
                for s in curveData.splines:
                    s.use_smooth = False
                rectOB = self.make_rect_object(self.bevel_depth_pole, self.object_aspect);
                sc.collection.objects.link(rectOB)
                curveData.bevel_mode = 'OBJECT'
                curveData.bevel_object = rectOB
                return rectOB

            if (bevel_type == 'LSHAPE'):
                for s in curveData.splines:
                    s.use_smooth = False
                LshapeOB = self.make_Lshape_object(self.bevel_depth_pole, 1);
                sc.collection.objects.link(LshapeOB)
                curveData.bevel_mode = 'OBJECT'
                curveData.bevel_object = LshapeOB
                curveData.use_fill_caps = False
                return LshapeOB

            if (bevel_type == 'LSHAPE_POLE'):
                for s in curveData.splines:
                    s.use_smooth = False
                LshapeOB = self.make_Lshape_object_pole(self.bevel_depth_pole, 1);
                sc.collection.objects.link(LshapeOB)
                curveData.bevel_mode = 'OBJECT'
                curveData.bevel_object = LshapeOB
                curveData.use_fill_caps = False
                return LshapeOB

    def make_Lshape_object(self, depth, aspect):
        newCurveData = bpy.data.curves.new('Lshape', type='CURVE')
        newCurveData.dimensions = '3D'
        newCurveData.resolution_u = 1
        newCurveData.use_fill_caps = False
        
        polyline = newCurveData.splines.new('POLY')
        polyline.use_cyclic_u = False
        polyline.use_smooth = False
        points = polyline.points
        points.add(2)

        co_list = [Vector((+1, -1, 0, 1)) * depth,
                   Vector((+1, +1, 0, 1)) * depth,
                   Vector((-1, +1, 0, 1)) * depth]
        self.set_point_from_list3d(points, co_list)

        return bpy.data.objects.new(newCurveData.name, newCurveData)

    def make_Lshape_object_pole(self, depth, aspect):
        newCurveData = bpy.data.curves.new('Lshape_pole', type='CURVE')
        newCurveData.dimensions = '3D'
        newCurveData.resolution_u = 1
        newCurveData.use_fill_caps = False
        
        polyline = newCurveData.splines.new('POLY')
        polyline.use_cyclic_u = False
        polyline.use_smooth = False
        points = polyline.points
        points.add(5)
        
        co_list = [Vector((0,   -1.2  , 0, 1)) * depth,
                   Vector((0,   -0.6, 0, 1)) * depth,
                   Vector((0.6, -0.3, 0, 1)) * depth,
                   Vector((0.6, +0.3, 0, 1)) * depth,
                   Vector((0,   +0.6, 0, 1)) * depth,
                   Vector((0,   +1.2  , 0, 1)) * depth ]
        self.set_point_from_list3d(points, co_list)

        return bpy.data.objects.new(newCurveData.name, newCurveData)


        
            
            
    
    def is_pole_vertical(self, co0, co1, threshold = -1):
        if (threshold == -1):
            threshold = self.pole_vertical_threshold
        binormal = (co1-co0).normalized()
        dot = binormal.dot((0,0,1))
        if abs(dot) > (1-threshold):
            return True
        return False

    def get_num_v_co_2(self, co_1, co_0, interval, fit, fit_offset = 0):
        v_dist = (co_1 - co_0).length
        num_vf = v_dist / interval
        num_v = ceil(num_vf)
        if (fit == True):
            num_vf = num_v
            num_v += fit_offset
        return v_dist, num_vf, num_v

    def make_pole_L_rim(self, curveData, edge_list_list):
            
        vertices_for_vis = []

        unit_z = Vector((0,0,1))
        radius = self.bevel_depth_pole

        def get_wire_mesh_info(co_00, co_01):
            interval_v = self.interval_mesh_vertical
            v_dist, num_vf, num_v = self.get_num_v_co_2(co_01, co_00, interval_v, False)

            if (True): # 縦方向は必ずFit扱いにする
                interval_v *= num_vf / num_v
                num_v += 1
                
            v0_step = (co_01 - co_00) / num_v

            return num_v, v0_step

        def get_co_vis(i, num_v, normal, co_00, co_01, v0_step, vis_offset):
            i_idx = int(i*num_v / (self.fence_support_num+1))
            if (i == self.fence_support_num+1):
                i_idx = num_v - 1
                
            co_vis = (i_idx*co_00 + (num_v-1-i_idx)*co_01)/(num_v-1)
            if (i > 0 and i < self.fence_support_num+1):
                co_vis -= 0.2 * v0_step
                
            #co_vis += normal * self.bevel_depth_pole * (self.pole_radius_ratio_horizontal + 0.6 + self.pole_vis_offset_B)
            co_vis += normal * self.bevel_depth_pole * (self.pole_radius_ratio_horizontal + vis_offset)
            
            return co_vis
            
        def make_vertical(e, e1, index):

            vertices_for_vis_vertical = []
            
            # 端のとその１つ内側
            co0_list = self.get_co_list_from_edge(e, 0, 0)
            co_00 = co0_list[0]; co_01 = co0_list[1]
            normal = self.get_edge_normal(e)

            co1_list = self.get_co_list_from_edge(e1, 0, 0)
            co_10 = co1_list[0]; co_11 = co1_list[1]

            if (self.is_pole_vertical(co_00, co_01, threshold = 1e-4)):
                co_01 += normal * 1e-3
                
            unit_v = (co_01-co_00).normalized()
            normal -= (normal.dot(unit_v)*unit_v)
            normal.normalized()
            
            if (self.pole_vis == True):

                num_v, v0_step = get_wire_mesh_info(co_00, co_01)
                tangent = (co_01-co_11).normalized()
                
                rot_matrix = self.get_rot_matrix_from_2v(tangent, -normal)

                for i in range(0, self.fence_support_num+2):

                    co_vis = get_co_vis(i, num_v, normal, co_00, co_01, v0_step, self.pole_vis_offset_A)
                    #face_index_vis = self.make_instance_plane(rot_matrix, co_vis, vertices_for_vis, faces_for_vis, face_index_vis)
                    vertices_for_vis_vertical += self.set_instancer_verts(rot_matrix, co_vis)
                    

            co_01 += self.extend_pole_top * unit_v
            co_00 -= self.extend_pole_bottom * unit_v

            if (index == 0):
                tilt = self.get_tilt((co_01-co_00).normalized(), unit_z, normal)
                self.set_poly_from3d(curveData, co_00, co_01, tilt)

            if (index == -1):
                tilt = self.get_tilt((co_00-co_01).normalized(), unit_z, normal)
                self.set_poly_from3d(curveData, co_01, co_00, tilt)

            return vertices_for_vis_vertical

        def make_horizontal(idx):

            vertices_for_vis_horizontal = []
            
            points = curveData.splines.new('POLY').points
            points.add(len(edge_list)-1)
            co_pre = None
            num = len(edge_list)

            for i, (e, p) in enumerate(zip(edge_list, points)):
                co_list = self.get_co_list_from_edge(e, 0, 0)
                co_00 = co_list[0]; co_01 = co_list[1]
                normal = self.get_edge_normal(e)
                
                self.set_point_from3d(p, co_list[idx])
                p.radius = self.pole_radius_ratio_horizontal
                if (co_pre != None):
                    tangent = (co_list[idx]-co_pre).normalized()
                    tangent -= tangent.dot(normal)*normal
                    tangent.normalized()

                    p.tilt = self.get_tilt(tangent, unit_z, normal)
                    
                    if (self.pole_vis == True and i != num-1 and idx == 0):
                        
                        num_v, v0_step = get_wire_mesh_info(co_00, co_01)
                        rot_matrix = self.get_rot_matrix_along_tangent(tangent, normal, p.tilt)
                        
                        for i in range(0, self.fence_support_num+2):
                            
                            co_vis = get_co_vis(i, num_v, normal, co_00, co_01, v0_step, 0.6 + self.pole_vis_offset_B)
                            #face_index_vis = self.make_instance_plane(rot_matrix, co_vis, vertices_for_vis, faces_for_vis, face_index_vis)
                            vertices_for_vis_horizontal += self.set_instancer_verts(rot_matrix, co_vis)
                            
                co_pre = co_list[idx]
                
            axis0 = (points[1].co-points[0].co)
            axis0.resize_3d()
            points[0].tilt = self.get_tilt(axis0.normalized(), unit_z, self.get_edge_normal(edge_list[0]))
            points[0].co -= (points[1].co - points[0].co).normalized() * self.extend_pole_rim
            points[-1].co -= (points[-2].co - points[-1].co).normalized() * self.extend_pole_rim

            return vertices_for_vis_horizontal

        for edge_list in edge_list_list:

            if (self.pole_vertical):
                vertices_for_vis += make_vertical(edge_list[0], edge_list[1], 0)
                vertices_for_vis += make_vertical(edge_list[-1], edge_list[-2], -1)

            if (self.pole_not_vertical):
                vertices_for_vis += make_horizontal(1)

                edge_list.reverse()
                vertices_for_vis += make_horizontal(0)

        return vertices_for_vis
        
                
                    
    def make_pole_L_pole(self, curveData, edge_list_list):

        unit_z = Vector((0,0,1))
        
        for edge_list in edge_list_list:

            if (self.pole_vertical):
                for e in edge_list[1:-1]:
                    co_list = self.get_co_list_from_edge(e, 0, 0)

                    co_0 = co_list[0]; co_1 = co_list[1]
                    normal = self.get_edge_normal(e)

                    if (self.is_pole_vertical(co_0, co_1, threshold = 1e-4)):
                        co_1 += normal * 1e-3
                        
                    unit_v = (co_1-co_0).normalized()
                    normal -= (normal.dot(unit_v)*unit_v)
                    normal.normalized()

                    co_1 += self.extend_pole_top * unit_v + self.bevel_depth_pole * self.pole_radius_ratio_horizontal * normal
                    co_0 -= self.extend_pole_bottom * unit_v - self.bevel_depth_pole * self.pole_radius_ratio_horizontal * normal

                    tilt = self.get_tilt((co_1-co_0).normalized(), unit_z, normal)
                    self.set_poly_from3d(curveData, co_0, co_1, tilt)

        


    def make_pole_L(self, context, object, data, material):

        new_objects = []
        
        radius = self.bevel_depth_pole
        curveData = self.make_curve('pole_rim')
        bm = bmesh.from_edit_mesh(data)

        edge_list_list = self.make_edge_list_list(bm)
        for e_list in edge_list_list:
            self.adjust_elist_direction(e_list)

        if (len(edge_list_list) == 0):
            return {'FINISHED'}

        vertices_for_vis = self.make_pole_L_rim(curveData, edge_list_list)

        self.set_bevel_depth(curveData)
        LshapeOB = self.proc_bevel_type(curveData, 'LSHAPE', context.scene)
        
        curveOB = self.make_curve_object(curveData.name, curveData, material, context)
        new_objects.append(curveOB)
        
        if (self.pole_solidify == True):
            mod = curveOB.modifiers.new(name="Solidify", type="SOLIDIFY")
            mod.offset = 1
            mod.use_even_offset = True
            mod.thickness = self.pole_solidify_thickness

        if (vertices_for_vis != []):
            partsOB, visOB, mod_dist = self.make_instancer_vis(context, vertices_for_vis, "vis_shape.A", "fence_vis", 4*radius*self.pole_vis_size)
            new_objects.append(visOB)
            
        LshapeOB.parent = curveOB

        if (self.pole_vertical == False):
            return {'FINISHED'}

        curveData = self.make_curve('pole_pole')
        bm = bmesh.from_edit_mesh(data)
        self.make_pole_L_pole(curveData, edge_list_list)

        if (len(curveData.splines) == 0):
            bpy.data.curves.remove(curveData)
            return {'FINISHED'}
            
        self.set_bevel_depth(curveData)
        LshapeOB = self.proc_bevel_type(curveData, 'LSHAPE_POLE', context.scene)

        curveOB_pole = self.make_curve_object(curveData.name, curveData, material, context)
        new_objects.append(curveOB_pole)
        
        curveOB_pole.parent = curveOB
        if (self.pole_solidify == True):
            mod = curveOB_pole.modifiers.new(name="Solidify", type="SOLIDIFY")
            mod.offset = -1
            mod.use_even_offset = True
            mod.thickness = self.pole_solidify_thickness

        LshapeOB.parent = curveOB_pole
        
        return new_objects


    def make_pole(self, context, object, data, material):

        if (self.pole_bevel_type == 'LSHAPE'):
            return self.make_pole_L(context, object, data, material)

        new_objects = []

        radius = self.bevel_depth_pole
        curveData = self.make_curve('pole')
        bm = bmesh.from_edit_mesh(data)
        
        unit_z = Vector((0,0,1))

        base_co_list = []
        base_tilt_list = []

        wave_length = 0.3 / self.pole_distort_freq

        for e in bm.edges:
            if e.select == False:
                continue

            co0, co1, normal = self.get_edge_info(e)
            is_vertical = self.is_pole_vertical(co0, co1)
            
            if (self.pole_vertical and is_vertical):

                v0, v1, co0, co1 = self.get_sorted_by_z(e, co0, co1)

                unit_v = (v1.co-v0.co).normalized()

                top = False
                e_candidates = self.search_edge_cont(bm, e, v1)

                if not any([e.select for e in e_candidates]):
                    co1 += self.extend_pole_top * unit_v
                    top = True

                bottom = False
                e_candidates = self.search_edge_cont(bm, e, v0)
                if not any([e.select for e in e_candidates]):
                    co0 -= self.extend_pole_bottom * unit_v
                    bottom = True

                tilt = atan2(normal.y, normal.x)
                tilt *= self.tilt_pole

                if (self.pole_distort != 0):
                    polyline = self.set_poly_resample_from3d(curveData, co0, co1, tilt, 0.3*wave_length)
                else:
                    polyline = self.set_poly_from3d(curveData, co0, co1, tilt)
                    if top:
                        v_unit = (co1-co0).normalized()
                        size = self.pole_cap_size
                        if (self.pole_top_cap == "ROUND"):
                            self.append_round_top(polyline, v_unit, co1, tilt, radius*size)

                    if (self.pole_top_cap == 'CAP'):
                        line_list = self.make_cap(curveData, co1, radius * self.pole_cap_size, self.bevel_depth_pole, normal, v_unit, normal, "TOP", self.pole_cap_type)
                        for s in line_list:
                            for p in s.points:
                                p.tilt = tilt

                if bottom:
                    base_co_list.append(co0)
                    base_tilt_list.append(tilt)

            def search_vertical_neighbour(e0, v):
                
                for e in v.link_edges:
                    if e == e0:
                        continue
                        co0, co1, normal = self.get_edge_info(e)
                        if self.is_pole_vertical(co0, co1):
                            return e
                        
                return e0
                            
            if (self.pole_not_vertical and (is_vertical == False)):
                co0 = e.verts[0].co.copy()
                co1 = e.verts[1].co.copy()
                
                tangent = (co1-co0).normalized()

                e_candidates = self.search_edge_cont(bm, e, e.verts[0])
                if not any([e.select for e in e_candidates]):
                    co0 -= self.extend_pole_rim * tangent

                e_candidates = self.search_edge_cont(bm, e, e.verts[1])
                if not any([e.select for e in e_candidates]):
                    co1 += self.extend_pole_rim * tangent

                e_vertical0 = search_vertical_neighbour(e, e.verts[0])
                e_vertical1 = search_vertical_neighbour(e, e.verts[1])
                normal0 = self.get_edge_normal(e_vertical0)
                normal1 = self.get_edge_normal(e_vertical1)

                co0 += normal0 * radius * self.pole_radius_ratio_horizontal * self.pole_offset_horizontal
                co1 += normal1 * radius * self.pole_radius_ratio_horizontal * self.pole_offset_horizontal
                
                tilt = self.get_tilt((co1-co0).normalized(), unit_z, normal)
                tilt *= self.tilt_pole

                if (self.pole_distort !=0):
                    polyline = self.set_poly_resample_from3d(curveData, co0, co1, tilt, 0.3*wave_length, radius = self.pole_radius_ratio_horizontal)
                else:
                    polyline = self.set_poly_from3d(curveData, co0, co1, tilt, radius = self.pole_radius_ratio_horizontal)

        if (self.pole_merge == True):
            self.merge_poly(curveData)

        if (self.pole_distort != 0):
            for s in curveData.splines:
                for p in s.points:
                    co = p.co.copy(); co.resize_3d()
                    noise = mathutils.noise.noise_vector(co / wave_length)
                    p.co.x += noise.x * radius * self.pole_distort
                    p.co.y += noise.y * radius * self.pole_distort
                    p.co.z += noise.z * radius * self.pole_distort
                    
        self.set_bevel_depth(curveData)
        rectOB = self.proc_bevel_type(curveData, self.pole_bevel_type, context.scene)
        curveOB = self.make_curve_object(curveData.name, curveData, material, context)
        if (rectOB != None):
            rectOB.parent = curveOB

        new_objects.append(curveOB)
            
        #if (self.pole_base_type != "NONE" and base_co_list != []):
        #    baseOB, base_shapeOB = self.make_base_instanceB(base_co_list, base_tilt_list, material, context)
        #    new_objects.append(baseOB)
        #    
        #    mod_dist = self.add_GN_modifier(baseOB, "EdgeDistributor.Basic")
        #    scale = Vector((1, 1, self.pole_base_aspect)) * self.pole_header_footer_size
        #    
        #    self.modifier_input(mod_dist, "Object", base_shapeOB)
        #    self.modifier_input(mod_dist, "Scale", scale)

        #    mod_edge = self.add_edge_split_modifier(baseOB)
            
        #    if (self.pole_footer_taper == True):
        #        self.add_taper_modifier(base_shapeOB, self.pole_footer_taper_facror)
                
        return new_objects;


    def make_pole_header_footer(self, context, object, data, material):

        new_objects = []

        bm = bmesh.from_edit_mesh(data)

        co_list = []
        tilt_list = []

        for e in bm.edges:
            if e.select == False:
                continue

            co0, co1, normal = self.get_edge_info(e)
            is_vertical = self.is_pole_vertical(co0, co1)

            if (is_vertical == False):
                continue

            v0, v1, co0, co1 = self.get_sorted_by_z(e, co0, co1)

            unit_v = (v1.co-v0.co).normalized()

            top = False
            e_candidates = self.search_edge_cont(bm, e, v1)
            if not any([e.select for e in e_candidates]):
                co1 += self.extend_pole_top * unit_v
                top = True

            bottom = False
            e_candidates = self.search_edge_cont(bm, e, v0)
            if not any([e.select for e in e_candidates]):
                co0 -= self.extend_pole_bottom * unit_v
                bottom = True

            tilt = atan2(normal.y, normal.x)
            tilt *= self.tilt_pole

            if (self.pole_appendix_type == 'FOOTER'):
                if bottom:
                    co_list.append(co0)
                    tilt_list.append(tilt)

            if (self.pole_appendix_type == 'HEADER'):
                if top:
                    co_list.append(co1)
                    tilt_list.append(tilt)
                
        instancerOB = self.make_header_footer_instance(context, co_list, tilt_list, material, "pole." + self.pole_appendix_type, self.pole_header_footer_size)

        if (instancerOB == None):
            return []
        
        new_objects.append(instancerOB)

        function = None
        dataname = None

        if (self.pole_appendix_type == 'HEADER'):
            if (self.pole_header_type == "REFLECTOR"):
                dataname = {"TYPEA":"appendix_reflector.A",
                            "TYPEB":"appendix_reflector.B",
                            "TYPEC":"appendix_reflector.C"}[self.pole_reflector_type]
            

        if (self.pole_appendix_type == 'FOOTER'):
            if (self.pole_footer_type == 'BLOCK'):
                dataname = 'Cube'
            if (self.pole_footer_type == 'CYLINDER'):
                dataname = 'Cylinder'
            
        instanceeOB = self.make_instanceeOB(context, material, function = function, dataname = dataname)

        if (instanceeOB == None):
            if self.header_footer_custom_target in bpy.data.objects:
                instanceeOB = bpy.data.objects[self.header_footer_custom_target]
        
                
        mod_dist = self.add_GN_modifier(instancerOB, "EdgeDistributor.Taper")

        aspect = 1
        if (self.pole_appendix_type == 'FOOTER'):
            if (self.pole_footer_type in ["BLOCK", "CYLINDER"]):
                aspect = self.pole_base_aspect
                if (self.pole_footer_taper == True):
                    self.modifier_input(mod_dist, "Taper", self.pole_footer_taper_facror)
                    #self.add_taper_modifier(instanceeOB, self.pole_footer_taper_facror)
                    
        scale = Vector((1, 1, aspect)) * self.pole_header_footer_size * self.bevel_depth_pole
        
        self.modifier_input(mod_dist, "Object", instanceeOB)
        self.modifier_input(mod_dist, "Scale", scale)

        #mod_edge = self.add_edge_split_modifier(instancerOB)
        
        return new_objects

    
    def make_pole_connector_ring(self, context, object, data, material):

        new_objects = []

        bm = bmesh.from_edit_mesh(data)
        unit_z = Vector((0,0,1))

        vertices_for_connector = []

        def get_vertices(co, unit_v, f):
            
            vertices = []
            
            normal = f.normal
            bitangent = unit_v
            tangent = normal.cross(bitangent).normalized()

            if (tangent.dot(f.calc_center_bounds() - co0) < 0):
                tangent *= -1

            vertices.append(co)
            vertices.append(co + tangent*0.1)

            return vertices

        for e in bm.edges:
            if e.select == False:
                continue

            co0, co1, normal = self.get_edge_info(e)
            is_vertical = self.is_pole_vertical(co0, co1)

            if is_vertical == False:
                continue

            v0, v1, co0, co1 = self.get_sorted_by_z(e, co0, co1)
            unit_v = (v1.co-v0.co).normalized()

            co_base = self.midpos(co0, co1, self.pole_appendix_ypos)

            if len(e.link_faces) == 2:
                vertices_for_connector += get_vertices(co_base, unit_v, e.link_faces[0])
                vertices_for_connector += get_vertices(co_base, unit_v, e.link_faces[1])
                
            if len(e.link_faces) == 1:
                vertices_for_connector += get_vertices(co_base, unit_v, e.link_faces[0])
                
        bm.free()

        ring_name = {"RING_A":"PoleConnectorRing.A",
                     "RING_B":"PoleConnectorRing.B"}[self.pole_connector_type]

        partsOB, ConnectorOB, mod = self.make_instancer_curve(context, vertices_for_connector, ring_name, "pole_connector.Ring", self.pole_appendix_size, self.pole_appendix_resolution_u, self.bevel_depth)

        bpy.ops.object.mode_set(mode="OBJECT")
        bpy.ops.object.mode_set(mode="EDIT")
        
        new_objects.append(ConnectorOB)

        self.modifier_input(mod, "Translation", Vector((self.bevel_depth_pole,0,0)))
        #force update
        ConnectorOB.data.vertices[0].co[0] = ConnectorOB.data.vertices[0].co[0] 
        
        return new_objects
    
    def get_connector_1face_vertices(self, co0, co1, f, v_offset, rim_type):

        vertices = []

        normal = f.normal
        bitangent = (co1-co0).normalized()
        tangent = normal.cross(bitangent).normalized()
        
        if (tangent.dot(f.calc_center_bounds() - co0) < 0):
            tangent *= -1

        num = self.fence_support_num
        ini = 1
        fin = num+1
        if (self.fill_top == True):
            ini = 0
        if (self.fill_bottom == True):
            fin = num+2
            
        for i in range(ini, fin):

            if (rim_type == "RIM_ONLY"):
                if (i != 0 and i != num+1):
                    continue

            if (rim_type == "EXCLUDE_RIM"):
                if (i == 0 or i == num+1):
                    continue

            pos_v_offset = v_offset * self.bevel_depth * self.pole_appendix_connector_offset
            if (num == 1 and i == 1):
                pos_base = (co0+co1)/2
            else:
                pos_base = (co0*i + co1*(num+1-i)) / (num+1)
                
            #print("v_offset=", v_offset)
            pos_base += bitangent * pos_v_offset
            
            vertices.append(pos_base)
            vertices.append(pos_base + tangent*0.1)

        return vertices


    def make_pole_connector_main(self, context, object, data, type, rim_type):

        col = self.get_asset_colelction(context, "FenceParts")

        if (type == "CLIP"):
            partsOB     = self.load_object(context, "PoleConnector", col)
            partsOBDual = self.load_object(context, "PoleConnectorDual", col)
            offset_overlap = 1

        if (type == "POLE"):
            partsOB = self.load_object(context, "PoleConnector.Pole", col)
            partsOBDual = self.load_object(context, "PoleConnectorDual.Pole", col)
            offset_overlap = 0
            
        bm = bmesh.from_edit_mesh(data)
        unit_z = Vector((0,0,1))

        vertices_for_connector = []
        vertices_for_dual = []

        for e in bm.edges:
            if e.select == False:
                continue

            co0, co1, normal = self.get_edge_info(e)
            is_vertical = self.is_pole_vertical(co0, co1)

            if is_vertical == False:
                continue

            v0, v1, co0, co1 = self.get_sorted_by_z(e, co0, co1)
            unit_v = (v1.co-v0.co).normalized()

            if len(e.link_faces) == 2:
                n1 = e.link_faces[0].normal
                n2 = e.link_faces[1].normal
                if (n1.dot(n2) > 0.98):
                    pre = len(vertices_for_connector)
                    vertices_for_connector += self.get_connector_1face_vertices(co0, co1, e.link_faces[0], v_offset = 0, rim_type = rim_type)
                    post = len(vertices_for_connector)
                    vertices_for_dual += list(range(pre, post))
                else:
                    vertices_for_connector += self.get_connector_1face_vertices(co0, co1, e.link_faces[0], v_offset = -offset_overlap, rim_type = rim_type)
                    vertices_for_connector += self.get_connector_1face_vertices(co0, co1, e.link_faces[1], v_offset = +offset_overlap, rim_type = rim_type)

            if len(e.link_faces) == 1:
                vertices_for_connector += self.get_connector_1face_vertices(co0, co1, e.link_faces[0], v_offset = 0, rim_type = rim_type)
                
        bm.free()
        
        edges = []
        for index in range(0, len(vertices_for_connector), 2):
            edges.append([index, index+1])

        mesh_name = "pole_connector." + type
            
        mesh = bpy.data.meshes.new(mesh_name)
        mesh.from_pydata(vertices_for_connector, edges, [])

        bpy.ops.object.mode_set(mode="OBJECT")
        bpy.ops.object.mode_set(mode="EDIT")
        
        ConnectorOB = self.make_object(mesh, None, context)
        ConnectorOB.vertex_groups.new(name="Dual")

        vg = ConnectorOB.vertex_groups['Dual']
        vg.add(vertices_for_dual, 1, 'REPLACE')
        
        mod_dist = self.add_GN_modifier(ConnectorOB, "EdgeDistributor.Connector")
        self.modifier_attrib(mod_dist, "Ring", "Ring")
        self.modifier_attrib(mod_dist, "Flat", "Flat")
        self.modifier_attrib(mod_dist, "Dual", "Dual")
        self.modifier_attrib(mod_dist, "ExtendPlus", "ExtendPlus")
        self.modifier_attrib(mod_dist, "ExtendMinus", "ExtendMinus")
        self.modifier_attrib(mod_dist, "ScaleZ", "ScaleZ")
        self.modifier_attrib(mod_dist, "ScaleXZ", "ScaleXZ")
        self.modifier_attrib(mod_dist, "ScaleXYZ", "ScaleXYZ")
        self.modifier_input(mod_dist, "Radius", float(self.bevel_depth))
        self.modifier_input(mod_dist, "Scale", float(self.bevel_depth_pole))
        self.modifier_input(mod_dist, "ScaleB", float(self.pole_appendix_connector_scaleB))
        if (type == "CLIP"):
            self.modifier_input(mod_dist, "Extend", float(self.pole_appendix_connector_extend))
        if (type == "POLE"):
            self.modifier_input(mod_dist, "Extend", float(self.pole_appendix_connector_extend_for_ring))
        self.modifier_input(mod_dist, "Object", partsOB)
        self.modifier_input(mod_dist, "ObjectDual", partsOBDual)

        return ConnectorOB
        
    
    def make_pole_connector(self, context, object, data, material):

        if ("RING" in self.pole_connector_type):
            return self.make_pole_connector_ring(context, object, data, material)

        new_objects = []

        if (self.pole_appendix_connector_for_ring == True):
            objClip = self.make_pole_connector_main(context, object, data, "CLIP", "EXCLUDE_RIM")
            objPole = self.make_pole_connector_main(context, object, data, "POLE", "RIM_ONLY")
            new_objects.append(objClip)
            new_objects.append(objPole)
        else:
            objClip = self.make_pole_connector_main(context, object, data, "CLIP", "ALL")
            new_objects.append(objClip)
        
        return new_objects


    def make_pole_support(self, context, object, data, material):

        new_objects = []

        curveData = self.make_curve('pole_support')
        self.set_bevel_depth(curveData)

        bm = bmesh.from_edit_mesh(data)
        
        unit_z = Vector((0,0,1))

        #app_co_list = []
        #app_tilt_list = []
        #app_normal_list = []
        #app_v_unit_list = []

        for e in bm.edges:
            if e.select == False:
                continue

            co0, co1, normal = self.get_edge_info(e)
            is_vertical = self.is_pole_vertical(co0, co1)

            if is_vertical == False:
                continue

            v0, v1, co0, co1 = self.get_sorted_by_z(e, co0, co1)
            unit_v = (v1.co-v0.co).normalized()
            
            bottom = False
            e_candidates = self.search_edge_cont(bm, e, v0)
            if not any([e.select for e in e_candidates]):
                co0 -= self.extend_pole_bottom * unit_v
                bottom = True

            if bottom == False:
                continue

            angle = 3.14159/2 * self.pole_support_angle
            co_top = co0 + unit_v * self.pole_support_size
            
            if (self.pole_support_offset == 0):
                co_base = co_top - (cos(angle) * unit_v + sin(angle) * normal) / cos(angle) * (1+self.pole_support_extend) * self.pole_support_size

                self.set_poly_from3d(curveData, co_base, co_top, 0, radius = 1)

            else:
                polyline = curveData.splines.new('POLY')
                points = polyline.points
                points.add(2)
                
                co_mid = co_top - self.pole_support_offset * normal * self.pole_support_size
                co_base = co_mid - (cos(angle) * unit_v + sin(angle) * normal) / cos(angle) * (1+self.pole_support_extend) * self.pole_support_size
                
                self.set_point_from3d(points[0], co_base)
                self.set_point_from3d(points[1], co_mid)
                self.set_point_from3d(points[2], co_top)


        bevel_type = self.pole_bevel_type
        if (bevel_type == 'LSHAPE'):
            bevel_type = 'LSHAPE_POLE'
        self.proc_bevel_type(curveData, bevel_type, context.scene, 'POLE_SUPPORT')
        if (bevel_type == 'LSHAPE_POLE'):
            self.set_tilt_all(curveData, -3.14159/2)
            

        supportOB = self.make_curve_object('pole_support', curveData, material, context)
        new_objects.append(supportOB)

        if (bevel_type == 'LSHAPE_POLE' and self.pole_solidify == True):
            mod = supportOB.modifiers.new(name="Solidify", type="SOLIDIFY")
            mod.offset = 1
            mod.use_even_offset = True
            mod.thickness = self.pole_solidify_thickness
        
        return new_objects

    def make_pole_appendix(self, context, object, data, material):

        new_objects = []

        #print(self.pole_appendix_type)

        if (self.pole_appendix_type == 'REFLECTOR'):
            new_objects += self.make_pole_reflector(context, object, data, material)


            
        if (self.pole_appendix_type in ['HEADER', 'FOOTER']):
            new_objects += self.make_pole_header_footer(context, object, data, material)

        if (self.pole_appendix_type == 'CONNECTOR'):
            new_objects += self.make_pole_connector(context, object, data, material)

        if (self.pole_appendix_type == 'SUPPORT'):
            new_objects += self.make_pole_support(context, object, data, material)

        return new_objects
    
    def execute(self, context):

        object = context.object
        data = object.data
        material = self.get_material(object)

        if (self.action == 'POLE'):
            new_objects = self.make_pole(context, object, data, material)

        if (self.action == 'POLE_APPENDIX'):
            new_objects = self.make_pole_appendix(context, object, data, material)

        #print(new_objects)
        self.move_objects_to_collection(context, object, new_objects)
        self.set_parenting_(object, new_objects)
        
        return {'FINISHED'}
        
    def make_fence_connect_shape(self):

        self.load_mesh("fence_connect_shape")
        return bpy.data.meshes["fence_connect_shape"]
